//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_ANP_DCSweep.h,v $
//
// Purpose        : DC sweep analysis class
//
// Special Notes  : Specify any "hidden" or subtle details of the class here.
//                  Portability details, error handling information, etc.
//
// Creator        : Richard Schiek, SNL, Electrical and Microsystem Modeling
//
// Creation Date  : 01/24/08
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.57 $
//
// Revision Date  : $Date: 2015/10/27 19:24:39 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_ANP_DCSweep_h
#define Xyce_N_ANP_DCSweep_h

// ----------   Xyce Includes   ----------
#include <N_ANP_fwd.h>
#include <N_TOP_fwd.h>

#include <N_ANP_AnalysisBase.h>
#include <N_ANP_RegisterAnalysis.h>
#include <N_UTL_OptionBlock.h>

namespace Xyce {
namespace Analysis {

//-------------------------------------------------------------------------
// Class         : DCSweep
// Purpose       : Transient analysis class
// Special Notes :
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//-------------------------------------------------------------------------
class DCSweep : public AnalysisBase
{
public:
  DCSweep(
    AnalysisManager &                   analysis_manager,
    Nonlinear::Manager &                nonlinear_manager,
    Loader::Loader &                    loader,
    Topo::Topology &                    topology,
    IO::InitialConditionsManager &      initial_conditions_manager,
    HB *                                hb_analysis = 0);  // HACK TO GET US MOVING FORWARD, should pass output adapter or something

  virtual ~DCSweep();

  void setTIAParams(const TimeIntg::TIAParams &tia_params)
  {
    tiaParams_ = tia_params;
  }

  const TimeIntg::TIAParams &getTIAParams() const
  {
    return tiaParams_;
  }

  TimeIntg::TIAParams &getTIAParams()
  {
    return tiaParams_;
  }

  bool setAnalysisParams(const Util::OptionBlock & paramsBlock); // override;
  bool setTimeIntegratorOptions(const Util::OptionBlock &option_block);

  bool outputFailureStats(std::ostream &os);

  bool getDCOPFlag() const // override
  {
    return true;
  }

protected:
  bool doRun(); // override;
  bool doInit(); // override;
  bool doLoopProcess(); // override;
  bool doProcessSuccessfulStep(); // override;
  bool doProcessFailedStep(); // override;
  bool doFinish(); // override;
  bool doHandlePredictor(); // override;

public:
  // Two Level specific
  bool twoLevelStep();

  void printStepHeader(std::ostream &os);
  bool printLoopInfo(int start, int finish);

private:
  void dcSweepOutput();
  void initializeSolution_();
  void takeStep_ ();

private:
  AnalysisManager &                     analysisManager_;
  Loader::Loader &                      loader_;
  Nonlinear::Manager &                  nonlinearManager_;
  Topo::Topology &                      topology_;
  IO::InitialConditionsManager &        initialConditionsManager_;
  OutputMgrAdapter &                    outputManagerAdapter_;
  HB *                                  hbAnalysis_;
  TimeIntg::TIAParams                   tiaParams_;
  bool                                  sensFlag_;
  bool                                  dcLoopInitialized_;
  int                                   dcLoopSize_;
  std::vector<int>                      dcSweepFailures_;
  SweepVector                           dcSweepVector_;
  std::vector<double>                   objectiveVec_;
  std::vector<double>                   dOdpVec_;
  std::vector<double>                   dOdpAdjVec_;
  std::vector<double>                   scaled_dOdpVec_;
  std::vector<double>                   scaled_dOdpAdjVec_;
};

bool registerDCSweepFactory(FactoryBlock &factory_block);

} // namespace Analysis
} // namespace Xyce

#endif
