//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-NOISE04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_ANP_NOISE.C,v $
// Purpose       : NOISE analysis functions.
// Special Notes :
// Creator       : Eric Keiter
// Creation Date :  12/8/2014
//
// Revision Information:
// ---------------------
// Revision Number: $Revision: 1.45 $
// Revision Date  : $Date: 2015/09/03 22:11:39 $
// Current Owner  : $Author: rlschie $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iomanip>

#include <fstream>

#include <N_ANP_NOISE.h>
#include <N_ANP_NoiseData.h>

#include <N_ANP_AnalysisManager.h>
#include <N_ANP_OutputMgrAdapter.h>
#include <N_ERH_Message.h>
#include <N_ERH_ErrorMgr.h>
#include <N_IO_CircuitBlock.h>
#include <N_IO_CmdParse.h>
#include <N_IO_InitialConditions.h>
#include <N_IO_OptionBlock.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_SpiceSeparatedFieldTool.h>
#include <N_LAS_BlockMatrix.h>
#include <N_LAS_BlockSystemHelpers.h>
#include <N_LAS_BlockVector.h>
#include <N_LAS_Builder.h>
#include <N_LAS_Matrix.h>
#include <N_LAS_MultiVector.h>
#include <N_LAS_System.h>
#include <N_LOA_Loader.h>
#include <N_NLS_Manager.h>

#include <N_TIA_DataStore.h>
#include <N_TIA_WorkingIntegrationMethod.h>
#include <N_TIA_StepErrorControl.h>
#include <N_TIA_NoTimeIntegration.h>

#include <N_UTL_Diagnostic.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_LogStream.h>
#include <N_UTL_Math.h>
#include <N_UTL_Timer.h>
#include <N_PDS_Comm.h>
#include <N_PDS_MPI.h>
#include <N_PDS_Manager.h>
#include <N_PDS_Serial.h>

#include <N_PDS_ParMap.h>

#include <N_TOP_Topology.h>

#include <Epetra_SerialComm.h>
#include <Epetra_Map.h>
#include <Epetra_BlockMap.h>
#include <Epetra_CrsMatrix.h>
#include <Epetra_CrsGraph.h>
#include <Epetra_MultiVector.h>
#include <Epetra_Vector.h>
#include <Epetra_Export.h>
#include <Epetra_LinearProblem.h>
#include <Amesos.h>

#include<N_UTL_ExtendedString.h>
#include<N_UTL_ExpressionData.h>
#include<N_NLS_ReturnCodes.h>

#include <Teuchos_RCP.hpp>
using Teuchos::RCP;
using Teuchos::rcp;

// putting these here for now (from spice3's noisedefs.h file)
#define N_MINLOG          1E-38       /* the smallest number we can take the log of */
#define N_MINGAIN         1E-20  // the smallest input-output gain we can tolerate 
                                 // (to calculate input-referred noise we divide 
					                       // the output noise by the gain) 

#define N_INTFTHRESH   1E-10       // the largest slope (of a log-log noise spectral 
					   //    density vs. freq plot) at which the noise 
					   //    spectum is still considered flat. (no need for 
					   //    log curve fitting) 
#define N_INTUSELOG      1E-10       // decides which expression to use for the integral of 
					   //    x**k.  If k is -1, then we must use a 'ln' form. 
					   //    Otherwise, we use a 'power' form.  This 
					   //    parameter is the region around (k=) -1 for which 
					   //    use the 'ln' form. 

namespace Xyce {
namespace Analysis {
    
//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setTimeIntegratorOptions
// Purpose       :
// Special Notes : These are from '.options timeint'
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/18/02
//-----------------------------------------------------------------------------
bool NOISE::setTimeIntegratorOptions(
  const Util::OptionBlock &     option_block)
{
  for (Util::ParamList::const_iterator it = option_block.begin(), 
      end = option_block.end(); it != end; ++it)
  {
    const Util::Param &param = *it;

    if (param.uTag() == "DEBUGLEVEL" )
    {
      IO::setTimeIntegratorDebugLevel(analysisManager_.getCommandLine(), param.getImmutableValue<int>());
    }
    else if (nonlinearManager_.setReturnCodeOption(param))
    {
      ;
    }
    else if (tiaParams_.setTimeIntegratorOption(param))
    {
      ;
    }
    else if (setDCOPOption(param))
    {
      ;
    }
    else
    {
      Report::UserError() << param.uTag() 
        << " is not a recognized time integration option";
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::NOISE
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
NOISE::NOISE(
  AnalysisManager &                     analysis_manager,
  Linear::System &                      linear_system,
  Nonlinear::Manager &                  nonlinear_manager,
  Loader::Loader &                      loader,
  Topo::Topology &                      topology,
  IO::InitialConditionsManager &        initial_conditions_manager)
  : AnalysisBase(analysis_manager, "NOISE"),
    StepEventListener(&analysis_manager),
    analysisManager_(analysis_manager),
    loader_(loader),
    linearSystem_(linear_system),
    nonlinearManager_(nonlinear_manager),
    topology_(topology),
    initialConditionsManager_(initial_conditions_manager),
    outputManagerAdapter_(analysis_manager.getOutputManagerAdapter()),
    bVecRealPtr(linearSystem_.builder().createVector()),
    bVecImagPtr(linearSystem_.builder().createVector()),
    bNoiseVecRealPtr(linearSystem_.builder().createVector()),
    bNoiseVecImagPtr(linearSystem_.builder().createVector()),
    hackOutputCalledBefore_(false),
    useAdjoint_(true),
    outputNodeSingle_(true),
    outputNode1_(""),
    outputNode2_(""),
    specifiedSource_(""),
    ACMatrix_(0),
    B_(0),
    X_(0),
    saved_AC_X_(0),
    blockSolver_(0),
    blockProblem_(0),
    noiseLoopSize_(0),
    type_("DEC"),
    stepFlag_(false),
    np_(10.0),
    fStart_(1.0),
    fStop_(1.0),
    stepMult_(0.0),
    fstep_(0.0),
    pts_per_summary_(0),
    pts_per_summary_Given(false),
    delFreq_(0.0),
    lastFreq_(0.0),
    currentFreq_(0.0),
    lnFreq_(0.0),
    lnLastFreq_(0.0),
    delLnFreq_(0.0),
    GainSqInv_(0.0),
    lnGainInv_(0.0),
    totalOutputNoise_(0.0),
    totalInputNoise_(0.0),
    totalOutputNoiseDens_(0.0),
    totalInputNoiseDens_(0.0)
{
  bVecRealPtr->putScalar(0.0);
  bVecImagPtr->putScalar(0.0);
  bNoiseVecRealPtr->putScalar(0.0);
  bNoiseVecImagPtr->putScalar(0.0);
}

//-----------------------------------------------------------------------------
// Function      : NOISE::~NOISE()
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
NOISE::~NOISE()
{
  delete bVecRealPtr;
  delete bVecImagPtr;
  delete bNoiseVecRealPtr;
  delete bNoiseVecImagPtr;
  delete ACMatrix_;
  delete B_;
  delete X_;
  delete saved_AC_X_;
  delete blockSolver_;
  delete blockProblem_;

  int size = noiseDataVec_.size();
  for (int i=0;i<size;++i)
  {
    delete (noiseDataVec_[i]);
  }
  noiseDataVec_.clear();
}

//-----------------------------------------------------------------------------
// Function      : NOISE::notify
// Purpose       :
// Special Notes : 
// Scope         : public
// Creator       : 
// Creation Date : 
//-----------------------------------------------------------------------------
void NOISE::notify(const StepEvent &event) 
{
  if (event.state_ == StepEvent::STEP_STARTED)
  {
    AnalysisBase::resetForStepAnalysis();

    stepFlag_ = true;
    analysisManager_.getStepErrorControl().resetAll(tiaParams_);

    bVecRealPtr->putScalar(0.0);
    bVecImagPtr->putScalar(0.0);
    bNoiseVecRealPtr->putScalar(0.0);
    bNoiseVecImagPtr->putScalar(0.0);
  }
}

//-----------------------------------------------------------------------------
// Function      : NOISE::setAnalysisParams
// Purpose       :
// Special Notes : These are from the .NOISE statement.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::setAnalysisParams(const Util::OptionBlock & paramsBlock)
{
  for (Util::ParamList::const_iterator it = paramsBlock.begin(), 
      end = paramsBlock.end(); it != end; ++it)
  {
    if ((*it).uTag() == "V")
    {
      if ((*it).getImmutableValue<double>()==1.0)
      {
        outputNodeSingle_ = true;
        Util::ParamList::const_iterator itNode = it;
        itNode++;
        outputNode1_ = (*itNode).uTag();
      }
      else if ((*it).getImmutableValue<double>()==2.0)
      {
        outputNodeSingle_ = false;
        Util::ParamList::const_iterator itNode = it;
        itNode++;
        outputNode1_ = (*itNode).uTag();
        itNode++;
        outputNode2_ = (*itNode).uTag();
      }
    }
    else if ((*it).uTag() == "SOURCE")
    {
      specifiedSource_ = (*it).stringValue();
    }
    else if ((*it).uTag() == "TYPE")
    {
      type_ = (*it).stringValue();
    }
    else if ((*it).uTag() == "NP")
    {
      np_ = (*it).getImmutableValue<double>();
    }
    else if ((*it).uTag() == "FSTART")
    {
      fStart_ = (*it).getImmutableValue<double>();
    }
    else if ((*it).uTag() == "FSTOP")
    {
      fStop_ = (*it).getImmutableValue<double>();
    }
    else if ((*it).uTag() == "PTS_PER_SUMMARY")
    {
      pts_per_summary_ = (*it).getImmutableValue<int>();
    }
  }

  if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
  {
    dout() << section_divider << std::endl
           << "NOISE simulation parameters" 
           << std::endl;

    if (outputNodeSingle_)
    {
      dout() << "Output Node: V(" << outputNode1_ << ")" <<std::endl;
    }
    else
    {
      dout() << "Output Node: V(" << outputNode1_ << ","<<outputNode2_<<")" <<std::endl;
    }

    dout() << "specified source = " << specifiedSource_ << std::endl
           << "number of points  = " << np_ << std::endl
           << "starting frequency = " << fStart_ << std::endl
           << "stop frequency = " << fStop_ << std::endl
           << "pts_per_summary = " << pts_per_summary_ 
             << std::endl;
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::getDCOPFlag()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::getDCOPFlag() const
{
  return getIntegrationMethod() == TimeIntg::NoTimeIntegration::type;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::run()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::doRun()
{
  return doInit() && doLoopProcess() && doFinish();
}

//-----------------------------------------------------------------------------
// Function      : NOISE::init()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::doInit()
{
  bool bsuccess = true;

  noiseLoopSize_ = setupSweepParam_();

  // Get set to do the operating point.
  baseIntegrationMethod_ = TimeIntg::NoTimeIntegration::type;
  analysisManager_.createTimeIntegratorMethod(tiaParams_, baseIntegrationMethod_);

  stepNumber = 0;
  setDoubleDCOPEnabled(loader_.isPDESystem());

  static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish
    (AnalysisEvent(AnalysisEvent::INITIALIZE, AnalysisEvent::NOISE_IC));

  // set initial guess, if there is one to be set.
  // this setInitialGuess call is to up an initial guess in the
  // devices that have them (usually PDE devices).  This is different than
  // the "intializeProblem" call, which sets IC's.  (initial conditions are
  // different than initial guesses.
  loader_.setInitialGuess (analysisManager_.getDataStore()->nextSolutionPtr);

  // If available, set initial solution (.IC, .NODESET, etc).
  // setInputOPFlag(
  //   outputManagerAdapter_.setupInitialConditions(
  //     *analysisManager_.getDataStore()->nextSolutionPtr,
  //     *analysisManager_.getDataStore()->flagSolutionPtr));

  setInputOPFlag(
    initialConditionsManager_.setupInitialConditions(outputManagerAdapter_.getComm(),
                                                     topology_.getSolutionNodeNameMap(),
                                                     *analysisManager_.getDataStore()->nextSolutionPtr,
                                                     *analysisManager_.getDataStore()->flagSolutionPtr));

  // Set a constant history for operating point calculation
  analysisManager_.getDataStore()->setConstantHistory();
  analysisManager_.getDataStore()->computeDividedDifferences();
  analysisManager_.getWorkingIntegrationMethod().obtainCorrectorDeriv();

  // solving for DC op
  doHandlePredictor();
  loader_.updateSources();
  analysisManager_.getStepErrorControl().newtonConvergenceStatus = nonlinearManager_.solve();
  analysisManager_.getDataStore()->stepLinearCombo ();
  gatherStepStatistics(stats_, nonlinearManager_.getNonlinearSolver(), analysisManager_.getStepErrorControl().newtonConvergenceStatus);
  analysisManager_.getStepErrorControl().evaluateStepError(loader_, tiaParams_);

  if ( analysisManager_.getStepErrorControl().newtonConvergenceStatus <= 0)
  {
    Report::UserError() << "Solving for DC operating point failed! Cannot continue NOISE analysis";
    return false;
  }

  // only output DC op if the .op was specified in the netlist
  // or if this NOISE analysis is called from a .step loop
  if ( stepFlag_ || analysisManager_.getDotOpSpecified() ) 
  {
    outputManagerAdapter_.dcOutput(
        stepNumber, 
        *analysisManager_.getDataStore()->nextSolutionPtr, 
        *analysisManager_.getDataStore()->nextStatePtr, 
        *analysisManager_.getDataStore()->nextStorePtr,
        *analysisManager_.getDataStore()->nextLeadCurrentPtr,
        *analysisManager_.getDataStore()->nextLeadDeltaVPtr,
        *analysisManager_.getDataStore()->nextLeadCurrentQDerivPtr,
        objectiveVec_, 
          dOdpVec_, dOdpAdjVec_, scaled_dOdpVec_, scaled_dOdpAdjVec_);
    outputManagerAdapter_.finishOutput();
  }

  // This for saving the data from the DC op.  different from the above where we are
  // concerned with generating normal output.
  // outputManagerAdapter_.outputDCOP(*analysisManager_.getDataStore()->nextSolutionPtr);
  initialConditionsManager_.outputDCOP(outputManagerAdapter_.getComm(), topology_.getSolutionNodeNameMap(), *analysisManager_.getDataStore()->nextSolutionPtr);

  loader_.loadBVectorsforAC (bVecRealPtr, bVecImagPtr);

  static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish
    (AnalysisEvent(AnalysisEvent::FINISH, AnalysisEvent::NOISE_IC));

  processOutputNodes ();

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::loopProcess()
// Purpose       : Conduct the stepping loop.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::doLoopProcess()
{
  analysisManager_.getDataStore()->daeQVectorPtr->putScalar(0.0);
  analysisManager_.getDataStore()->daeFVectorPtr->putScalar(0.0);

  analysisManager_.getDataStore()->dFdxdVpVectorPtr->putScalar(0.0);
  analysisManager_.getDataStore()->dQdxdVpVectorPtr->putScalar(0.0);
  analysisManager_.getDataStore()->dQdxMatrixPtr->put(0.0);
  analysisManager_.getDataStore()->dFdxMatrixPtr->put(0.0);

  loader_.updateState
                ((analysisManager_.getDataStore()->nextSolutionPtr),
               (analysisManager_.getDataStore()->currSolutionPtr),
               (analysisManager_.getDataStore()->lastSolutionPtr),
               (analysisManager_.getDataStore()->nextStatePtr),
               (analysisManager_.getDataStore()->currStatePtr),
               (analysisManager_.getDataStore()->lastStatePtr),
               (analysisManager_.getDataStore()->nextStorePtr),
               (analysisManager_.getDataStore()->currStorePtr),
               (analysisManager_.getDataStore()->lastStorePtr)
               );

  loader_.loadDAEVectors
              ((analysisManager_.getDataStore()->nextSolutionPtr),
               (analysisManager_.getDataStore()->currSolutionPtr),
               (analysisManager_.getDataStore()->lastSolutionPtr),
               (analysisManager_.getDataStore()->nextStatePtr),
               (analysisManager_.getDataStore()->currStatePtr),
               (analysisManager_.getDataStore()->lastStatePtr),
               (analysisManager_.getDataStore()->nextStateDerivPtr),
               (analysisManager_.getDataStore()->nextStorePtr),
               (analysisManager_.getDataStore()->currStorePtr),
               (analysisManager_.getDataStore()->lastStorePtr),
               (analysisManager_.getDataStore()->nextStoreLeadCurrQPtr),
               (analysisManager_.getDataStore()->nextLeadCurrentPtr),
               (analysisManager_.getDataStore()->currLeadCurrentPtr),
               (analysisManager_.getDataStore()->lastLeadCurrentPtr),
               (analysisManager_.getDataStore()->nextLeadCurrentQDerivPtr),
               (analysisManager_.getDataStore()->nextLeadDeltaVPtr),
               (analysisManager_.getDataStore()->currLeadDeltaVPtr),
               (analysisManager_.getDataStore()->lastLeadDeltaVPtr),
               (analysisManager_.getDataStore()->daeQVectorPtr),
               (analysisManager_.getDataStore()->daeFVectorPtr),
               (analysisManager_.getDataStore()->daeBVectorPtr),
               (analysisManager_.getDataStore()->dFdxdVpVectorPtr),
               (analysisManager_.getDataStore()->dQdxdVpVectorPtr) );

  loader_.loadDAEMatrices(analysisManager_.getDataStore()->nextSolutionPtr,
      analysisManager_.getDataStore()->nextStatePtr, 
      analysisManager_.getDataStore()->nextStateDerivPtr,
      analysisManager_.getDataStore()->nextStorePtr,
      analysisManager_.getDataStore()->dQdxMatrixPtr,  
      analysisManager_.getDataStore()->dFdxMatrixPtr);

  C_ = analysisManager_.getDataStore()->dQdxMatrixPtr;
  G_ = analysisManager_.getDataStore()->dFdxMatrixPtr;

#if 0
  if (DEBUG_TIME && isActive(Diag::TIME_PARAMETERS))
  {
    Xyce::dout() << "dQdxMatrixPtr:" << std::endl;
    analysisManager_.getDataStore()->dQdxMatrixPtr->printPetraObject( Xyce::dout() );

    Xyce::dout() << "dFdxMatrixPtr:" << std::endl;
    analysisManager_.getDataStore()->dFdxMatrixPtr->printPetraObject( Xyce::dout() );

    Xyce::dout() << std::endl;
  }
#endif

  createACLinearSystem_();

  static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish
    (AnalysisEvent(AnalysisEvent::INITIALIZE, AnalysisEvent::NOISE));

  int numNoiseDevices = loader_.getNumNoiseDevices();
  noiseDataVec_.resize(numNoiseDevices);
  for (int i=0;i<numNoiseDevices;++i)
  {
    noiseDataVec_[i] = new NoiseData();
  }

  loader_.setupNoiseSources(noiseDataVec_);

  // clear out the integral arrays
  for (int i=0;i<numNoiseDevices;++i)
  {
    int numNoiseThisDevice = noiseDataVec_[i]->numSources;
    for (int j=0;j<numNoiseThisDevice;++j)
    {
      noiseDataVec_[i]->inputNoiseTotal[j] = 0.0;
      noiseDataVec_[i]->outputNoiseTotal[j] = 0.0;
    }
  }

  // loop over all the specified frequencies, do an AC solve and a NOISE solve at each.
  if (useAdjoint_)
  {
    setupAdjointRHS_();
  }

  N_PDS_Manager &pds_manager = *analysisManager_.getPDSManager();
  N_PDS_Comm & comm = *(pds_manager.getPDSComm());
  int myPID = comm.procID();

  ///////////////////////////////////////////////////////////////////////////
  // frequency loop
  for (int currentStep = 0; currentStep < noiseLoopSize_; ++currentStep)
  {
    // solve the AC system, to get up-to-date currents and voltages
    updateCurrentFreq_(currentStep);

    resetACLinearSystem_();

    static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish
      (AnalysisEvent(AnalysisEvent::STEP_STARTED, AnalysisEvent::NOISE, currentFreq_, currentStep));

    bool stepAttemptStatus;
    {
      Stats::StatTop _nonlinearStat("Nonlinear Solve");
      Stats::TimeBlock _nonlinearTimer(_nonlinearStat);

      stepAttemptStatus = solveACLinearSystem_();
    }

    // save a copy of X_ (the AC solution, already computed), for output purposes, etc.
    delete saved_AC_X_;
    saved_AC_X_ = new Linear::BlockVector (*X_);

    // Compute AC gain.  
    Linear::Vector & Xreal = X_->block( 0 );
    Linear::Vector & Ximag = X_->block( 1 );

    double v1r = 0.0;
    double v1i = 0.0;
    double v2r = 0.0;
    double v2i = 0.0;

    //comm.barrier();
    int root=-1;
    if (outputVarGIDs_.size()>0)
    {
      if (outputVarGIDs_[0] > -1)
      {
        v1r = Xreal.getElementByGlobalIndex(outputVarGIDs_[0]); 
        v1i = Ximag.getElementByGlobalIndex(outputVarGIDs_[0]);
        root = myPID;
      }
      Xyce::Parallel::AllReduce(comm.comm(), MPI_MAX, &root, 1);
      comm.bcast( &v1r, 1, root );
      comm.bcast( &v1i, 1, root );
    }

    root=-1;
    if (outputVarGIDs_.size()>1)
    {
      if (outputVarGIDs_[1] > -1)
      {
        v2r = Xreal.getElementByGlobalIndex(outputVarGIDs_[1]);
        v2i = Ximag.getElementByGlobalIndex(outputVarGIDs_[1]);
        root = myPID;
      }
      Xyce::Parallel::AllReduce(comm.comm(), MPI_MAX, &root, 1);
      comm.bcast( &v2r, 1, root );
      comm.bcast( &v2i, 1, root );
    }
    //comm.barrier();

    double realVal = v1r-v2r;
    double imagVal = v1i-v2i;
    GainSqInv_ = 1.0 / std::max(((realVal*realVal) + (imagVal*imagVal)),N_MINGAIN);
    lnGainInv_ = std::log(GainSqInv_);

    // save previous (last) lnNoise densities
    for (int i=0;i<noiseDataVec_.size();++i)
    {
      int numNoiseThisDevice = noiseDataVec_[i]->numSources;
      for (int j=0;j<numNoiseThisDevice;++j)
      {
        noiseDataVec_[i]->lastLnNoiseDens[j] = noiseDataVec_[i]->lnNoiseDens[j];
      }
    }

    // do NOISE analysis for this frequency.
    if (useAdjoint_)
    {
      resetAdjointNOISELinearSystem_();
      solveAdjointNOISE_();
    }
    else// only here as diagnostic, not implemented yet.
    {
      solveDirectNOISE_();
    }

    // perform total noise integrals
    if (currentStep != 0)
    {
      for (int i=0;i<noiseDataVec_.size();++i)
      {
        int numNoiseThisDevice = noiseDataVec_[i]->numSources;
        for (int j=0;j<numNoiseThisDevice;++j)
        {
          double noizDens = noiseDataVec_[i]->outputNoiseDens[j];
          double lnDens = noiseDataVec_[i]->lnNoiseDens[j];
          double lnlastDens = noiseDataVec_[i]->lastLnNoiseDens[j];

          double tempOutNoise = noiseIntegral( noizDens, lnDens, lnlastDens,
                 delLnFreq_, delFreq_, lnFreq_, lnLastFreq_);

          double tempInNoise = noiseIntegral( 
                 noizDens * GainSqInv_, 
                 lnDens + lnGainInv_,
                 lnlastDens + lnGainInv_,
                 delLnFreq_, delFreq_, lnFreq_, lnLastFreq_);


          noiseDataVec_[i]->outputNoiseTotal[j] += tempOutNoise;
          noiseDataVec_[i]->inputNoiseTotal[j] += tempInNoise;

          totalOutputNoise_+=tempOutNoise;
          totalInputNoise_+=tempInNoise;
        }
      }
    }
   
    // process success/failure 
    if (stepAttemptStatus)
    {
      static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish
        (AnalysisEvent(AnalysisEvent::STEP_SUCCESSFUL, AnalysisEvent::NOISE, currentFreq_, currentStep));
      doProcessSuccessfulStep();
    }
    else // stepAttemptStatus  (ie do this if the step FAILED)
    {
      static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish
        (AnalysisEvent(AnalysisEvent::STEP_FAILED, AnalysisEvent::NOISE, currentFreq_, currentStep));
      doProcessFailedStep();
    }
  }

  //Report::safeBarrier(comm.comm());
  Xyce::Parallel::AllReduce(comm.comm(), MPI_SUM, &totalOutputNoise_, 1);
  Xyce::Parallel::AllReduce(comm.comm(), MPI_SUM, &totalInputNoise_, 1);

  // Outputs to the screen:
  Xyce::lout() << "Total Output Noise = " << totalOutputNoise_ <<std::endl;
  Xyce::lout() << "Total Input Noise = " << totalInputNoise_ <<std::endl;

  static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish
    (AnalysisEvent(AnalysisEvent::FINISH, AnalysisEvent::NOISE));

  return true;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::createACLinearSystem_()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::createACLinearSystem_()
{
  bool bsuccess = true;

  N_PDS_Manager &pds_manager = *analysisManager_.getPDSManager();

  RCP<N_PDS_ParMap> baseMap = rcp(pds_manager.getParallelMap( Parallel::SOLUTION ), false);
  Epetra_CrsGraph &baseFullGraph = *pds_manager.getMatrixGraph(Parallel::JACOBIAN);

  int numBlocks = 2;

  RCP<N_PDS_ParMap> blockMap = Linear::createBlockParMap(numBlocks, *baseMap);

  // Create a block vector
  delete B_;
  B_ = new Linear::BlockVector(numBlocks, blockMap, baseMap);

  // -----------------------------------------------------
  // Now test block graphs.
  // -----------------------------------------------------

  std::vector<std::vector<int> > blockPattern(2);
  blockPattern[0].resize(2);
  blockPattern[0][0] = 0; blockPattern[0][1] = 1;
  blockPattern[1].resize(2);
  blockPattern[1][0] = 0; blockPattern[1][1] = 1;

  int offset = Linear::generateOffset( *baseMap );

  RCP<Epetra_CrsGraph> blockGraph = Linear::createBlockGraph( offset, blockPattern, *blockMap, baseFullGraph);

  delete ACMatrix_;
  ACMatrix_ = new Linear::BlockMatrix( numBlocks, offset, blockPattern, *blockGraph, baseFullGraph);

  // First diagonal block
  ACMatrix_->put( 0.0 ); // Zero out whole matrix
  ACMatrix_->block( 0, 0 ).add(*G_);

  // Second diagonal block
  ACMatrix_->block( 1, 1 ).add(*G_);

  B_->putScalar( 0.0 );
  B_->block( 0 ).addVec( 1.0, *bVecRealPtr);
  B_->block( 1 ).addVec( 1.0, *bVecImagPtr);

  Amesos amesosFactory;

  delete X_;
  X_ = new Linear::BlockVector (numBlocks, blockMap, baseMap);
  X_->putScalar( 0.0 );

  delete saved_AC_X_;
  saved_AC_X_ = new Linear::BlockVector (numBlocks, blockMap, baseMap);
  saved_AC_X_->putScalar( 0.0 );

  delete blockProblem_;
  blockProblem_ = new Epetra_LinearProblem(&ACMatrix_->epetraObj(), &X_->epetraObj(), &B_->epetraObj() );

  delete blockSolver_;
  blockSolver_ = amesosFactory.Create( "Klu", *blockProblem_ );

  // Need to reindex the linear system because of the noncontiguous block map.
  Teuchos::ParameterList params;
  params.set( "Reindex", true );
  blockSolver_->SetParameters( params );

#if 0
  // Call symbolic factorization without syncronizing the values, since they are not necessary here. 
  int linearStatus = blockSolver_->SymbolicFactorization();
  if (linearStatus != 0)
  {
    Xyce::dout() << "Amesos symbolic factorization exited with error: " << linearStatus;
    bsuccess = false;
  }
#endif

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::resetACLinearSystem_
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::resetACLinearSystem_()
{
  double omega =  2.0 * M_PI * currentFreq_;

  ACMatrix_->block( 0, 1).put( 0.0);
  ACMatrix_->block( 0, 1).add(*C_);
  ACMatrix_->block( 0, 1).scale(-omega);

  ACMatrix_->block(1, 0).put( 0.0);
  ACMatrix_->block(1, 0).add(*C_);
  ACMatrix_->block(1, 0).scale(omega);

  // Copy the values loaded into the blocks into the global matrix for the solve.
  ACMatrix_->assembleGlobalMatrix();

  ACMatrix_->setUseTranspose (false);
  blockSolver_->SetUseTranspose(false);

  // restore the AC RHS
  B_->putScalar( 0.0 );
  B_->block( 0 ).addVec( 1.0, *bVecRealPtr);
  B_->block( 1 ).addVec( 1.0, *bVecImagPtr);

  // clear out X_
  X_->putScalar( 0.0 );

  return true;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::solveACLinearSystem_()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::solveACLinearSystem_()
{
  bool bsuccess = true;
 // Solve the block problem

  int linearStatus = blockSolver_->SymbolicFactorization();
  if (linearStatus != 0)
  {
    Xyce::dout() << "Amesos symbolic factorization exited with error: " << linearStatus;
    bsuccess = false;
  }
  
  linearStatus = blockSolver_->NumericFactorization();
  if (linearStatus != 0)
  {
    Xyce::dout() << "Amesos numeric factorization exited with error: " << linearStatus;
    bsuccess = false;
  }

  linearStatus = blockSolver_->Solve();
  if (linearStatus != 0)
  {
    Xyce::dout() << "Amesos solve exited with error: " << linearStatus;
    bsuccess = false;
  }

  return bsuccess;
}


//-----------------------------------------------------------------------------
// Function      : NOISE::processOutputNodes
// Purpose       : determines the GIDs for the nodes specified in the first argument 
//                 of the .NOISE line.
// Special Notes : 
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/15/2014
//-----------------------------------------------------------------------------
void NOISE::processOutputNodes ()
{
  // setup the names:
  outputVarNames_.clear();
  outputVarNames_.push_back(outputNode1_);
  if (!outputNodeSingle_)
  {
    outputVarNames_.push_back(outputNode2_);
  }
  int numOutVars = outputVarNames_.size();

  // set up the gid's:
  int found(0);
  int found2(0);
  bool foundLocal(false);
  bool foundLocal2(false);

  N_PDS_Manager &pds_manager = *analysisManager_.getPDSManager();
  N_PDS_Comm & comm = *(pds_manager.getPDSComm());
  int myPID = comm.procID();

  outputVarGIDs_.resize( numOutVars, -1 );
  for (int iout = 0; iout < numOutVars; ++iout)
  {
    std::vector<int> svGIDList1, dummyList;
    char type1;
    foundLocal = topology_.getNodeSVarGIDs(NodeID(outputVarNames_[iout], Xyce::_VNODE), 
        svGIDList1, dummyList, type1);

    found = static_cast<int>(foundLocal);
    Xyce::Parallel::AllReduce(comm.comm(), MPI_LOR, &found, 1);

    foundLocal2 = false;
    if (!found)// if looking for this as a voltage node failed, try a "device" (i.e. current) node.
    {
      foundLocal2 = topology_.getNodeSVarGIDs(NodeID(outputVarNames_[iout], Xyce::_DNODE),
          svGIDList1, dummyList, type1);
    }
    found2 = static_cast<int>(foundLocal2);
    Xyce::Parallel::AllReduce(comm.comm(), MPI_LOR, &found2, 1);

    if (!found && !found2)
    {
      Report::UserError() << "Output function variable " << outputVarNames_[iout] << " not found";
    }

    if (found || found2)
    {
      int tmpGID=-1;
      if(svGIDList1.size()==1)
      {
        tmpGID = svGIDList1.front();
      }
      outputVarGIDs_[iout] = tmpGID;
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : NOISE::solveDirectNOISE_
// Purpose       : Solves for NOISE using the direct method.  
// Special Notes : Not practical, implemented only as a test for adjoint.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/15/2014
//-----------------------------------------------------------------------------
bool NOISE::solveDirectNOISE_()
{
  bool bsuccess = true;
  // Solve the block problem

  // get the number of noise sources
  // int numSources = loader_.getNumNoiseSources();
  // for (int isource=0;isource<numSources;++isources)
  // {
  //   // need different BVector and linear solve set up for each noise source
  //  
  //  // substitute the noise Bvector into the linear system object
  //
  //  // solve linear problem for dX/dNoise
  //
  //  // perform matvec to get this dOutput/dNoise for this noise source
  // }


  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::setupAdjointRHS_
// Purpose       : Sets up stuff that only needs to be set up once, 
//                 but is unique to the adjoint solve.
//                    (i.e. is not used by the direct form)
// Special Notes : 
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/15/2014
//-----------------------------------------------------------------------------
void NOISE::setupAdjointRHS_()
{
  // set up the RHS vector of 1's and 0's vector to represent dOutput/dX.  
  // If thinking of this as the input to the adjoint circuit problem 
  // (or adjoint operator) this can also be thought of as current sources to 
  // the adjoint circuit.
  //
  // For the adjoint problem
  // assume that the phase is zero (as this is essentially the input to the
  // adjoint operator) and just set 1's on the real part of the block vector.
  // For this purpose re-use the B_ vector from the AC problem.

  // replace RHS in linear system.  Assume that the phase is zero, so no 
  // imaginary contribution.
 
  N_PDS_Manager &pds_manager = *analysisManager_.getPDSManager();
  N_PDS_Comm & comm = *(pds_manager.getPDSComm());
  int myPID = comm.procID();


  bNoiseVecRealPtr->putScalar(0.0);
  bNoiseVecImagPtr->putScalar(0.0);

  int numOutVars = outputVarNames_.size();
  for (int iout=0;iout<numOutVars;++iout)
  {
    int tmpGID=outputVarGIDs_[iout];
    if (tmpGID > -1)
    {
      double val=1.0;
      if (iout>0) val=-1.0;
      bNoiseVecRealPtr->setElementByGlobalIndex( tmpGID, val, 0);
    }
  }
  bNoiseVecRealPtr->fillComplete();
}

//-----------------------------------------------------------------------------
// Function      : NOISE::resetAdjointNOISELinearSystem_
// Purpose       : 
// Special Notes : 
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/15/2014
//-----------------------------------------------------------------------------
void NOISE::resetAdjointNOISELinearSystem_()
{
  // clear out the X_ vector used by the linear solver.
  X_->putScalar(0.0);


  // setup the B_ vector RHS for the adjoint solve:
  B_->putScalar( 0.0 );
  B_->block( 0 ).addVec( 1.0, *bNoiseVecRealPtr);
  B_->block( 1 ).addVec( 1.0, *bNoiseVecImagPtr);

#if 0
  //-------------------------------------------
  Xyce::dout()<<"adjoint noise B vector:"<<std::endl;
  B_->printPetraObject(Xyce::dout());
  //-------------------------------------------
#endif

  // solve transpose (adjoint) system to obtain dOutput/dF.  If thinking of this
  // as the adjoint circuit problem (or adjoint operator) this can also be
  // thought of as the adjoint voltages, or more precisely a transImpedance (gain).

  ACMatrix_->setUseTranspose (true);
  blockSolver_->SetUseTranspose(true);

#if 0
  //-------------------------------------------
  Xyce::dout()<<"matrix:"<<std::endl;
  ACMatrix_->printPetraObject(Xyce::dout());
  //-------------------------------------------
#endif
}

//-----------------------------------------------------------------------------
// Function      : NOISE::solveAdjointNOISE_
// Purpose       : Solves for NOISE using the adjoint method.  
// Special Notes : This is the production solve.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/15/2014
//-----------------------------------------------------------------------------
bool NOISE::solveAdjointNOISE_()
{
  bool bsuccess = true;

  // do the solve of the tranpose system to get dOutput/dF
  int linearStatus = blockSolver_->SymbolicFactorization();
  if (linearStatus != 0)
  {
    Xyce::dout() << "Amesos symbolic factorization exited with error: " << linearStatus;
    bsuccess = false;
  }

  linearStatus = blockSolver_->NumericFactorization();
  if (linearStatus != 0)
  {
    Xyce::dout() << "Amesos numeric factorization exited with error: " << linearStatus;
    bsuccess = false;
  }

  linearStatus = blockSolver_->Solve();
  if (linearStatus != 0)
  {
    Xyce::dout() << "Amesos solve exited with error: " << linearStatus;
    bsuccess = false;
  }

  double omega =  2.0 * M_PI * currentFreq_;

  int numNoiseDevices = noiseDataVec_.size();
  for (int i=0;i<numNoiseDevices;++i)
  {
    (noiseDataVec_[i])->omega = omega;
    (noiseDataVec_[i])->freq = currentFreq_;
  }

  loader_.getNoiseSources(noiseDataVec_);

  std::vector< Teuchos::RCP<Linear::Vector> > outputVectors;
  outputVectors.push_back( rcp(linearSystem_.builder().createVector()) );
  outputVectors.push_back( rcp(linearSystem_.builder().createVector()) );

  copyFromBlockVector( *X_, outputVectors );
  Linear::Vector & Xreal = *(outputVectors[0]);
  Linear::Vector & Ximag = *(outputVectors[1]);

#if 0
  std::cout << "Xreal: ------------------------------------"<<std::endl;
  Xreal.printPetraObject( Xyce::dout() );
  std::cout << "Ximag: ------------------------------------"<<std::endl;
  Ximag.printPetraObject( Xyce::dout() );
#endif

  N_PDS_Manager &pds_manager = *analysisManager_.getPDSManager();
  N_PDS_Comm & comm = *(pds_manager.getPDSComm());
  int myPID = comm.procID();

  totalOutputNoiseDens_ = 0.0;
  totalInputNoiseDens_ = 0.0;
  for (int i=0;i<numNoiseDevices;++i)
  {
    int numNoiseThisDevice = noiseDataVec_[i]->numSources;

    noiseDataVec_[i]->totalNoise = 0.0;
    noiseDataVec_[i]->totalOutputNoise = 0.0;
    for (int j=0;j<numNoiseThisDevice;++j)
    {
      int li_Pos = noiseDataVec_[i]->li_Pos[j];
      int li_Neg = noiseDataVec_[i]->li_Neg[j];

      double realVal = Xreal[li_Pos] - Xreal[li_Neg];
      double imagVal = Ximag[li_Pos] - Ximag[li_Neg];
      double gain = (realVal*realVal) + (imagVal*imagVal);

      noiseDataVec_[i]->totalNoise += noiseDataVec_[i]->noiseDens[j];

      noiseDataVec_[i]->outputNoiseDens[j] = gain * noiseDataVec_[i]->noiseDens[j];
#if 0
      std::string namestring = noiseDataVec_[i]->noiseNames[j];
      std::cout << std::left << "noise: " << std::setw(20)<< namestring <<std::right << ":";
      std::cout.setf(std::ios::scientific);
      std::cout << " li_Pos="<<li_Pos<<" li_Neg="<<li_Neg
                << " gain="<<gain<< " noiseDens="<<noiseDataVec_[i]->noiseDens[j]
                << " onoise="<<noiseDataVec_[i]->outputNoiseDens[j] <<std::endl;
#endif

      noiseDataVec_[i]->lnNoiseDens[j] = std::log(std::max( noiseDataVec_[i]->outputNoiseDens[j],N_MINLOG) ); 

      noiseDataVec_[i]->inputNoiseDens[j] = noiseDataVec_[i]->outputNoiseDens[j] * GainSqInv_;

      noiseDataVec_[i]->totalOutputNoise += noiseDataVec_[i]->outputNoiseDens[j];
    }
#if 0
    std::cout << std::endl;
#endif
    noiseDataVec_[i]->totalInputNoise = noiseDataVec_[i]->totalOutputNoise * GainSqInv_;

    totalOutputNoiseDens_ += noiseDataVec_[i]->totalOutputNoise;
  }
  Xyce::Parallel::AllReduce(comm.comm(), MPI_SUM, &totalOutputNoiseDens_, 1);
  totalInputNoiseDens_ += totalOutputNoiseDens_ * GainSqInv_;

  if (comm.isSerial() )
  {
    // FIX:  replace this output call!
    hackTecplotOutput();
  }

  return bsuccess;
}


//-----------------------------------------------------------------------------
// Function      : NOISE::hackTecplotOutput
// Purpose       : provides an output function until I setup the output 
//                 manager properly for NOISE.
// Author        : Eric Keiter
// Date          : 1/5/2015
//-----------------------------------------------------------------------------
void NOISE::hackTecplotOutput()
{
  int numNoiseDevices = noiseDataVec_.size();
  std::string fileName = analysisManager_.getNetlistFilename() + "_noise.dat";

  std::ofstream output_stream(fileName.c_str(), 
            !hackOutputCalledBefore_ ? std::ios_base::out : std::ios_base::app);

  if (!hackOutputCalledBefore_)
  {
    // header output
    output_stream << "TITLE=\"noise output\"\tVARIABLES=\"frequency\" "<<std::endl;
    output_stream << "\t\"Re(V(" << outputNode1_ << "))\""<<std::endl;
    output_stream << "\t\"Im(V(" << outputNode1_ << "))\""<<std::endl;

    if(!outputNodeSingle_)
    {
      output_stream << "\t\"Re(V(" << outputNode2_ << "))\""<<std::endl;
      output_stream << "\t\"Im(V(" << outputNode2_ << "))\""<<std::endl;
    }
    output_stream << "\t\"onoise_spectrum \""<<std::endl;
    output_stream << "\t\"inoise_spectrum \""<<std::endl;

    if(pts_per_summary_>0)
    {
      for (int i=0;i<numNoiseDevices;++i)
      {
        int numNoiseThisDevice = noiseDataVec_[i]->numSources;
        for (int j=0;j<numNoiseThisDevice;++j)
        {
          std::string namestring = noiseDataVec_[i]->noiseNames[j];
          std::transform(namestring.begin(), namestring.end(), namestring.begin(), ::tolower);
          output_stream << "\t\"" << namestring << "\"";
        }

        // spice doesn't output input noise for individual devices, so exclude.
        if (numNoiseThisDevice > 1)
        {
          std::string totalOutputNoiseName = "onoise_" + noiseDataVec_[i]->deviceName;

          std::transform(totalOutputNoiseName.begin(), totalOutputNoiseName.end(), totalOutputNoiseName.begin(), ::tolower);

          output_stream << "\t\""<<totalOutputNoiseName<<"\"";
        }
      }
    }
 
    output_stream << std::endl;
    output_stream << "ZONE F=POINT  T=\"Xyce data\""<<std::endl;
    hackOutputCalledBefore_ = true;
  }

  // data output
  output_stream.setf(std::ios::scientific);

  // output the AC solution for the output node(s) (for double checking)
  Linear::Vector & Xreal = saved_AC_X_->block( 0 );
  Linear::Vector & Ximag = saved_AC_X_->block( 1 );
  double v1r = Xreal.getElementByGlobalIndex(outputVarGIDs_[0]);
  double v1i = Ximag.getElementByGlobalIndex(outputVarGIDs_[0]);

  output_stream << currentFreq_ << "\t" << v1r << "\t" << v1i;

  if(!outputNodeSingle_)
  {
    double v2r = 0.0;
    double v2i = 0.0;
    v2r = Xreal.getElementByGlobalIndex(outputVarGIDs_[1]);
    v2i = Ximag.getElementByGlobalIndex(outputVarGIDs_[1]);
    output_stream << "\t" << v2r << "\t" << v2i;
  }

  output_stream << "\t" << totalOutputNoiseDens_ << "\t" << totalInputNoiseDens_;

  if(pts_per_summary_>0)
  {
    for (int i=0;i<numNoiseDevices;++i)
    {
      int numNoiseThisDevice = noiseDataVec_[i]->numSources;
      for (int j=0;j<numNoiseThisDevice;++j)
      {
        output_stream << "\t" << noiseDataVec_[i]->outputNoiseDens[j];
      }

      if (numNoiseThisDevice > 1)
      {
        output_stream << "\t" << noiseDataVec_[i]->totalOutputNoise;
      }
      //output_stream << std::endl;
    }
  }
  output_stream << std::endl;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::noiseIntegral
// Purpose       : based on the spice3 function: Nintegrate.
//  This subroutine evaluates the integral of the function
//
//                                             EXPONENT
//                      NOISE = a * (FREQUENCY)
//
//   given two points from the curve.  If EXPONENT is relatively close
//   to 0, the noise is simply multiplied by the change in frequency.
//   If it isn't, a more complicated expression must be used.  Note that
//   EXPONENT = -1 gives a different equation than EXPONENT <> -1.
//   Hence, the reason for the constant 'N_INTUSELOG'.
//
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
double NOISE::noiseIntegral(double noizDens, double lnNdens, double lnNlstDens, 
      double delLnFreq, double delFreq, double lnFreq, double lnLastFreq)
{
  double exponent;
  double a;

  exponent = (lnNdens - lnNlstDens) / delLnFreq;
  if ( fabs(exponent) < N_INTFTHRESH ) 
  {
    return (noizDens * delFreq);
  } else 
  {
    a = std::exp(lnNdens - exponent*lnFreq);
    exponent += 1.0;
    if (fabs(exponent) < N_INTUSELOG) 
    {
      return (a * (lnFreq - lnLastFreq));
    } 
    else 
    {
      return (a * ((std::exp(exponent * lnFreq) - std::exp(exponent * lnLastFreq)) /
            exponent));
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : NOISE::processSuccessfulStep()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::doProcessSuccessfulStep()
{
  bool bsuccess = true;

  // FIX OUTPUT
  outputManagerAdapter_.outputAC (currentFreq_, saved_AC_X_->block(0), saved_AC_X_-> block(1));

  outputManagerAdapter_.outputNoise 
    (currentFreq_, totalOutputNoiseDens_, totalInputNoiseDens_, noiseDataVec_);

  if ( !firstDoubleDCOPStep() )
  {
      stepNumber += 1;
      stats_.successStepsThisParameter_ += 1;
      stats_.successfulStepsTaken_ += 1;
  }

  // This output call is for device-specific output (like from a PDE device,
  // outputting mesh-based tecplot files).  It will only work in parallel if on
  // a machine where all processors have I/O capability.
  loader_.outputPlotFiles();

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::processFailedStep
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::doProcessFailedStep()
{
  bool bsuccess = true;

  stepNumber += 1;
  noiseSweepFailures_.push_back(stepNumber);
  stats_.failedStepsAttempted_  += 1;
  analysisManager_.getStepErrorControl().numberSuccessiveFailures += 1;

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::finish
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::doFinish()
{
  bool bsuccess = true;

  if (DEBUG_ANALYSIS)
  {
    Xyce::dout() << "Calling NOISE::doFinish() outputs!" << std::endl;
  }

  // FIX OUTPUT
  outputManagerAdapter_.finishOutput();

  if (!(noiseSweepFailures_.empty()))
  {
    bsuccess = false;
  }

  return bsuccess;
}


//-----------------------------------------------------------------------------
// Function      : NOISE::doHandlePredictor
// Purpose       :
// Special Notes :
// Scope         : private
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::doHandlePredictor()
{
  analysisManager_.getDataStore()->setErrorWtVector(tiaParams_);
  analysisManager_.getWorkingIntegrationMethod().obtainPredictor();
  analysisManager_.getWorkingIntegrationMethod().obtainPredictorDeriv();

  // In case this is the upper level of a 2-level sim, tell the
  // inner solve to do its prediction:
  bool        beginIntegrationFlag = analysisManager_.getBeginningIntegrationFlag();           // system_state.beginIntegrationFlag;
  double      nextTimeStep = analysisManager_.getStepErrorControl().currentTimeStep;           // system_state.nextTimeStep;
  double      nextTime = analysisManager_.getStepErrorControl().nextTime;                      // system_state.nextTime;
  int         currentOrder = analysisManager_.getWorkingIntegrationMethod().getOrder();        // system_state.currentOrder;

  loader_.startTimeStep(beginIntegrationFlag, nextTimeStep, nextTime, currentOrder);

  return true;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::updateCurrentFreq_(int stepNumber )
// Purpose       :
// Special Notes : Used for NOISE analysis classes.
// Scope         : public
// Creator       : Eric Keiter, SNL.
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool NOISE::updateCurrentFreq_(int stepNumber)
{
  lastFreq_ = currentFreq_;
  if (type_ == "LIN")
  {

    currentFreq_  = fStart_ + static_cast<double>(stepNumber)*fstep_;
  }
  else if(type_ == "DEC" || type_ == "OCT")
  {

    currentFreq_ = fStart_*pow(stepMult_, static_cast<double>(stepNumber) );
  }
  else
  {
    Report::DevelFatal().in("NOISE::updateCurrentFreq_") 
      << "NOISE::updateCurrentFreq_: unsupported STEP type";
  }

  delFreq_ = currentFreq_ - lastFreq_;

  lnFreq_     = std::log(std::max(currentFreq_,N_MINLOG));
  lnLastFreq_ = std::log(std::max(lastFreq_,N_MINLOG));
  delLnFreq_  = lnFreq_ - lnLastFreq_;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : NOISE::setupSweepParam_
// Purpose       : Processes sweep parameters.
// Special Notes : Used for NOISE analysis classes.
// Scope         : public
// Creator       : Eric Keiter, SNL.
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
int NOISE::setupSweepParam_()
{
  double fstart, fstop;
  double fcount = 0.0;

  fstart = fStart_;
  fstop = fStop_;

  if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
  {
    Xyce::dout() << std::endl << std::endl;
    Xyce::dout() << section_divider << std::endl;
    Xyce::dout() << "NOISE::setupSweepParam_" << std::endl;
  }

    if (type_ == "LIN")
    {
      int np = static_cast<int>(np_);

      if ( np == 1)
        fstep_ = 0;
      else
        fstep_  = (fstop - fstart)/(np_ - 1.0);

      fcount = np_;
      if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
      {
        Xyce::dout() << "fstep   = " << fstep_  << std::endl;
      }
    }
    else if (type_ == "DEC")
    {
      stepMult_ = std::pow(10.0, 1.0/np_);
      fcount   = floor(fabs(std::log10(fstart) - std::log10(fstop)) * np_ + 1.0);
      if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
      {
        Xyce::dout() << "stepMult_ = " << stepMult_  << std::endl;
      }
    }
    else if (type_ == "OCT")
    {
      stepMult_ = std::pow(2.0, 1.0/np_);

      // changed to remove dependence on "log2" function, which apparently
      // doesn't exist in the math libraries of FreeBSD or the mingw
      // cross-compilation suite.   Log_2(x)=log_e(x)/log_e(2.0)
      double ln2 = std::log(2.0);
      fcount   = floor(fabs(std::log(fstart) - std::log(fstop))/ln2 * np_ + 1.0);
      if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
      {
        Xyce::dout() << "stepMult_ = " << stepMult_  << std::endl;
      }
    }
    else
    {
      Report::DevelFatal().in("NOISE::setupSweepParam") << "Unsupported type";
    }

  // At this point, pinterval equals the total number of steps
  // for the step loop.
  return static_cast<int> (fcount);
}

namespace {

typedef Util::Factory<AnalysisBase, NOISE>  NOISEFactoryBase;

//-----------------------------------------------------------------------------
// Class         : NOISEFactory
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Thu Jan 29 12:53:02 2015
//-----------------------------------------------------------------------------
///
/// Factory for parsing NOISE parameters from the netlist and creating NOISE analysis.
///
class NOISEFactory : public NOISEFactoryBase
{
public:
  //-----------------------------------------------------------------------------
  // Function      : NOISEFactory
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:54:09 2015
  //-----------------------------------------------------------------------------
  ///
  /// Constructs the NOISE analysis factory
  ///
  /// @invariant Stores the results of parsing, so if more than one of the analysis and
  /// associated lines are parsed, the second options simply overwrite the previously parsed
  /// values.
  ///
  /// @invariant The existence of the parameters specified in the constructor cannot
  /// change.
  ///
  /// @param analysis_manager 
  /// @param linear_system 
  /// @param nonlinear_manager 
  /// @param topology 
  ///
  NOISEFactory(
    Analysis::AnalysisManager & analysis_manager,
    Linear::System &            linear_system,
    Nonlinear::Manager &        nonlinear_manager,
    Loader::Loader &            loader,
    Topo::Topology &            topology,
    IO::InitialConditionsManager &      initial_conditions_manager)
    : NOISEFactoryBase(),
      analysisManager_(analysis_manager),
      linearSystem_(linear_system),
      nonlinearManager_(nonlinear_manager),
      loader_(loader),
      topology_(topology),
      initialConditionsManager_(initial_conditions_manager)
  {}

  virtual ~NOISEFactory()
  {}

  //-----------------------------------------------------------------------------
  // Function      : create
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:59:00 2015
  //-----------------------------------------------------------------------------
  ///
  /// Create a new NOISE analysis and applies the analysis and time integrator option blocks.
  ///
  /// @return new NOISE analysis object
  ///
  NOISE *create() const
  {
    analysisManager_.setAnalysisMode(ANP_MODE_NOISE);
    NOISE *noise = new NOISE(analysisManager_, linearSystem_, nonlinearManager_, loader_, topology_, initialConditionsManager_);
    noise->setAnalysisParams(noiseAnalysisOptionBlock_);
    noise->setTimeIntegratorOptions(timeIntegratorOptionBlock_);

    return noise;
  }

  //-----------------------------------------------------------------------------
  // Function      : setNOISEAnalysisOptionBlock
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 13:00:14 2015
  //-----------------------------------------------------------------------------
  ///
  /// Saves the analysis parsed options block in the factory.
  ///
  /// @invariant Overwrites any previously specified analysis option block.
  ///
  /// @param option_block parsed option block
  ///
  void setNOISEAnalysisOptionBlock(const Util::OptionBlock &option_block)
  {
    noiseAnalysisOptionBlock_ = option_block;
  }

  //-----------------------------------------------------------------------------
  // Function      : setTimeIntegratorOptionBlock
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 13:01:27 2015
  //-----------------------------------------------------------------------------
  ///
  /// Saves the time integrator parsed option block.
  ///
  /// @invariant Overwrites any previously specified time integrator option block.
  ///
  /// @param option_block parsed option block
  ///
  bool setTimeIntegratorOptionBlock(const Util::OptionBlock &option_block)
  {
    timeIntegratorOptionBlock_ = option_block;

    return true;
  }

public:
  AnalysisManager &                     analysisManager_;
  Linear::System &                      linearSystem_;
  Nonlinear::Manager &                  nonlinearManager_;
  Loader::Loader &                      loader_;
  Topo::Topology &                      topology_;
  IO::InitialConditionsManager &        initialConditionsManager_;

private:
  Util::OptionBlock     noiseAnalysisOptionBlock_;
  Util::OptionBlock     timeIntegratorOptionBlock_;
};

// .NOISE
struct NOISEAnalysisReg : public IO::PkgOptionsReg
{
  NOISEAnalysisReg(
    NOISEFactory &   factory )
    : factory_(factory)
  {}

  bool operator()(const Util::OptionBlock &option_block)
  {
    factory_.setNOISEAnalysisOptionBlock(option_block);

    factory_.analysisManager_.addAnalysis(&factory_);

    return true;
  }

  NOISEFactory &              factory_;
};

//-----------------------------------------------------------------------------
// Function      : extractNOISEData
// Purpose       : Extract the parameters from a netlist .NOISE line held in
//                 parsed_line.
// Special Notes : noise specification:
//
//  .noise v(output <,ref>) src ( dec | lin | oct ) pts fstart fstop + <pts_per_summary>
//
// for now ignoring the optional points per summary
//
// Scope         : public
// Creator       : Eric R. Keiter
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
bool
extractNOISEData(
  IO::PkgOptionsMgr &           options_manager,
  IO::CircuitBlock &            circuit_block,
  const std::string &           netlist_filename,
  const IO::TokenVector &       parsed_line)
{
  Util::OptionBlock option_block("NOISE", Util::OptionBlock::NO_EXPRESSIONS, netlist_filename, parsed_line[0].lineNumber_);

  int numFields = parsed_line.size();

  // Check that the minimum required number of fields are on the line.
  if (numFields < 10)
  {
    Report::UserError0().at(netlist_filename, parsed_line[0].lineNumber_)
      << ".NOISE line has an unexpected number of fields.  NumFields = " << numFields;
    return false;
  }

  int linePosition = 1;   // Start of parameters on .param line.

  Util::Param parameter("", "");

  // output node(s):
  ExtendedString stringVal("");
  std::ostringstream msg;
  int p_err=0;
  if( parsed_line[linePosition].string_ == "V" || parsed_line[linePosition].string_ == "v")
  {
    if( parsed_line[linePosition+3].string_ == ")" )
    {
      stringVal = parsed_line[linePosition].string_;
      stringVal.toUpper();
      parameter.setTag(stringVal);
      parameter.setVal( 1.0 );
      option_block.addParam( parameter );

      stringVal = parsed_line[linePosition+2].string_;
      stringVal.toUpper();
      parameter.setTag( stringVal );
      parameter.setVal( 0.0 );
      option_block.addParam( parameter );

      linePosition += 4;
    }
    else if( parsed_line[linePosition+5].string_ == ")" )
    {
      stringVal = parsed_line[linePosition].string_;
      stringVal.toUpper();
      parameter.setTag(stringVal);
      parameter.setVal( 2.0 );
      option_block.addParam( parameter );

      stringVal = parsed_line[linePosition+2].string_;
      stringVal.toUpper();
      parameter.setTag( stringVal );
      parameter.setVal( 0.0 );
      option_block.addParam( parameter );

      stringVal = parsed_line[linePosition+4].string_;
      stringVal.toUpper();
      parameter.setTag( stringVal );
      parameter.setVal( 0.0 );
      option_block.addParam( parameter );

      linePosition += 6;
    }
    else
    {
      msg << "Unrecognized parenthetical specification for NOISE output ";
      p_err = linePosition;
    }
  }
  else
  {
    msg << "Incorrect format for NOISE output.";
    p_err = linePosition;
  }

  // source is required
  stringVal = parsed_line[linePosition].string_;
  stringVal.toUpper();
  parameter.setTag( "SOURCE" );
  parameter.setVal(std::string(stringVal));
  option_block.addParam( parameter );
#if 0
  std::cout << "Source is at linePosition="<<linePosition<<"  source="<<stringVal<<std::endl;
#endif
  ++linePosition;     // Advance to next parameter.

  // type is required
  stringVal = parsed_line[linePosition].string_;
  stringVal.toUpper();
  parameter.setTag( "TYPE" );
  parameter.setVal(std::string(stringVal));
  option_block.addParam( parameter );
#if 0
  std::cout << "Type is at linePosition="<<linePosition<<"  type="<<stringVal<<std::endl;
#endif
  ++linePosition;     // Advance to next parameter.

  // np is required
  parameter.setTag( "NP" );
  parameter.setVal( parsed_line[linePosition].string_ );
  option_block.addParam( parameter );
#if 0
  std::cout << "NP is at linePosition="<<linePosition<<"  np="
    << parsed_line[linePosition].string_  <<std::endl;
#endif
  ++linePosition;     // Advance to next parameter.

  // fstart is required
  parameter.setTag( "FSTART" );
  parameter.setVal( parsed_line[linePosition].string_ );
  option_block.addParam( parameter );
#if 0
  std::cout << "fstart is at linePosition="<<linePosition<<"  fstart="
    << parsed_line[linePosition].string_  <<std::endl;
#endif
  ++linePosition;     // Advance to next parameter.


  // fstop is required
  parameter.setTag( "FSTOP" );
  parameter.setVal( parsed_line[linePosition].string_ );
  option_block.addParam( parameter );
#if 0
  std::cout << "fstop is at linePosition="<<linePosition<<"  fstop="
    << parsed_line[linePosition].string_  <<std::endl;
#endif

  // pts_per_summary is optional.  If value is negative, assume it wasn't set.
  parameter.setTag( "PTS_PER_SUMMARY" );
  if (numFields >= 11)
  {
    ++linePosition;     // Advance to next parameter.

    parameter.setVal( parsed_line[linePosition].string_ );

#if 0
    std::cout << "pts_per_summary is at linePosition="<<linePosition<<"  pts_per_summary="
      << parsed_line[linePosition].string_  <<std::endl;
#endif
  }
  else
  {
    parameter.setVal( std::string("-1") );
#if 0
    std::cout << "pts_per_summary was not specified.  It is set to:  pts_per_summary=-1" <<std::endl;
#endif
  }
  option_block.addParam( parameter );

  circuit_block.addOptions(option_block);

  return true;
}

} // namespace <unnamed>

//-----------------------------------------------------------------------------
// Function      : NOISEFactory::registerFactory
// Purpose       :
//-----------------------------------------------------------------------------
bool
registerNOISEFactory(
  FactoryBlock &        factory_block)
{
  NOISEFactory *factory = new NOISEFactory(factory_block.analysisManager_, factory_block.linearSystem_, factory_block.nonlinearManager_, factory_block.loader_, factory_block.topology_, factory_block.initialConditionsManager_);

  addAnalysisFactory(factory_block, factory);

  factory_block.optionsManager_.addCommandParser(".NOISE", extractNOISEData);

  factory_block.optionsManager_.addCommandProcessor("NOISE", new NOISEAnalysisReg(*factory));

  factory_block.optionsManager_.addOptionsProcessor("TIMEINT", IO::createRegistrationOptions<NOISEFactory>(*factory, &NOISEFactory::setTimeIntegratorOptionBlock));

  return true;
}

} // namespace Analysis
} // namespace Xyce
