//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_IO_MeasureManager.C,v $
// Purpose       : This file contains the functions to manage measure objects
// Special Notes :
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//
// Revision Information:
// ---------------------
// Revision Number: $Revision: 1.71 $
// Revision Date  : $Date: 2015/10/27 14:56:28 $
// Current Owner  : $Author: peshola $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <utility>
#include <sstream>

#include <Teuchos_oblackholestream.hpp>

#include <N_ERH_ErrorMgr.h>
#include <N_IO_CircuitBlock.h>
#include <N_IO_MeasureAverage.h>
#include <N_IO_MeasureDerivativeEvaluation.h>
#include <N_IO_MeasureDuty.h>
#include <N_IO_MeasureEquationEvaluation.h>
#include <N_IO_MeasureFindWhen.h>
#include <N_IO_MeasureFourier.h>
#include <N_IO_MeasureFrequency.h>
#include <N_IO_MeasureIntegralEvaluation.h>
#include <N_IO_MeasureManager.h>
#include <N_IO_MeasureMax.h>
#include <N_IO_MeasureMin.h>
#include <N_IO_MeasureOffTime.h>
#include <N_IO_MeasureOnTime.h>
#include <N_IO_MeasurePeakToPeak.h>
#include <N_IO_MeasureRMS.h>
#include <N_IO_MeasureRelativeError.h>
#include <N_IO_MeasureRiseFallDelay.h>
#include <N_IO_OpBuilders.h>
#include <N_IO_OptionBlock.h>
#include <N_IO_OutputPrn.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_SpiceSeparatedFieldTool.h>
#include <N_PDS_ParMap.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_NoCase.h>
#include <N_UTL_OptionBlock.h>

namespace Xyce {
namespace IO {
namespace Measure {

//-----------------------------------------------------------------------------
// Function      : Manager::Manager
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
Manager::Manager(
  const std::string &   netlist_filename)
  : netlistFilename_(netlist_filename)
{}

//-----------------------------------------------------------------------------
// Function      : Manager::Manager
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
Manager::~Manager()
{
  for (MeasurementVector::iterator it = allMeasuresList_.begin(); it != allMeasuresList_.end(); ++it)
    delete (*it);
}

void
Manager::makeMeasureOps(Parallel::Machine comm, const Util::Op::BuilderManager &op_builder_manager) 
{
  for (MeasurementVector::iterator it = allMeasuresList_.begin(); it != allMeasuresList_.end(); ++it)
    (*it)->makeMeasureOps(comm, op_builder_manager);
}

void
Manager::notify(
  const Analysis::StepEvent &   step_event)
{
  switch (step_event.state_) {
    case Analysis::StepEvent::INITIALIZE:
      break;

    case Analysis::StepEvent::STEP_STARTED:
      for (MeasurementVector::iterator it = allMeasuresList_.begin(); it != allMeasuresList_.end(); ++it) 
      {
        (*it)->reset();
      }
      // move any inactive measures back to the active list
      activeMeasuresList_ = allMeasuresList_;
      break;

    case Analysis::StepEvent::STEP_COMPLETED:
      {
        if( !allMeasuresList_.empty() )
        {
          std::ostringstream converterBuff;
          converterBuff << step_event.count_;
          std::string filename = netlistFilename_ + ".mt" + converterBuff.str();
          std::ofstream outputFileStream;
          outputFileStream.open( filename.c_str() );
          outputResults( outputFileStream );
          outputFileStream.close();
          outputVerboseResults(Xyce::lout(), step_event.finalSimTime_  );
        }
      }

      break;

    case Analysis::StepEvent::FINISH:
      break;
  }
}


//-----------------------------------------------------------------------------
// Function      : Manager::addMeasure
// Purpose       : Entry point when .measure lines are passed in the netlist
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
bool Manager::addMeasure(const Util::OptionBlock & measureBlock )
{
  // based on what's in the option block passed in, we
  // create the needed measure instance
  if (DEBUG_IO)
    Xyce::dout() << "In Manager::addMeasure" << std::endl
                 << "measureLine passed it was: " << std::endl << measureBlock << std::endl;

  // here's the base data we should pull from the option block while
  std::string type;

  Util::ParamList::const_iterator it = std::find_if(measureBlock.begin(), measureBlock.end(), Util::EqualParam("TYPE"));
  if (it != measureBlock.end())
  {
    type = (*it).stringValue();
  }
  else
  {
    // this shouldn't happen, but catch it if does
    Report::UserError0() << "Missing TYPE in Manager";
  }

  Base  * theMeasureObject = 0;
  // have enough info to make the correct measure class
  if( type=="TRIG" || type=="TARG" )
  {
    theMeasureObject = new Measure::RiseFallDelay( measureBlock);
  }
  else if( type=="AVG" )
  {
    theMeasureObject = new Measure::Average( measureBlock);
  }
  else if( type=="MAX" )
  {
    theMeasureObject = new Measure::Max( measureBlock);
  }
  else if( type=="MIN" )
  {
    theMeasureObject = new Measure::Min( measureBlock);
  }
  else if( type=="PP" )
  {
    theMeasureObject = new Measure::PeakToPeak( measureBlock);
  }
  else if( type=="RMS" )
  {
    theMeasureObject = new Measure::RMS( measureBlock);
  }
  else if( type=="FREQ" )
  {
    theMeasureObject = new Measure::Frequency( measureBlock);
  }
  else if( type=="DUTY" )
  {
    theMeasureObject = new Measure::Duty( measureBlock);
  }
  else if( type=="ON_TIME" )
  {
    theMeasureObject = new Measure::OnTime( measureBlock);
  }
  else if( type=="OFF_TIME" )
  {
    theMeasureObject = new Measure::OffTime( measureBlock);
  }
  else if( type=="FIND" || type=="WHEN" )
  {
    theMeasureObject = new Measure::FindWhen( measureBlock);
  }
  else if( type=="PARAM" || type=="EQN"  )
  {
    theMeasureObject = new Measure::EquationEvaluation( measureBlock);
  }
  else if( type=="DERIVATIVE" || type=="DERIV" )
  {
    theMeasureObject = new Measure::DerivativeEvaluation( measureBlock);
  }
  else if( type=="INTEGRAL" || type=="INTEG" )
  {
    theMeasureObject = new Measure::IntegralEvaluation( measureBlock);
  }
  else if( type=="ERROR" )
  {
    theMeasureObject = new Measure::RelativeError( measureBlock);
  }
  else if( type=="FOUR" )
  {
    theMeasureObject = new Measure::Fourier( measureBlock);
  }
  else
  {
    // unknown type issue warning.
    Xyce::Report::UserWarning() << "Unknown MEASURE type requested, \"" << type << "\".  This measure will be ignored";
  }

  // if the measure object is supported, then add it to the active and all lists
  if (theMeasureObject && theMeasureObject->typeSupported_ )
  {
    allMeasuresList_.push_back( theMeasureObject );
    activeMeasuresList_.push_back( theMeasureObject );
  }
  else
    delete theMeasureObject;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::updateTranMeasures
// Purpose       : Called during the simulation to update the measure objects
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
void Manager::updateTranMeasures(Parallel::Machine comm, const double circuitTime, const Linear::Vector *solnVec, const Linear::Vector *stateVec, const Linear::Vector *storeVec, const Linear::Vector *lead_current_vector, const Linear::Vector *junction_voltage_vector, const Linear::Vector *lead_current_dqdt_vector)
{
  // loop over active masure objects and get them to update themselves.
  for (MeasurementVector::iterator it = activeMeasuresList_.begin(); it != activeMeasuresList_.end(); ++it) 
  {
    (*it)->updateTran(comm, circuitTime, solnVec, stateVec, storeVec, lead_current_vector, junction_voltage_vector, lead_current_dqdt_vector);
  }
  activeMeasuresList_.erase(std::remove_if(activeMeasuresList_.begin(), activeMeasuresList_.end(), std::mem_fun(&Measure::Base::finishedCalculation)),
                            activeMeasuresList_.end());
}


//-----------------------------------------------------------------------------
// Function      : Manager::updateDcMeasures
// Purpose       : Called during the simulation to update the measure objects
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
void Manager::updateDCMeasures(Parallel::Machine comm, const std::vector<Analysis::SweepParam> & dcParamsVec, const Linear::Vector *solnVec, const Linear::Vector *stateVec, const Linear::Vector *storeVec, const Linear::Vector *lead_current_vector, const Linear::Vector *junction_voltage_vector, const Linear::Vector *lead_current_dqdt_vector)
{
  for (MeasurementVector::iterator it = activeMeasuresList_.begin(); it != activeMeasuresList_.end(); ++it) 
  {
    (*it)->updateDC(comm, dcParamsVec, solnVec, stateVec, storeVec, lead_current_vector, junction_voltage_vector, lead_current_dqdt_vector);
  }
  activeMeasuresList_.erase(std::remove_if(activeMeasuresList_.begin(), activeMeasuresList_.end(), std::mem_fun(&Measure::Base::finishedCalculation)),
                            activeMeasuresList_.end());
}

//-----------------------------------------------------------------------------
// Function      : Manager::updateAcMeasures
// Purpose       : Called during the simulation to update the measure objects
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/21/2014
//-----------------------------------------------------------------------------
void Manager::updateACMeasures(Parallel::Machine comm, const double frequency, const Linear::Vector *real_solution_vector, const Linear::Vector *imaginary_solution_vector)
{
  for (MeasurementVector::iterator it = activeMeasuresList_.begin(); it != activeMeasuresList_.end(); ++it) 
  {
    (*it)->updateAC(comm, frequency, real_solution_vector, imaginary_solution_vector);
  }
  activeMeasuresList_.erase(std::remove_if(activeMeasuresList_.begin(), activeMeasuresList_.end(), std::mem_fun(&Measure::Base::finishedCalculation)),
                            activeMeasuresList_.end()); }

//-----------------------------------------------------------------------------
// Function      : Manager::outputResults
// Purpose       : Output measure results, to .mt files, at end of simulation
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
std::ostream &Manager::outputResults( std::ostream& outputStream ) const
{
  if (!allMeasuresList_.empty())
  {
    // loop over active measure objects and get the results
    for (MeasurementVector::const_iterator it = allMeasuresList_.begin(), end = allMeasuresList_.end(); it != end; ++it)
    {
      (*it)->printMeasureResult( outputStream );
    }
  }
  
  return outputStream;
}

//-----------------------------------------------------------------------------
// Function      : Manager::outputVerboseResults
// Purpose       : Output measure results, to std output, at end of simulation
// Special Notes :
// Scope         : public
// Creator       : Pete Sholander, SNL, Electrical and Microsystem Modeling
// Creation Date : 02/05/2015
//-----------------------------------------------------------------------------
std::ostream &Manager::outputVerboseResults( std::ostream& outputStream, double endSimTime ) const
{
  if (!allMeasuresList_.empty())
  {
    outputStream << std::endl
                 << " ***** Measure Functions ***** " << std::endl
                 << std::endl;

    // loop over active measure objects and get the results
    for (MeasurementVector::const_iterator it = allMeasuresList_.begin(), end = allMeasuresList_.end(); it != end; ++it)
    {
      // this function prints out diagnostic information if the measure failed.  An example
      // is a nonsensical time window.  It currently prints no info for frequency-based measures
      (*it)->printMeasureWarnings( endSimTime );
      // this function prints the measure value, and measure times if appropriate for a given measure such as
      // MIN, MAX, PP, TRIG/TARG
      (*it)->printMeasureResult( outputStream, true ); 
      // this function print out information about the time window used.  It prints out
      // no information for frequency-based measures  
      (*it)->printMeasureWindow( outputStream, endSimTime );
      // this function prints out information about the time window used for Rise/Fall/Cross
      // if one of those keywords was specified
      (*it)->printRFCWindow( outputStream );
      outputStream << std::endl; // separate verbose output by one line
    }
  }
  
  return outputStream;
}

//-----------------------------------------------------------------------------
// Function      : Manager::getMeasureValue
// Purpose       : Get the value of a .measure test
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 04/29/2010
//-----------------------------------------------------------------------------
bool Manager::getMeasureValue(const std::string &name, double &value) const
{
  for (MeasurementVector::const_iterator it = allMeasuresList_.begin(); it != allMeasuresList_.end(); ++it)
  {
    if (equal_nocase((*it)->name_, name))
    {
      value = (*it)->getMeasureResult();
      return true;
    }
  }
  return false;
}

const Base *Manager::find(const std::string &name) const
{
  for (MeasurementVector::const_iterator it = allMeasuresList_.begin(); it != allMeasuresList_.end(); ++it)
  {
    if (equal_nocase((*it)->name_, name))
    {
      return *it;
    }
  }
  return 0;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::remeasure
// Purpose       : Recompute measure functions based on existing Xyce output
// Special Notes :
// Creator       : Rich Schiek
// Creation Date : 4/8/13
//-----------------------------------------------------------------------------
void Manager::remeasure(N_PDS_Comm &pds_comm, const std::string &netlist_filename, const std::string &remeasure_path, bool transient_flag, Util::Op::BuilderManager &op_builder_manager, OutputMgr &output_manager, Util::SymbolTable &symbol_table)
{
  Xyce::lout() << "In OutputMgr::remeasure " << std::endl
               << "file to reprocess through measure functions. " << remeasure_path << std::endl;

  // open file for reading
  // just support PRN format for now
  OutputFileBase * fileToReMeasure = new OutputPrn();
  if (!fileToReMeasure->openFileForRead(remeasure_path))
  {
    // open failed.  report error and exit remeasure routine

    return;
  }

  // load data-names & allocate space for a line
  std::vector<std::string> fileVarNames;
  if (!(fileToReMeasure->getOutputVarNames(fileVarNames)))
  {
    // reading var names failed.  report error and exit remeasure routine.

  }

  // get column names for building up a symbol table
  fileToReMeasure->convertOutputNamesToSolVarNames(fileVarNames);
  
  // assume analysis type is transient Xyce hasn't processed the analysis type yet
  // when this function is called.  In the next loop if we find a column of
  // data for TIME then we can treat this as a transient analysis.

  int timeIndex=0;
  // set up allNodes map to map var names to indices in the solution vector
  int numVars = fileVarNames.size();
  int numLocalVars = 0;

  // create an Linear::Vector to hold the data from the file.  This is
  // needed to support getPrintValue_ for evaluating the measure functions.
  std::vector<int> lbMap;
  if (Parallel::rank(pds_comm.comm()) == 0)
  {
    numLocalVars = numVars;
  }
  else
  {
    numLocalVars = 0;
  }

  lbMap.resize(numLocalVars);
  for (int i=0;i<numLocalVars; i++)
  {
    lbMap[i]=i;
  }

  for (int i=0; i<numVars; i++)
  {
    if(Parallel::rank(pds_comm.comm()) == 0)
      symbol_table[Util::SOLUTION_SYMBOL][fileVarNames[i]] = i;
    else
      symbol_table[Util::SOLUTION_SYMBOL][fileVarNames[i]] = -1;

    // while scanning fileVarNames look for "TIME" as a name in the 0th or 1st
    // column.  We will use this as a key to figure out if the output file is
    // transient or DC data(no support for .measure in other analysis types yet)
    if ((i<2) &&(fileVarNames[i]=="TIME"))
    {
      if(Parallel::rank(pds_comm.comm()) == 0)
        timeIndex = i;
      else
        timeIndex = -1;
    }
  }

  // call make_ops to get the measure functions ready to get data
  makeMeasureOps(pds_comm.comm(), op_builder_manager);
  N_PDS_ParMap aParMap(numVars, numLocalVars, lbMap, 0, pds_comm);
  Linear::Vector varValuesVec(aParMap);
  varValuesVec.putScalar(0);
  
  // run though lines in the file calling update measure as we go.
  int reading[1], summedReading[1];
  reading[0] = 0;
  summedReading[0] = 1;
  while (summedReading[0]==1)
  {
    if (Parallel::rank(pds_comm.comm()) == 0)
    {  
      reading[0] = fileToReMeasure->getOutputNextVarValues(&varValuesVec);
      summedReading[0] = reading[0];
 
    }
    
    // reduce all procs on reading[] so all procs know when we are done.
    //varValuesVec.fillComplete();
    //varValuesVec.importOverlap();
    pds_comm.sumAll( reading, summedReading,  1);
    
    if( summedReading[0] == 1 )
    {
      if (transient_flag)
      {
        double lt[1], sumlt[1];
        lt[0] = 0.0;
        sumlt[0] = 0.0;
        if( numLocalVars > 0 )
        {
          lt[0] = varValuesVec[timeIndex];
        }
        pds_comm.sumAll( lt, sumlt,  1);
        output_manager.setCircuitTime(sumlt[0]);
        updateTranMeasures(pds_comm.comm(), output_manager.getCircuitTime(), &varValuesVec, 0, 0, 0, 0, 0 );
      }
      else
      {
        updateDCMeasures(pds_comm.comm(), output_manager.getDCSweepVector(), &varValuesVec, 0, 0, 0, 0, 0 );
      }
      varValuesVec.putScalar(0);
    }

  }

  // The use of this black hole stream was replaced with calls to Xyce::lout()
  // This is a null output stream that helps ensure a function that needs to be called
  // on every processor in parallel only outputs on one processor.
  //Teuchos::oblackholestream outputBHS;

  // Output the Measure results to Xyce::dout().
  // Make sure the function gets called on all processors, but only one outputs it.
  if (Parallel::rank(pds_comm.comm()) == 0)
  {
    outputVerboseResults( Xyce::lout(), output_manager.getCircuitTime() );
  }
  else
  {
    outputVerboseResults( Xyce::lout(), output_manager.getCircuitTime() );
  }

  // Output the Measure results to file.
  if (Parallel::rank(pds_comm.comm()) == 0)
  {
    // Adding "0" to the end of this string for the step number, which was always 0
    // for .measure previously anyways.
    std::string filename = netlist_filename + ".mt0";
    std::ofstream outputFileStream(filename.c_str());
    outputResults(outputFileStream);
    outputFileStream.close();
  }
  else {
    outputResults( Xyce::lout() );
  }

  fileToReMeasure->closeFileForRead();

  delete fileToReMeasure;
  fileToReMeasure = 0;
}

namespace {

//-----------------------------------------------------------------------------
// Class         : MeasureOptionsReg
// Purpose       : functor for registering Measure options
// Special Notes : Used by package manager addOptionsProcessor method
// Creator       : dgbaur
// Creation Date : 06/28/2013
//-----------------------------------------------------------------------------
struct MeasureOptionsReg : public PkgOptionsReg
{
  MeasureOptionsReg( Measure::Manager &mgr )
    : measureManager_(mgr)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return measureManager_.addMeasure( options ); }

  Measure::Manager &measureManager_;
};

//-----------------------------------------------------------------------------
// Function      : extractMEASUREData
// Purpose       : Convert a .measure line to an options block
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, Electrical and Microsystems Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
bool
extractMEASUREData(
  IO::PkgOptionsMgr &           options_manager,
  IO::CircuitBlock &            circuit_block,
  const std::string &           netlist_filename,
  const IO::TokenVector &       parsed_line)
{
  // .measure has a lot of optonal and complex syntax.
  //
  // .measure mode name type output_var... options...
  //
  // where mode = AC, DC or TRAN
  //       name = any text string for a name (can't be AC, DC or TRAN)
  //       type = keywords that let us figure out what type of measurement is needed
  //              TRIG and/or TARG = RiseFallDelay
  //              AVG, MAX, MIN, PP, RMS, INTEG = Statistics
  //              FIND/WHEN = FindWhen
  //              PARAM = Equation evaluation
  //              DERIVATIVE or DERIV = Derivative
  //              INTEGRAL = Integral
  //              ERROR = Error
  //              FOUR = Fourier analysis (similar to .FOUR)
  //       output_var = simulation variale to be measured.  This can be any of the following
  //              v(a), v(a,b), v(a)=number v(a)=v(b), i(a) ix(a) or an expression.
  //              So it really could be one or more
  //              variables and or a real number.  The comparison is always equity.
  //       options = these are keywords=value pairs that set limits on when or how
  //              the measurement is done. Value is usually a number, but can be a string in
  //              at least one case.
  //

  // we will use these sets to test for measure types and keywords.
  std::set<std::string> typeSet;
  typeSet.insert( std::string("TRIG") );
  typeSet.insert( std::string("TARG") );
  typeSet.insert( std::string("AVG") );
  typeSet.insert( std::string("MAX") );
  typeSet.insert( std::string("MIN") );
  typeSet.insert( std::string("PP") );
  typeSet.insert( std::string("RMS") );
  typeSet.insert( std::string("FREQ") );
  typeSet.insert( std::string("FIND") );
  typeSet.insert( std::string("WHEN") );
  typeSet.insert( std::string("PARAM") );
  typeSet.insert( std::string("EQN") );
  typeSet.insert( std::string("DERIVATIVE") );
  typeSet.insert( std::string("DERIV") );
  typeSet.insert( std::string("DUTY") );
  typeSet.insert( std::string("INTEGRAL") );
  typeSet.insert( std::string("INTEG") );
  typeSet.insert( std::string("ERROR") );
  typeSet.insert( std::string("ON_TIME") );
  typeSet.insert( std::string("OFF_TIME") );
  typeSet.insert( std::string("FOUR") );

  std::set<std::string> keywords;
  std::set<std::string> simpleKeywords;
  std::set<std::string> numOrTextKeywords;

  simpleKeywords.insert( std::string("TD") );
  simpleKeywords.insert( std::string("GOAL") );
  simpleKeywords.insert( std::string("WEIGHT") );
  simpleKeywords.insert( std::string("MINVAL") );
  simpleKeywords.insert( std::string("AT") );
  simpleKeywords.insert( std::string("FROM") );
  simpleKeywords.insert( std::string("TO") );
  simpleKeywords.insert( std::string("IGNORE") );
  simpleKeywords.insert( std::string("YMIN") );
  simpleKeywords.insert( std::string("YMAX") );
  simpleKeywords.insert( std::string("ON") );
  simpleKeywords.insert( std::string("OFF") );
  simpleKeywords.insert( std::string("FRAC_MAX") );
  simpleKeywords.insert( std::string("MIN_THRESH") );
  simpleKeywords.insert( std::string("MAX_THRESH") );
  simpleKeywords.insert( std::string("NUMFREQ") );
  simpleKeywords.insert( std::string("GRIDSIZE") );
  simpleKeywords.insert( std::string("DEFAULT_VAL") );
  simpleKeywords.insert( std::string("INDEPVARCOL") );
  simpleKeywords.insert( std::string("INDEPVAR2COL") );
  simpleKeywords.insert( std::string("DEPVARCOL") );

  numOrTextKeywords.insert( std::string("RISE") );
  numOrTextKeywords.insert( std::string("FALL") );
  numOrTextKeywords.insert( std::string("CROSS") );
  numOrTextKeywords.insert( std::string("FILE") );
  numOrTextKeywords.insert( std::string("COMP_FUNCTION") );
  numOrTextKeywords.insert( std::string("INDEPVARCOL") );
  numOrTextKeywords.insert( std::string("DEPVARCOL") );
  numOrTextKeywords.insert( std::string("PRECISION") );
  // make a union for the keywords set
  set_union( simpleKeywords.begin(), simpleKeywords.end(), numOrTextKeywords.begin(),
             numOrTextKeywords.end(), std::inserter<std::set<std::string> >(keywords, keywords.begin()) );

  Util::OptionBlock option_block("MEASURE", Util::OptionBlock::ALLOW_EXPRESSIONS, netlist_filename, parsed_line[0].lineNumber_);

  int numFields = parsed_line.size();

  if( numFields < 4 )
  {
    Report::UserError().at(netlist_filename, parsed_line[0].lineNumber_)
      << "Too few items on .MEASURE line.  Need at least .MEASURE <mode> <name> <type>";
    return false;
  }

  Util::Param parameter;

  // look for the fixed items
  ExtendedString currentWord(parsed_line[1].string_);
  currentWord.toUpper();
  if (currentWord == "TR")
  {
    currentWord = "TRAN"; // TR is a synonym for TRAN
  }

  if( currentWord == "DC" || currentWord == "TRAN" || currentWord == "AC" )
  {
    parameter.set("MODE", std::string(currentWord));
    option_block.addParam(parameter);
  }
  else
  {
    Report::UserError0().at(netlist_filename, parsed_line[1].lineNumber_) << "Unknown mode in .MEASURE line.  Should be DC, AC or TRAN/TR";
  }

  currentWord = parsed_line[2].string_;
  currentWord.toUpper();
  if (currentWord == "TR")
  {
    currentWord = "TRAN"; // TR is a synonym for TRAN
  }
  if( currentWord != "DC" && currentWord != "TRAN" && currentWord != "AC" )
  {
    parameter.set("NAME", std::string(currentWord));
    option_block.addParam(parameter);
  }
  else
  {
    Report::UserError0().at(netlist_filename, parsed_line[2].lineNumber_) << "Illegal name in .MEASURE line.  Cannot be AC, DC or TRAN/TR ";
  }

  currentWord = parsed_line[3].string_;
  currentWord.toUpper();
  if( typeSet.find( currentWord ) != typeSet.end() )
  {
    parameter.set("TYPE", std::string(currentWord));
    option_block.addParam(parameter);
  }
  else
  {
    Report::UserError0().at(netlist_filename, parsed_line[3].lineNumber_) << "Illegal type in .MEASURE line.  Must be one of: TRIG, TARG, AVG, MAX, MIN, PP, RMS, INTEG, FIND, WHEN, PARAM, DERIVATIVE, DERIV, INTEGRAL, ERROR, FOUR, DUTY, OFF_TIME, ON_TIME";
  }

  // already got MEASURE <DC|AC|TRAN> name TYPE
  // now try and parse of the rest of the line catching keywords when they're found
  int position = 4;
  while ( position < numFields )
  {
    currentWord = parsed_line[position].string_;
    currentWord.toUpper();
    // The type tag, TARG can get repeated, so check for it with other params.
    if( typeSet.find( currentWord ) != typeSet.end() )
    {
      parameter.set("TYPE", std::string(currentWord));
      option_block.addParam(parameter);
    }
    else if( numOrTextKeywords.find( currentWord ) != numOrTextKeywords.end() )
    {
      // these can be in the form of TAG=value or TAG=keyword where keyword = LAST
      // and potentially the "=" is optional.
      // IF and first ELSE-IF block needed to protect against core dumps from invalid .MEASURE lines.
      if ( (position+1) >= parsed_line.size() )
      {
        // will core dump later if this is true, so error out now
  Report::UserError0().at(netlist_filename, parsed_line[position].lineNumber_) << "Invalid Measure Line";
      } 
      else if ( (parsed_line[(position+1)].string_ == "=") && ( (position+2) >= parsed_line.size() ) )
      {
        // will core dump later if this is true, so error out now
  Report::UserError0().at(netlist_filename, parsed_line[position].lineNumber_) << "Invalid Measure Line";
      }
      else if( ((position+1) < numFields) || ((parsed_line[(position+1)].string_ == "=") && ((position+2) < numFields)) )
      {
        // value is in next or next + 1 position
        int valPosition = ++position;
        if( parsed_line[valPosition].string_ == "=" )
        {
          valPosition = ++position;
        }
        const std::string & value = parsed_line[valPosition].string_;
        if( Util::isInt(value) )
        {
          parameter.set(currentWord, Util::Ival(value) );
        }
        else if( Util::isValue(value) )
        {
          int valAsInt = static_cast<int>(Util::Value(value));
          parameter.set(currentWord, valAsInt);
        }
        else
        {
          parameter.set(currentWord, value);
        }
        option_block.addParam(parameter);
      }
      else
      {
        Report::UserError0().at(netlist_filename, parsed_line[position].lineNumber_) << "Incomplete .MEASURE line.  RISE, FALL, CROSS must be followed by a value or LAST";
      }

    }
    else if( simpleKeywords.find( currentWord ) != simpleKeywords.end() )
    {
      // these are in the form of TAG=value and potentially the "=" is optional.
      // IF and first ELSE-IF block needed to protect against core dumps from invalid lines.
      if ( (position+1) >= parsed_line.size() )
      {
        // will core dump if this is true, so error out now
  Report::UserError0().at(netlist_filename, parsed_line[position].lineNumber_) << "Invalid Measure Line";
      } 
      else if ( (parsed_line[(position+1)].string_ == "=") && ( (position+2) >= parsed_line.size() ) )
      {
        // will core dump if this is true
  Report::UserError0().at(netlist_filename, parsed_line[position].lineNumber_) << "Invalid Measure Line";
      }
      else if( ((position+1) < numFields) || ((parsed_line[(position+1)].string_ == "=") && ((position+2) < numFields)) )
      {
        // value is in next or next + 1 position
        int valPosition = ++position;
        if( parsed_line[valPosition].string_ == "=" )
        {
          valPosition = ++position;
        }
        const std::string & value = parsed_line[valPosition].string_;
        if( Util::isValue(value) )
        {
          parameter.set(currentWord, Util::Value(value) );
        }
        else if( Util::isInt(value) )
        {
          parameter.set(currentWord, Util::Ival(value) );
        }
        else
        {
          Report::UserError0().at(netlist_filename, parsed_line[position].lineNumber_) << "Incomplete .MEASURE line.  TD, GOAL, WEIGHT, MINVAL, AT, TO, IGNORE, YMIN, YMAX must be followed by a value";
        }
        option_block.addParam(parameter);
      }
      else
      {
        Report::UserError0().at(netlist_filename, parsed_line[position].lineNumber_) << "Incomplete .MEASURE line.  TD, GOAL, WEIGHT, MINVAL, AT, TO, IGNORE, YMIN, YMAX must be followed by a value";
      }
    }
    else
    {
      if( (currentWord[0]=='{') && (currentWord[currentWord.size()-1]=='}') )
      {
        parameter.set(currentWord, std::string(currentWord));
        option_block.addParam(parameter);
      }
      else
      {
      // the last form of legal syntax is out_var, out_var=val or out_var1=out_var2
      // first figure out how many positions between here and the next legal keyword or
      // the end of the line
      int endPosition=position;
      bool endNotFound=true;
      ExtendedString nextWord("");
      while( endNotFound )
      {
        endPosition++;
        if( endPosition == numFields)
        {
          endNotFound=false;
        }
        else
        {
          nextWord = parsed_line[endPosition].string_;
          nextWord.toUpper();
          if( (keywords.find( nextWord ) != keywords.end()) || (typeSet.find( nextWord ) != typeSet.end()) )
          {
            // found a keyword or type was found so we're at the end of space to
            // use in finding out_var, out_var, out_var=val or out_var1=out_var2
            endNotFound=false;
          }
        }
      }
      // ok, package up the output vars
      // it will be in the form ( [] indicate optional items )
      // V or I nodeName [nodeName] [ val | V or I nodeName [nodeName] ]
      // or { } expression delimited.  It might also use N, P or W.

      while( position < endPosition )
      {
        nextWord = parsed_line[position].string_;
        nextWord.toUpper();
        // the second part of this if clause is to ensure we don't catch keywords that start
        // with I, V, N, P or W, and mistake them for Ixxx( ) or Vxxx()
        if( (nextWord[0] == 'I' || nextWord[0] == 'V' || nextWord[0] == 'N' 
             || nextWord[0] == 'P' ||  nextWord[0] == 'W') &&
            (simpleKeywords.find( nextWord ) == simpleKeywords.end() ) )
        {
          // need to do a bit of look ahead here to see if this is a V(a) or V(a,b)
          int numNodes = 1;
          if( ((position+3) < endPosition) && (parsed_line[position+3].string_ == ",") )
          {
            numNodes = 2;
          }
          parameter.set(nextWord, numNodes);
          option_block.addParam( parameter );

          if( (position+2) < endPosition )
          {
            position+=2;
            nextWord = parsed_line[position].string_;
            nextWord.toUpper();
            parameter.set( nextWord, 0.0 );
            option_block.addParam( parameter );

            if( ((position+2) < endPosition) && (parsed_line[position+1].string_ == ",") )
            {
              // may have a voltage difference request
              position+=2;
              nextWord = parsed_line[position].string_;
              nextWord.toUpper();
              parameter.set( nextWord, 0.0 );
              option_block.addParam( parameter );
            }
          }
          else
          {
            Report::UserError0().at(netlist_filename, parsed_line[position].lineNumber_) << "Error in .MEASURE line.  Could not parse voltage/current variable";
          }

        }
        else if( nextWord.isValue() )
        {
          std::string objValue("OBJVAL");
          parameter.set( objValue, nextWord.Value());
          option_block.addParam( parameter );
        }
        else if( nextWord.isInt() )
        {
          std::string objValue( "OBJVAL");
          parameter.set( objValue, nextWord.Ival());
          option_block.addParam( parameter );
        }
        else if( nextWord.possibleParam() )
        {
          // could be a param or measure name that will be evaluated
          // during the simulation.
          std::string objValue( "OBJVAL");
          parameter.set( objValue, std::string(nextWord));
          option_block.addParam( parameter );
        }
        position++;
      }
      // reset the position indicator to the end - 1 because
      // we're going to increment it at the end of the while loop
      position = endPosition - 1;
      }

    }
    position++;
  }

  circuit_block.addOptions(option_block);

  return true;
}

} // namespace <unnamed>

//-----------------------------------------------------------------------------
// Function      : registerPkgOptionsMgr
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, 1437
// Creation Date : 10/21/08
//-----------------------------------------------------------------------------
bool registerPkgOptionsMgr(Manager &manager, PkgOptionsMgr &options_manager)
{
  options_manager.addCommandParser(".MEAS", extractMEASUREData);
  options_manager.addCommandParser(".MEASURE", extractMEASUREData);

  options_manager.addCommandProcessor("MEASURE", new MeasureOptionsReg(manager));

  return true;
}

} // namespace Measure
} // namespace IO
} // namespace Xyce
