//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_IO_CircuitBlock.h,v $
//
// Purpose        : Declare the circuit level containers for holding netlist
//                  circuit data and the associated circuit level methods.
//
// Special Notes  :
//
// Creator        : Lon Waters, SNL
//
// Creation Date  : 09/06/2001
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.118 $
//
// Revision Date  : $Date: 2015/11/04 01:15:49 $
//
// Current Owner  : $Author: hkthorn $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_IO_CircuitBlock_h
#define Xyce_N_IO_CircuitBlock_h

#include <fstream>
#include <iostream>
#include <list>
#include <map>
#include <string>
#include <vector>

#include <N_IO_fwd.h>
#include <N_TOP_fwd.h>

#include <N_IO_SpiceSeparatedFieldTool.h>
#include <N_IO_ParameterBlock.h>
#include <N_IO_DeviceBlock.h>
#include <N_UTL_OptionBlock.h>

namespace Xyce {
namespace IO {

//-----------------------------------------------------------------------------
// Class         : CircuitBlock
// Purpose       :
// Special Notes :
// Creator       : Lon Waters, SNL
// Creation Date : 09/06/2001
//-----------------------------------------------------------------------------

class CircuitBlock
{
public:
  CircuitBlock(
     const std::string &                                                fileName,
     const CmdParse &                                                   command_line,
     HangingResistor &                                                  hanging_resistor, 
     CircuitMetadata &                                                  md,
     unordered_set<std::string> &                                       mn,
     std::map<std::string,FileSSFPair> &                                ssfm,
     std::map<std::string,IncludeFileInfo> &                            iflm,
     CircuitContext &                                                   cc,
     int &                                                              uc,
     Topo::Topology &                                                   topology,
     Device::DeviceMgr &                                                device_manager,
     Device::DeviceNameInstanceBlockMap &                               dNames,
     unordered_set<std::string> &                                       nNames,
     AliasNodeMap &                                                     alias_node_map,
     const std::vector< std::pair< std::string, std::string> > &        externalNetlistParams);

  CircuitBlock(
     const std::string &                                                fileName,
     const CmdParse &                                                   command_line,
     HangingResistor &                                                  hanging_resistor, 
     CircuitMetadata &                                                  md,
     unordered_set<std::string> &                                       mn,
     std::map<std::string,FileSSFPair> &                                ssfm,
     std::map<std::string,IncludeFileInfo> &                            iflm,
     CircuitContext &                                                   cc,
     int &                                                              uc,
     CircuitBlock *                                                     mainCircPtr,
     CircuitBlock *                                                     parentCircPtr,
     Topo::Topology &                                                   topology,
     Device::DeviceMgr &                                                device_manager,
     Device::DeviceNameInstanceBlockMap &                               dNames,
     unordered_set<std::string> &                                       nNames,
     AliasNodeMap &                                                     alias_node_map,
     const std::vector< std::pair< std::string, std::string> > &        externalNetlistParams,
     std::vector<bool> &                                                pFilter,
     bool                                                               removeRedundant);

  ~CircuitBlock();

private:
  // Copy Constructor.
  CircuitBlock(CircuitBlock const& rhsCB);
  CircuitBlock &operator=(const CircuitBlock& rhsCB);

public:

  // Get the name of the circuit, if this is the root
  const std::string &getTitle() const
  {
    return title_;
  }

  // Get the name of the subcircuit, if this is a subcircuit  
  const std::string& getName() const
  {
    return name_;
  }

  // Get the name of the netlist file
  const std::string& getNetlistFilename() const
  {
    return netlistFilename_;
  }

  AliasNodeMap &getAliasNodeMap()
  {
    return aliasNodeMap_;
  }

  const unordered_set< std::string > &getAliasNodeMapHelper() const 
  {
    return aliasNodeMapHelper_;
  }

  const std::vector<bool>& getPreprocessFilter() const
  {
    return preprocessFilter_;
  }

  const HangingResistor &getHangingResistor() const
  {
    return hangingResistor_;
  }

  HangingResistor &getHangingResistor()
  {
    return hangingResistor_;
  }

  CircuitContext* getCircuitContextPtr()
  {
    return &circuitContext_;
  }

  bool isSubcircuit() const
  {
    return parentCircuitPtr_ != NULL;
  }

  // This function parses the netlist file and fills in the
  // details of the circuit. This is phase 1 of netlist parsing.
  // The devices cannot be completely handled in this phase.
  bool parseNetlistFilePass1(PkgOptionsMgr &options_manager);
  bool parseNetlistFilePass1(PkgOptionsMgr &options_manager, const std::string &libSelect, std::string libInside);

  // Perform special pass for mutual inductances
  bool parseMutualInductances();

  // Set data_->ssfPtr_ .
  void setSSFPtr( SpiceSeparatedFieldTool* ssfPtr )
  {
    ssfPtr_ = ssfPtr;
  }

  SpiceSeparatedFieldTool* getSSFPtr()
  {
    return ssfPtr_;
  }

  std::map<std::string,FileSSFPair>& getSSFMap()
  {
    return ssfMap_;
  }

  void setFilePosition(std::streampos const& position);
  void setLinePosition(int const& position);
  void setStartPosition();
  void setEndPosition();

  const std::streampos getStartPosition() const { 
    return fileStartPosition_; 
  }
  const std::streampos getEndPosition() const { 
    return fileEndPosition_; 
  }
  int getLineStartPosition() const { 
    return lineStartPosition_; 
  }
  int getLineEndPosition() const { 
    return lineEndPosition_; 
  }

  // At the top level reset the file and line positions and skip the comment line.
  void resetSSFPtr();

  // Add a mutual inductor to the circuit.
  void addMutualInductor( DeviceBlock& device, CircuitContext* context );

  // Add a model to the circuit.
  void addModel(const ParameterBlock * model, std::string const& modelPrefix);

  // Add a device to the circuit.
  void addTableData(DeviceBlock & device );

  // Add a set of options corresponding to a .OPTIONS netlist line
  // to the circuit.
  void addOptions(const Util::OptionBlock &options);

  void addParams(const Util::OptionBlock &options);
  void addGlobalParams(const Util::OptionBlock &options);

  // Search the subcircuitInstanceTable of the current circuit block for the
  // subcircuit of the given name. If it is not found, recursively
  // search each parent subcircuit. Return a pointer to the circuit
  // block if it is found, otherwise return NULL.
  CircuitBlock* findSubcircuit( std::string const& subcircuitName );

  // Receive the circuit context (from the distribution tool).
  bool receiveCircuitContext(const CircuitContext&  ccIn);

  // Change netlist file name
  void setFileName ( const std::string & fileNameIn );

  // resolve expressions in optionBlocks like .print
  bool resolveExpressionsInOptionBlocks();

  // write out a copy of the netlist if it is requested.
  void writeOutNetlist();

  // Print the contents of CircuitBlock.
  void print();

  // Get information from handleAnalysis method.
  std::string& getAnalysisName() { return analysisName_; }
  bool getMORFlag() { return morFlag_; }

  const std::list<Util::OptionBlock>& getOptionsTable() const
  {
    return optionsTable_;
  }

  std::list<Util::OptionBlock>& getOptionsTable()
  {
    return optionsTable_;
  }

  CircuitMetadata& getMetadata()
  {
    return metadata_;
  }

public:

  // Lookup table for initial conditions
  std::map< std::string, TokenVector > initCondIndex;

private:
  std::string                           netlistFilename_;
  std::string                           title_;                   // For top level circuit, given by first line of netlist
  std::string                           name_;                    // For subcircuits
  std::string                           analysisName_;

  unordered_set<std::string> &          nodeNames_;
  unordered_set<std::string> &          modelNames_;
  ModelMap                              modMap_;
  Device::DeviceNameInstanceBlockMap &  deviceNames_;
  std::list<Util::OptionBlock>          optionsTable_;

  int & useCount_;  // Counts copies so deletion of circuitBlockTable_ 
                    // can be done properly in the class destructor.
  unordered_map<std::string, CircuitBlock *> circuitBlockTable_;

  // keep track of where include files are being called from.
  std::map<std::string, IncludeFileInfo> & includeFileLocation_; 
  
  // keep track of K lines that need extracted
  std::multimap< CircuitContext *, DeviceBlock > rawMIs_;

  const CmdParse &                      commandLine_;
  HangingResistor &                     hangingResistor_;
  
  // Circuit Context object to hold context information.
  CircuitContext & circuitContext_;

  CircuitMetadata & metadata_;

  std::vector< std::pair< std::string, std::string> > externalNetlistParams_;

  // Checking the netlist syntax.
  bool netlistSave_;

  // Check if MOR is requested.
  bool morFlag_;

  int devProcessedNumber_;

  // This is a map of node aliases.  Interface nodes to a subcircuit are removed as
  // the subcircuit is expanded into the netlist. We'll store the names of the interface
  // nodes in case the user accesses them elsewhere (as in a print statement)
  // The keys are the alias names and the values are the real cicuit node names
  AliasNodeMap &                aliasNodeMap_;
  unordered_set< std::string >  aliasNodeMapHelper_;

  std::ifstream* netlistIn_;

  SpiceSeparatedFieldTool* ssfPtr_;

  std::streampos fileStartPosition_;
  std::streampos fileEndPosition_;
  int lineStartPosition_;
  int lineEndPosition_;

  // Top level circuit pointer
  CircuitBlock* mainCircuitPtr_;

  // For subcircuits, points to the circuitBlock instance that contains
  // this subcircuit. NULL for top level.
  CircuitBlock* parentCircuitPtr_; 

  std::map<std::string,FileSSFPair> & ssfMap_;

  DeviceBlock device_;
  ParameterBlock tmpModel_;

  // This vector keeps track of any filters that are enabled to check for removal of 
  // "redundant" devices (where all device nodes are the same) or replace any occurrence 
  // of "GND", "GND!", or "GROUND" with "0" (so that all four terms are synonyms)
  std::vector<bool> preprocessFilter_;
  bool remove_any_redundant_;

  Topo::Topology &              topology_;
  Device::DeviceMgr  &          deviceManager_;

  //This function preprocesses the netlist file to provide the user the
  //option of removing "redundant" devices (devices where all the nodes are
  //the same.  The info gathered here affects the phase 1 and phase 2 parse.
  bool parsePreprocess();

  //This function will reproduce a copy of the netlist file under the name
  //netlistfilename_copy.cir (to be used, in the end, to produce netlist files
  //which contain large resistors connecting "dangling" nodes to ground.
  void produceUnflattenedNetlist();

  // Handle a netlist line, determine the line type and take the
  // appropriate action.
  bool handleLinePass1( bool & result,
                        PkgOptionsMgr &options_manager, std::map<std::string,int> & par,
                        std::map<std::string,int> & fun, ModelMap & modMap,
                        std::map<std::string,int> & sub, 
                        const std::string &libSelect, std::string &libInside );

  // Handle a line for Mutual Inductance Pass
  bool getLinePassMI();

  // Post process analysis commands.
  bool handleAnalysis();

  // Extract subcircuit data from parsedLine.
  bool extractSubcircuitData(std::string fileName,
                             TokenVector const& parsedLine);

  // Parse the given include file adding the contents
  // to the current CircuitBlock.
  bool parseIncludeFile(
    PkgOptionsMgr &options_manager,
    std::string const& includeFile, std::string const& libSelect,
    std::map<std::string,int> & par, std::map<std::string,int> & fun,
    ModelMap & modMap, std::map<std::string,int> & sub);

  // Retrieve separate IC= data from line or external file and temporarily
  // store in CircuitBlock
  void handleInitCond(TokenVector const& parsedLine );

};

//-----------------------------------------------------------------------------
// Function      : packCircuitOptions
// Purpose       : send option blocks to all procs
// Special Notes :
// Scope         : private
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
int packCircuitOptions(const std::list<Util::OptionBlock>& options, char* char_buffer,
                       int char_buffer_size, Xyce::Parallel::Communicator* pds_comm_ptr);

//-----------------------------------------------------------------------------
// Function      : unpackCircuitOptions
// Purpose       : unpack option blocks from proc 0
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool unpackCircuitOptions(std::list<Util::OptionBlock>& options, char* char_buffer,
                          int bsize, Xyce::Parallel::Communicator* pds_comm_ptr);

} // namespace IO
} // namespace Xyce

#endif // Xyce_N_IO_CircuitBlock_h
