//-------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Filename       : $RCSfile: N_IO_CircuitContext.C,v $
//
// Purpose        : Define the circuit "context".
//
// Special Notes  :
//
// Creator        : Lon Waters
//
// Creation Date  : 01/21/2003
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revsion$
//
// Revision Date   : $Date: 2015/11/06 01:07:01 $
//
// Current Owner  : $Author: hkthorn $
//----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iostream>
#include <sstream>

#include <N_UTL_fwd.h>

#include <N_DEV_DeviceBlock.h>
#include <N_ERH_ErrorMgr.h>
#include <N_IO_CircuitContext.h>
#include <N_IO_DeviceBlock.h>
#include <N_IO_Op.h>
#include <N_IO_ParameterBlock.h>
#include <N_PDS_Comm.h>
#include <N_UTL_Expression.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>

namespace Xyce {
namespace IO {

//----------------------------------------------------------------------------
// Function       : CircuitContext::CircuitContext
// Purpose        : Constructor
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 01/21/2003
//----------------------------------------------------------------------------
CircuitContext::CircuitContext(
  Util::Op::BuilderManager &    op_builder_manager,
  std::list<CircuitContext*> &  context_list,
  CircuitContext *&             current_context_pointer)
  : currentContextPtr_(current_context_pointer),
    opBuilderManager_(op_builder_manager),
    parentContextPtr_(NULL),
    contextList_(context_list),
    name_(""),
    deviceCountDone_(false),
    deviceCount_(0),
    linearDeviceCount_(0),
    subcircuitPrefix_(""),
    resolved_(false),
    resolvedParams_(),
    resolvedGlobalParams_()
{
  if (currentContextPtr_ == NULL)
  {
    currentContextPtr_ = this;
  }
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::~CircuitContext
// Purpose        : Destructor
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 01/21/2003
//----------------------------------------------------------------------------
CircuitContext::~CircuitContext()
{
  // Delete each subcircuit context in this context.
  unordered_map< std::string, CircuitContext * >::iterator itsc = circuitContextTable_.begin();
  unordered_map< std::string, CircuitContext * >::const_iterator itsc_end = circuitContextTable_.end();

  for ( ; itsc != itsc_end; ++itsc )
  {
    delete itsc->second;
  }

  circuitContextTable_.clear();


  // We delete all our own stored model pointers.  Because we also
  // delete all our subcircuit contexts, *their* destructors take care of
  // deleting their model pointers.
  ModelMap::iterator modelIter = models_.begin();
  ModelMap::iterator modelIterEnd = models_.end();
  for (; modelIter != modelIterEnd; ++modelIter)
    delete modelIter->second;
  models_.clear();
}

//-----------------------------------------------------------------------------
// Function      : CircuitContext::operator=
// Purpose       : assignment operator
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 01/18/2006
//-----------------------------------------------------------------------------
CircuitContext & CircuitContext ::operator=(const CircuitContext & right)
{
  if (this != &right)
  {
    currentContextPtr_ = right.currentContextPtr_;
    parentContextPtr_ = right.parentContextPtr_;
    contextList_ = right.contextList_;
    name_ = right.name_;

    deviceCountDone_ = right.deviceCountDone_;
    deviceCount_ = right.deviceCount_;
    linearDeviceCount_ = right.linearDeviceCount_;
    instanceList_ = right.instanceList_;
    usedContextList_ = right.usedContextList_;

    nodeList_ = right.nodeList_;
    subcircuitParameters_ = right.subcircuitParameters_;
    circuitContextTable_ = right.circuitContextTable_;
    models_ = right.models_;

    unresolvedParams_ = right.unresolvedParams_;
    unresolvedGlobalParams_ = right.unresolvedGlobalParams_;
    unresolvedFunctions_ = right.unresolvedFunctions_;
    mutualInductances_ = right.mutualInductances_;

    sharedInductorTable_ = right.sharedInductorTable_;
    allCoupledInductors_ = right.allCoupledInductors_;
    allIndexedMIs_ = right.allIndexedMIs_;
    kLines_ = right.kLines_;

    subcircuitPrefix_ = right.subcircuitPrefix_;
    nodeMap_ = right.nodeMap_;
    resolved_ = right.resolved_;
    resolvedParams_ = right.resolvedParams_;
    resolvedGlobalParams_ = right.resolvedGlobalParams_;
    resolvedFunctions_ = right.resolvedFunctions_;
  }

  return *this;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::addInstance
// Purpose        :
// Special Notes  :
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 02/13/2003
//----------------------------------------------------------------------------
void
CircuitContext::addInstance(std::string const& subcircuitName, std::string const& fileName, int const& lineNumber)
{
  std::string subcircuitNameUpper(ExtendedString(subcircuitName).toUpper());

  currentContextPtr_->instanceList_.push_back(subcircuitNameUpper);
  currentContextPtr_->instanceLocation_[subcircuitNameUpper].push_back( NetlistLocation(fileName, lineNumber) );
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::beginSubcircuitContext
// Purpose        : Add a subcircuit context to the current context.
// Special Notes  : This routine sets the current context to a newly
//                  created CircuitContext object. This context
//                  remains the current context until it is explicitly
//                  terminated with a call to endSubcircuitContext.
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 01/24/2003
//----------------------------------------------------------------------------
bool CircuitContext::beginSubcircuitContext(
  std::string const& netlistFileName,
  TokenVector & subcircuitLine)

{
  if (DEBUG_IO)
  {
    Xyce::dout() << "CircuitContext::beginSubcircuitContext" << std::endl
                 << "*******************************************" << std::endl
                 << " subcircuit file name: " << netlistFileName << std::endl
                 << "  size of line vector: " << subcircuitLine.size() << std::endl;
  }

  // remove parens
  TokenVector::iterator iterLine =
    subcircuitLine.begin();
  while( iterLine != subcircuitLine.end() )
  {
    if ( iterLine->string_ == "(" || iterLine->string_ == ")" )
    {
      subcircuitLine.erase( iterLine );
    }
    else
    {
      iterLine++;
    }
  }

  // Create a new circuit context for the subcircuit.
  CircuitContext* subcircuitContextPtr =
    new CircuitContext(opBuilderManager_, contextList_, currentContextPtr_);

  // Set the parent context, save the current context and reset it to the
  // newly created context.
  subcircuitContextPtr->parentContextPtr_ = currentContextPtr_;
  contextList_.push_front(currentContextPtr_);
  currentContextPtr_ = subcircuitContextPtr;

  // Save location of this subcircuit definition
  NetlistLocation newSubcktLoc( netlistFileName, subcircuitLine[0].lineNumber_ );
  subcircuitContextPtr->setLocation( newSubcktLoc );

  // Extract the subcircuit data from subcircuitLine.
  int numFields = subcircuitLine.size();

  if (numFields < 2)
  {
    Report::UserError0().at(netlistFileName, subcircuitLine[0].lineNumber_)
      << "Subcircuit name required";
    return false;
  }

  if (numFields < 3)
  {
    Report::UserError0().at(netlistFileName, subcircuitLine[0].lineNumber_)
      << "At least one node required for subcircuit " << subcircuitLine[1].string_;
    return false;
  }

  if (DEBUG_IO)
  {
    Xyce::dout() << "  Subcircuit name: " << subcircuitLine[1].string_ << std::endl
                 << "  Number of fields on subckt line = " << numFields << std::endl;
  }

  // Extract and set the subcircuit name.
  ExtendedString field ( subcircuitLine[1].string_ );
  field.toUpper();
  subcircuitContextPtr->setName(field);

  // Extract the subcircuit external nodes.
  int i;
  for (i = 2; i < numFields; ++i)
  {
    ExtendedString fieldES (  subcircuitLine[i].string_ );
    fieldES.toUpper();
    // Exit the loop if there are netlist parameters
    if (fieldES == "PARAMS:")
    {
      break;
    }
    if (i < numFields-1 && subcircuitLine[i+1].string_ == "=")
    {
      --i;
      break;
    }
    if (fieldES == "0")
    {
      Report::UserError0().at(netlistFileName, subcircuitLine[0].lineNumber_) 
        << "Ground node '0' appears in .SUBCKT line";
    }
    subcircuitContextPtr->nodeList_.push_back(fieldES);
  }

  // Extract the subcircuit parameters.
  Util::Param parameter("","");

  bool result = true;
  ++i; // Advance to start of parameters.
  while (i+2 < numFields)
  {
    ExtendedString fieldES = subcircuitLine[i].string_;
    fieldES.toUpper();
    if (!fieldES.possibleParam())
    {
      Report::UserError0().at(netlistFileName, subcircuitLine[i].lineNumber_)
        << "Parameter name " << subcircuitLine[i].string_ <<  " contains illegal character(s)";
      result = false;
    }
    parameter.setTag( fieldES );

    if ( (parameter.uTag() == "TEMP") || (parameter.uTag() == "VT") ||
         (parameter.uTag() == "GMIN") || (parameter.uTag() == "TIME") )
    {
      Report::UserError0().at(netlistFileName, subcircuitLine[i].lineNumber_)
        << "Parameter name " << parameter.uTag() << " not allowed in subcircuit parameter list for subcircuit "
        << getName();
      result = false;
    }

    if ( subcircuitLine[i+1].string_ != "=" )
    {
      Report::UserError0().at(netlistFileName, subcircuitLine[0].lineNumber_)
        << "Equal sign required between parameter and value in PARAM list for subcircuit " << getName();
      result = false;
    }

    else {
      i+=2; // Advance past "=" sign

      fieldES =  subcircuitLine[i].string_;
      fieldES.toUpper();
      parameter.setVal(std::string(fieldES));
      subcircuitContextPtr->subcircuitParameters_.push_back(parameter);
      if (DEBUG_IO)
      {
        Xyce::dout() << " Adding parameter " << parameter.uTag() << " with value " << fieldES << std::endl;
      }
    }
    ++i;
  }

  // check for truncated params: list
  if ( result && i < numFields )
  {
    Report::UserError0().at(netlistFileName, subcircuitLine[0].lineNumber_)
      << "Parameter list error in subcircuit " << getName();
    result = false;
  }

  if (DEBUG_IO)
  {
    Xyce::dout() << "End of CircuitContext::beginSubcircuitContext" << std::endl
                 << "*******************************************" << std::endl;
  }

  return result;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::endSubcircuitContext
// Purpose        : End the current context, push it onto the previous context's
//                  list of contexts, and reset the current context pointer to
//                  the previous context.
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 01/24/2003
//----------------------------------------------------------------------------
bool CircuitContext::endSubcircuitContext()
{
  if (DEBUG_IO)
  {
    Xyce::dout() << "CircuitContext::endSubcircuitContext" << std::endl;
  }

  // Get the previous context from the stack.
  if ( ! contextList_.empty() )
  {
    // Add the current context to the previous context's list of contexts.
    contextList_.front()->circuitContextTable_[ currentContextPtr_->name_ ] =
      currentContextPtr_;

    // Reset the current context to the previous context.
    currentContextPtr_ = contextList_.front();

    // remove from stack
    contextList_.pop_front();
    return true;
  }
  else
  {
    return false;
  }
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::addModel
// Purpose        : Add a model to the current context.
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 02/03/2003
//----------------------------------------------------------------------------
void CircuitContext::addModel(ParameterBlock * modelPtr)
{
  ParameterBlock* tmpModel;
  if (findModel(modelPtr->getName(), tmpModel))
  {
    Report::UserWarning0 message;
    message << "Reading model named " << modelPtr->getName() << " in the ";

    if (getCurrentContextName() == "")
      message << "main circuit";
    else
      message << "subcircuit " << getCurrentContextName();

    message << " and found one or more models previously defined in this scope";
  }

  currentContextPtr_->models_[modelPtr->getName()] = modelPtr;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::addParams
// Purpose        : Add a set of .PARAM parameters to the current context.
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 02/03/2003
//----------------------------------------------------------------------------
void CircuitContext::addParams(
  Util::ParamList::const_iterator paramIter, 
  Util::ParamList::const_iterator paramEnd)
{
  Util::Param parameter;
  for ( ; paramIter != paramEnd; ++paramIter)
  {
    parameter = *paramIter;
    resolveQuote(parameter);
    currentContextPtr_->unresolvedParams_.push_back(parameter);
  }
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::addGlobalParams
// Purpose        : Add a set of .GLOBAL_PARAM parameters to the current context.
// Special Notes  :
// Scope          :
// Creator        : Dave Shirley, PSSI
// Creation Date  : 11/17/2005
//----------------------------------------------------------------------------
void CircuitContext::addGlobalParams(
  Util::ParamList::const_iterator paramIter, 
  Util::ParamList::const_iterator paramEnd)
{
  Util::Param parameter;
  for ( ; paramIter != paramEnd; ++paramIter)
  {
    parameter = *paramIter;
    resolveQuote(parameter);
    currentContextPtr_->unresolvedGlobalParams_.push_back(parameter);
  }
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::addGlobalNode
// Purpose        : Add a global node from a .GLOBAL line
// Special Notes  :
// Scope          :
// Creator        : Dave Shirley, PSSI
// Creation Date  : 10/02/2009
//----------------------------------------------------------------------------
void CircuitContext::addGlobalNode( std::string & gnode)
{
  currentContextPtr_->globalNodes_.insert(gnode);
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::resolveQuote
// Purpose        : Resolve quoted parameters as soon as they are encountered.
// Special Notes  : This avoids every processor in a parallel run trying to
//                  access the same file.  Also exit handling does not work on
//                  error for parallel runs otherwise.
// Scope          :
// Creator        : Dave Shirley, PSSI
// Creation Date  :
//----------------------------------------------------------------------------
void CircuitContext::resolveQuote(Util::Param & parameter)
{
  if (parameter.isQuoted())
  {
    // The parameter is time dependent with its time history defined by the set
    // of time-value pairs in the file given by the value of the parameter.
    // Open and read these values, using them to build a "Table" expression
    // for the value of the parameter.
    std::ifstream paramDataIn;
    std::string parameterFile(parameter.stringValue().substr(1,parameter.stringValue().size()-2));
    paramDataIn.open(parameterFile.c_str(), std::ios::in);
    std::string table("table(time");

    if ( !paramDataIn.is_open() )
    {
      Report::UserFatal0() << "Could not find file " << parameterFile;
    }
    else
    {
      std::string time;
      std::string value;
      while ( paramDataIn >> time )
      {
        if ( paramDataIn >> value )
        {
          table += "," + time + "," + value;
        }
        else
        {
          Report::UserFatal0() << "Reached end of file in " << parameterFile << " while expecting another value";
        }
      }
    }

    paramDataIn.close();

    if( table.size() <= 10 ) // the length of "table(time" from above
    {
      Report::UserFatal0() << "Failed to successfully read contents of " << parameterFile;
    }

    table += ")";

    parameter.setVal( Util::Expression(table) );
    return;
  }
}
//----------------------------------------------------------------------------
// Function       : CircuitContext::addFunction
// Purpose        : Add a .FUNC function to the current context.
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 02/03/2003
//----------------------------------------------------------------------------
void CircuitContext::addFunction(FunctionBlock const& function)
{
  currentContextPtr_->unresolvedFunctions_.push_back(function);
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::resolve
// Purpose        : Resolve parameters in the current context.
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 02/04/2003
//----------------------------------------------------------------------------
bool CircuitContext::resolve( std::vector<Device::Param> const& subcircuitInstanceParams )
{
  Util::ParamList retryParams;
  std::vector<FunctionBlock> retryFunctions;

  if (currentContextPtr_->subcircuitParameters_.empty() &&
      subcircuitInstanceParams.empty() &&
      currentContextPtr_->resolved_)
  {
    // If there are no subcircuit parameters and this context has already
    // been resolved, then no work to do.
    return true;
  }

  if (DEBUG_IO)
  {
    Xyce::dout() << " CircuitContext::resolve has something to do..." << std::endl;
  }

  currentContextPtr_->resolved_ = true;

  // Clear resolved params and funcs.
  currentContextPtr_->resolvedParams_.clear();
  currentContextPtr_->resolvedGlobalParams_.clear();
  currentContextPtr_->resolvedFunctions_.clear();


  Util::ParamList asYetUnresolvedSubcircuitParameters=currentContextPtr_->subcircuitParameters_;
  Util::ParamList asYetUnresolvedParameters=currentContextPtr_->unresolvedParams_;
  Util::ParamList asYetUnresolvedGlobalParameters=currentContextPtr_->unresolvedGlobalParams_;
  std::vector<FunctionBlock> asYetUnresolvedFunctions=currentContextPtr_->unresolvedFunctions_;

  bool resolvedSomethingThisLoop=true;
  bool somethingLeftToDo= (!(asYetUnresolvedSubcircuitParameters.empty()) ||
                           !(asYetUnresolvedParameters.empty()) ||
                           !(asYetUnresolvedGlobalParameters.empty()) ||
                           !(asYetUnresolvedFunctions.empty())  ||
                           !(subcircuitInstanceParams.empty()) );

  while (resolvedSomethingThisLoop && somethingLeftToDo)
  {
    resolvedSomethingThisLoop=false;
    somethingLeftToDo=false;
    retryParams.clear();
    retryFunctions.clear();

    // Add subcircuitParameters_ to the set of resolved parameters.
    // currentContextPtr_->resolvedParams_.addParameters( currentContextPtr_->subcircuitParameters_ );
    Util::Param parameter;
    Util::ParamList::iterator paramIter;
    Util::ParamList::iterator start =
      asYetUnresolvedSubcircuitParameters.begin();
    Util::ParamList::iterator end =
      asYetUnresolvedSubcircuitParameters.end();
    for (paramIter = start; paramIter != end; ++paramIter)
    {
      Util::Param par = *paramIter;
      if (DEBUG_IO)
      {
        Xyce::dout() << " CircuitContext::resolve attempting to resolve" << par.uTag()<< std::endl;
      }

      if (par.getType() == Xyce::Util::STR && !par.isNumeric())
      {
        ExtendedString arg ( par.stringValue() );
        arg.toUpper();
        if (arg.possibleParam())
        {
          par.setVal(std::string("{" + arg + "}"));
        }
      }

      if (!resolveParameter(par))
      {
        if (DEBUG_IO)
        {
          Xyce::dout() << "Unable to resolve subcircuit param " << par.uTag() << std::endl;
        }

        retryParams.push_back(par);
        somethingLeftToDo=true;
      }
      else
      {
        if (DEBUG_IO)
        {
          Xyce::dout() << "resolveParameter returned true on parameter " << par.uTag() << " after resolution its type is " << par.getType() << "with value " ;
          switch (par.getType()) {
            case Xyce::Util::STR:
              Xyce::dout() << par.stringValue();
              break;
            case Xyce::Util::DBLE:
              Xyce::dout() << par.getImmutableValue<double>();
              break;
            case Xyce::Util::EXPR:
              Xyce::dout() << par.getValue<Util::Expression>().get_expression();
              break;
            default:
              Xyce::dout() << par.stringValue();
          }
          Xyce::dout() << std::endl;
        }

        currentContextPtr_->resolvedParams_.push_back(par);
        resolvedSomethingThisLoop=true;
      }
      if (DEBUG_IO)
      {
        Xyce::dout() << " CircuitContext::resolve done attempting to resolve" << par.uTag()<< std::endl;
      }
    }
    asYetUnresolvedSubcircuitParameters=retryParams;
    retryParams.clear();

    Util::Param* paramPtr;
    // Reset the subcircuit parameter values with the instance parameter
    // values as needed.
    int i;
    int numInstanceParameters = subcircuitInstanceParams.size();
    for ( i = 0; i < numInstanceParameters; ++i )
    {
      if (DEBUG_IO)
      {
        Xyce::dout() << " CircuitContext::resolve resetting subcircuit instance parameter" << subcircuitInstanceParams[i].uTag()<< " with value " << std::endl;
        switch (subcircuitInstanceParams[i].getType()) {
          case Xyce::Util::STR:
            Xyce::dout() << subcircuitInstanceParams[i].stringValue();
            break;
          case Xyce::Util::DBLE:
            Xyce::dout() << subcircuitInstanceParams[i].getImmutableValue<double>();
            break;
          case Xyce::Util::EXPR:
          {
            Util::Expression foo(subcircuitInstanceParams[i].getValue<Util::Expression>());
            Xyce::dout() << "EXPR(" << foo.get_expression() << ")";
            break;
          }
          default:
            Xyce::dout() << subcircuitInstanceParams[i].stringValue();
        }
        Xyce::dout() << std::endl;
      }

      // Look for the parameter in resolvedParams_, issue
      // a warning and continue if not found. Otherwise, if found
      // set the value.
      parameter.setTag(subcircuitInstanceParams[i].uTag());
      paramPtr = Util::findParameter(currentContextPtr_->resolvedParams_.begin(), currentContextPtr_->resolvedParams_.end(), parameter.tag());

      if ( paramPtr == NULL )
      {
        currentContextPtr_->resolvedParams_.push_back(subcircuitInstanceParams[i]);
        if (DEBUG_IO)
        {
          Xyce::dout() << " did not find in resolvdParams_, adding parameter " << subcircuitInstanceParams[i].uTag() << std::endl;
        }
      }
      else
      {
        if (DEBUG_IO)
        {
          Xyce::dout() << " found in resolvdParams_, setting parameter " << paramPtr->uTag() << std::endl;
        }

        paramPtr->setVal(static_cast<const Util::Param &>(subcircuitInstanceParams[i]));
      }
    }

    // Resolve any .PARAM parameters in the current context and add to
    // the resolvedParams_.
    start = asYetUnresolvedParameters.begin();
    end = asYetUnresolvedParameters.end();
    for (paramIter = start; paramIter != end; ++paramIter)
    {
      parameter = *paramIter;

      if (!resolveParameter(parameter))
      {
        // save it for later, because it might use functions that haven't
        // been resolved yet.
        retryParams.push_back(parameter);
        somethingLeftToDo=true;
      }
      else
      {
        currentContextPtr_->resolvedParams_.push_back(parameter);
        resolvedSomethingThisLoop=true;
      }
    }
    asYetUnresolvedParameters=retryParams;
    retryParams.clear();

    // Resolve any .GLOBAL_PARAM parameters in the current context and add to
    // the resolvedGlobalParams_.
    start = asYetUnresolvedGlobalParameters.begin();
    end = asYetUnresolvedGlobalParameters.end();
    for (paramIter = start; paramIter != end; ++paramIter)
    {
      parameter = *paramIter;
      if (DEBUG_IO)
      {
        Xyce::dout() << " CircuitContext::resolve Attempting to resolve global parameter " << parameter.uTag();
      }

      if (!resolveParameter(parameter))
      {
        retryParams.push_back(parameter);
        somethingLeftToDo = true;
      }
      else
      {
        if (parameter.getType() == Xyce::Util::EXPR)
        {
          std::vector<std::string> nodes, instances, leads, variables, specials;

          parameter.getValue<Util::Expression>().get_names(XEXP_NODE, nodes);
          parameter.getValue<Util::Expression>().get_names(XEXP_INSTANCE, instances);
          parameter.getValue<Util::Expression>().get_names(XEXP_LEAD, leads);
          parameter.getValue<Util::Expression>().get_names(XEXP_VARIABLE, variables);
          parameter.getValue<Util::Expression>().get_names(XEXP_SPECIAL, specials);

          if (!nodes.empty() || !instances.empty() || !leads.empty())
          {
            Report::UserError0 message;
            message << "The following are not allowed in global param expression: " << parameter.getValue<Util::Expression>().get_expression();
            
            // This should be caught earlier, but just in case it is checked here
            if (!nodes.empty())
            {
              message << std::endl << "node(s):";
              for (std::vector<std::string>::iterator s_i=nodes.begin() ; s_i!=nodes.end() ; ++s_i)
              {
                message << " " << *s_i;
              }
            }
            if (!instances.empty())
            {
              message << std::endl << "instance(s): ";
              for (std::vector<std::string>::iterator s_i=instances.begin() ; s_i!=instances.end() ; ++s_i)
              {
                message << " " << *s_i;
              }
            }
            if (!leads.empty())
            {
              message << std::endl << "lead(s): ";
              for (std::vector<std::string>::iterator s_i=leads.begin() ; s_i!=leads.end() ; ++s_i)
              {
                message << " " << *s_i;
              }
            }
          }

          if (!variables.empty())
          {
            // If variables are found, they must be previously defined global params
            for (std::vector<std::string>::iterator s_i=variables.begin() ; s_i!=variables.end() ; ++s_i)
            {
              if (findParameter(currentContextPtr_->resolvedGlobalParams_.begin(), currentContextPtr_->resolvedGlobalParams_.end(), *s_i) == NULL)
              {
                Report::UserError0() << "Unknown parameter (" << *s_i << ") found in global param expression: "
                                     << parameter.getValue<Util::Expression>().get_expression();
              }
            }
          }
          if (!specials.empty())
          {
            if (specials.size() > 1 || specials[0] != "TIME")
            {
              Report::UserError0 message;
              message << "Unknown special var(s):";
              for (std::vector<std::string>::iterator s_i=specials.begin() ; s_i!=specials.end() ; ++s_i)
              {
                message << " " << *s_i;
              }
            }
          }
        }
        currentContextPtr_->resolvedGlobalParams_.push_back(parameter);
        resolvedSomethingThisLoop=true;
      }
    }
    asYetUnresolvedGlobalParameters=retryParams;
    retryParams.clear();

    // Resolve functions in the current context.
    int numFunctions = asYetUnresolvedFunctions.size();
    for ( i = 0; i < numFunctions; ++i )
    {
      // Define the Util::Param for the function prototype (the function name
      // and arguments) and function body. Pass to the circuit for resolution.
      std::string functionName(asYetUnresolvedFunctions[i].functionName);
      std::string functionNameAndArgs(asYetUnresolvedFunctions[i].functionNameAndArgs);
      std::string functionBody(asYetUnresolvedFunctions[i].functionBody);

      std::vector<std::string> functionArgs =
        asYetUnresolvedFunctions[i].functionArgs;
      Util::Param functionParameter(functionNameAndArgs, functionBody);
      // this is only an error if we have no parameters left to resolve
      if (!resolveParameter(functionParameter, functionArgs))
      {
        retryFunctions.push_back(asYetUnresolvedFunctions[i]);
        somethingLeftToDo=true;
      }
      else
      {
        // After resolution, the only strings allowed in the function
        // are the function arguments, check that this holds.
        Util::Expression functionBodyExpression( functionParameter.stringValue() );
        std::vector<std::string> strings;
        bool canResolveAll=true;

        functionBodyExpression.get_names(XEXP_STRING, strings);

        int numStrings = strings.size();
        for (int j = 0; j < numStrings; ++j)
        {
          // Look for string in functionArgs.
          std::vector<std::string>::iterator stringIter =
            find(functionArgs.begin(), functionArgs.end(), strings[j]);
          if (stringIter == functionArgs.end()
              && findParameter(currentContextPtr_->resolvedGlobalParams_.begin(), currentContextPtr_->resolvedGlobalParams_.end(), strings[j]) == NULL)
          {
            retryFunctions.push_back(asYetUnresolvedFunctions[j]);
            somethingLeftToDo=true;
            canResolveAll=false;
          }
        }

        if (canResolveAll)
        {
          // Reset the function body.
          functionBody = functionParameter.stringValue();

          // Add the function to resolvedFunctions_.
          currentContextPtr_->resolvedFunctions_[functionName] =
            Util::Param(functionNameAndArgs, functionBody);
          resolvedSomethingThisLoop=true;
        }
      }
    }
    asYetUnresolvedFunctions=retryFunctions;
    retryFunctions.clear();
  }

  if (somethingLeftToDo)    // we failed to resolve everything, and the last loop did nothing.
  {
    for (Util::ParamList::iterator it = asYetUnresolvedSubcircuitParameters.begin(); it != asYetUnresolvedSubcircuitParameters.end(); ++it)
    {
      Report::UserError0() << "Unable to resolve .subckt parameter " << (*it).uTag() << " found in .PARAM statement";
    }

    for (Util::ParamList::iterator it = asYetUnresolvedParameters.begin(); it != asYetUnresolvedParameters.end(); ++it)
    {
      Report::UserError0() << "Unable to resolve parameter " << (*it).uTag() << " found in .PARAM statement";
    }

    for (Util::ParamList::iterator it = asYetUnresolvedGlobalParameters.begin(); it != asYetUnresolvedGlobalParameters.end(); ++it)
    {
      Report::UserError0() << "Unable to resolve global parameter " << (*it).uTag() << " found in .PARAM statement";
    }

    for (std::vector<FunctionBlock>::iterator func_it = asYetUnresolvedFunctions.begin(); func_it != asYetUnresolvedFunctions.end(); ++func_it)
    {
      const std::string &functionNameAndArgs = (*func_it).functionNameAndArgs;
      const std::string &functionBody = (*func_it).functionBody;
      const std::vector<std::string> &functionArgs = (*func_it).functionArgs;

      Util::Param functionParameter(functionNameAndArgs, functionBody);
      if (!resolveParameter(functionParameter, (*func_it).functionArgs))
      {
        Report::UserError0() << functionNameAndArgs << " contains an undefined parameter or function.";
      }
      else
      {
        std::vector<std::string> strings;

        Util::Expression functionBodyExpression(functionParameter.stringValue());
        functionBodyExpression.get_names(XEXP_STRING, strings);
        for (std::vector<std::string>::const_iterator it = strings.begin(); it != strings.end(); ++it)
        {
          if (find(functionArgs.begin(), functionArgs.end(), *it) == functionArgs.end()
              && findParameter(currentContextPtr_->resolvedGlobalParams_.begin(), currentContextPtr_->resolvedGlobalParams_.end(), *it) == NULL)
          {
            Report::UserError0() << functionNameAndArgs << " contains unknown parameter " << (*it);
          }
        }
      }
    }
  }

  return true;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::setContext
// Purpose        : Set the current context to that corresponding to the given
//                  subcircuit name. Save the previous context on the stack for
//                  later retrieval. If the given subcircuit name is not found,
//                  declare an error and abort.
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 02/07/2003
//----------------------------------------------------------------------------
bool CircuitContext::setContext(
  const std::string &                   subcircuitName,
  const std::string &                   subcircuitPrefixIn,
  const std::vector<std::string> &      instanceNodes,
  CircuitContext *                      previousContext ) const
{
  bool success = false;

  unordered_map< std::string, CircuitContext* >::iterator ccIter =
    currentContextPtr_->circuitContextTable_.find( subcircuitName );

  if ( ccIter != currentContextPtr_->circuitContextTable_.end() )
  {
    // Save current context and reset the current context pointer.
    if (previousContext == NULL)
      contextList_.push_front(currentContextPtr_);
    else
      contextList_.push_front(previousContext);

    currentContextPtr_ = ccIter->second;

    // If subcircuitPrefix and instanceNodes were given, set the prefix
    // in the new context, and build the node map.
    currentContextPtr_->nodeMap_.clear();
    currentContextPtr_->subcircuitPrefix_ = subcircuitPrefixIn;

    if (!instanceNodes.empty())
    {
      std::vector<std::string>::const_iterator nodeIter = instanceNodes.begin();
      for (unsigned int i = 0; i < currentContextPtr_->nodeList_.size() && nodeIter != instanceNodes.end(); ++i, ++nodeIter)
      {
        if (currentContextPtr_->nodeMap_.find(currentContextPtr_->nodeList_[i]) !=
            currentContextPtr_->nodeMap_.end())
        {
          if (currentContextPtr_->nodeMap_[currentContextPtr_->nodeList_[i]] !=
              *nodeIter)
          {
            Report::UserError0()
              <<  "Duplicate nodes in .subckt " << subcircuitName << " point to different nodes in X line invocation";
            return false;
          }
        }
        else
        {
          if ((currentContextPtr_->nodeList_[i].size() >= 2 &&
               currentContextPtr_->nodeList_[i].substr(0,2) == "$G") ||
              globalNode(currentContextPtr_->nodeList_[i]))
          {
            if (currentContextPtr_->nodeList_[i] != *nodeIter)
            {
              Report::UserError0()
                << "Global node in subcircuit invocation must match same name in .subckt";
              return false;
            }
          }
          currentContextPtr_->nodeMap_[currentContextPtr_->nodeList_[i]] = *nodeIter;
        }
      }
    }

    return true;
  }

  // The context with the given name was not found in the current context,
  // recursively search parent contexts.
  if (currentContextPtr_->parentContextPtr_ != NULL)
  {
    if (previousContext == NULL)
    {
      previousContext = currentContextPtr_;
    }

    currentContextPtr_ = currentContextPtr_->parentContextPtr_;
    success = setContext(subcircuitName, subcircuitPrefixIn, instanceNodes, previousContext);
  }

  // if (!success)
  // {
  //   Report::UserError0() << "Global node in subcircuit invocation must match same name in .subckt";
  // }

  return success;
}

void CircuitContext::setContext(CircuitContext* context) const
{
  contextList_.push_front(currentContextPtr_);
  currentContextPtr_ = context;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::restorePreviousContext
// Purpose        : Reset the context the context prior to the last invocation
//                  of setContext.
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 02/07/2003
//----------------------------------------------------------------------------
void CircuitContext::restorePreviousContext() const
{
  // Restore the previous context from the list of contexts unless
  // the list is empty.
  if (!contextList_.empty())
  {
    currentContextPtr_ = contextList_.front();
    contextList_.pop_front();
  }
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::globalNode
// Purpose        : 
// Special Notes  :
// Scope          :
// Creator        : Dave Shirley, PSSI
// Creation Date  : 10/20/2009
//----------------------------------------------------------------------------
bool CircuitContext::globalNode (const std::string &nodeName) const
{
  bool stat;

  if (!(currentContextPtr_->parentContextPtr_))
  {
    if (globalNodes_.find(nodeName) == globalNodes_.end())
      return false;
    else
      return true;
  }

  setContext(currentContextPtr_->parentContextPtr_);
  stat = globalNode ( nodeName );
  restorePreviousContext();

  return stat;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::resolveParameter
// Purpose        : Parameter whose value may be an expression that must be
//                  resolved. If the input parameter has an expression value,
//                  replace the parameters and functions in the
//                  expression with their actual values.
// Special Notes  :
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 02/10/2003
//----------------------------------------------------------------------------
bool CircuitContext::resolveParameter(Util::Param& parameter,
                                      std::vector<std::string> exceptionStrings)
{
  if (hasExpressionTag(parameter) || parameter.hasExpressionValue() )
  {
    if (DEBUG_IO)
    {
      Xyce::dout() << "CircuitContext::resolveParameter parameter " << parameter.uTag()
                   << " has expression value ";
    }

    // Extract the expression from the parameter value by stripping off
    // the enclosing braces.  Only strip if it's there!
    std::string expressionString;
    if (parameter.stringValue()[0] == '{')
    {
      expressionString = parameter.stringValue().substr(1, parameter.stringValue().size()-2);
    }
    else
    {
      expressionString = parameter.stringValue();
    }

    if (DEBUG_IO)
    {
      Xyce::dout() << expressionString << std::endl;
    }

    // Parse the expression:
    Util::Expression expression(expressionString);

    if (!expression.parsed())
    {
      return false;
    }

    // Resolve the strings in the expression. Unresolved strings
    // may be parameters defined in a .PARAM statement or global
    // parameters defined in .GLOBAL_PARAM statement or may
    // be due to function arguments if the expression is the
    // body of a function defined in a .FUNC statement.
    bool stringsResolved = resolveStrings(expression, exceptionStrings);

    // Resolve functions in the expression.
    bool functionsResolved = resolveFunctions(expression);

    // resolve variables in the function body
    if ( expression.get_num(XEXP_STRING) > 0 )
    {
      resolveStrings(expression, exceptionStrings);
    }

    if (expression.get_num( XEXP_LEAD ) > 0)
    {
      parameter.setVal(expression);
      return false;
    }

    if (stringsResolved && functionsResolved)
    {
      // Check the expression for nodes or instance (probably a B-source
      // expression, in which case the parameter value should be
      // expressionString. Also check for "specials", the only special
      // allowed is "time" for time dependent parameters.
      std::vector<std::string> nodes, instances, leads, variables, specials, nodecomps;
      expression.get_names(XEXP_NODE, nodes);
      expression.get_names(XEXP_INSTANCE, instances);
      expression.get_names(XEXP_LEAD, leads);
      expression.get_names(XEXP_VARIABLE, variables);
      expression.get_names(XEXP_SPECIAL, specials);
      expression.get_names(XEXP_NODAL_COMPUTATION, nodecomps);

      if (!nodes.empty() || !instances.empty() || !leads.empty() ||
          !variables.empty() || !specials.empty() || !nodecomps.empty())
      {
        if (DEBUG_IO)
        {
          Xyce::dout() << "CircuitContext::resolveParameter:  nodes, instances, leads, variables or specials not empty. " << std::endl;
          if (!nodes.empty())
          {
            Xyce::dout() << " Nodes: " << std::endl;
            for (int foo=0; foo<nodes.size(); ++foo)
              Xyce::dout() << foo << " : " << nodes[foo] << std::endl;
          }
          if (!instances.empty())
          {
            Xyce::dout() << " Instances: " << std::endl;
            for (int foo=0; foo<instances.size(); ++foo)
              Xyce::dout() << foo << " : " << instances[foo] << std::endl;
          }
          if (!leads.empty())
          {
            Xyce::dout() << " Leads: " << std::endl;
            for (int foo=0; foo<leads.size(); ++foo)
              Xyce::dout() << foo << " : " << leads[foo] << std::endl;
          }
          if (!variables.empty())
          {
            Xyce::dout() << " Variables: " << std::endl;
            for (int foo=0; foo<variables.size(); ++foo)
              Xyce::dout() << foo << " : " << variables[foo] << std::endl;
          }
          if (!specials.empty())
          {
            Xyce::dout() << " Specials: " << std::endl;
            for (int foo=0; foo<specials.size(); ++foo)
              Xyce::dout() << foo << " : " << specials[foo] << std::endl;
          }
        }

        parameter.setVal(expression);
        if (DEBUG_IO)
        {
          Xyce::dout() << "CircuitContext::resolveParameter: After all expression handling, get_expression returns "
                       << expression.get_expression() << std::endl;
          Xyce::dout() << " after setting the parameter " << parameter.uTag() << ", its type is " << parameter.getType() << std::endl;
          Xyce::dout() << " and its value is ";
          switch (parameter.getType()) {
            case Xyce::Util::STR:
              Xyce::dout() << parameter.stringValue();
              break;
            case Xyce::Util::DBLE:
              Xyce::dout() << parameter.getImmutableValue<double>();
              break;
            case Xyce::Util::EXPR:
              Xyce::dout() << parameter.getValue<Util::Expression>().get_expression();
              break;
            default:
              Xyce::dout() << parameter.stringValue();
          }
          Xyce::dout() << std::endl;
        }
      }
      else
      {
        if (exceptionStrings.empty())
        {
          // Reset the parameter value to the value of the expression.
          double value(0.0);
          expression.evaluateFunction ( value );
          parameter.setVal( value );
          // we have resolved the context so set it and the constant value to 
          // make later look ups easier.
          // parameter.addOp(Util::CONSTANT, new IO::ConstantOp(parameter.tag(), value));
          if (DEBUG_IO)
          {
            Xyce::dout() << " CircuitContext::resolveParameter --  Resetting parameter value from " << expressionString << " to " << value << " because exceptionStrings empty." << std::endl;
          }
        }
        else
        {
          parameter.setVal( expression );
        }
      }
    }
    else
    {
      // Reset the parameter value to the value of the expression with
      // as much resolution as could be achieved.
      parameter.setVal(expression);
    }

    if (DEBUG_IO)
    {
      Xyce::dout() << "CircuitContext::resolveParameter: right before returns "
                   << std::endl;
      Xyce::dout() << " after setting the parameter " << parameter.uTag() << ", its type is " << parameter.getType() << std::endl;
      Xyce::dout() << " and its value is ";
      switch (parameter.getType()) {
        case Xyce::Util::STR:
          Xyce::dout() << parameter.stringValue();
          break;
        case Xyce::Util::DBLE:
          Xyce::dout() << parameter.getImmutableValue<double>();
          break;
        case Xyce::Util::EXPR:
          Xyce::dout() << parameter.getValue<Util::Expression>().get_expression();
          break;
        default:
          Xyce::dout() << parameter.stringValue();
      }
      Xyce::dout() << std::endl;
    }

    return stringsResolved && functionsResolved;

  }
  // Handle quoted parameters e.g. "filename" (which get turned into
  // TABLEs)
  resolveQuote(parameter);

  // The parameter does not have an expression value and does not need
  // resolving.
  return true;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::resolveStrings
// Purpose        : Determine if expression has any unresolved strings
//                  and resolve appropriately. Return true if all strings are
//                  resolved otherwise return false.
// Special Notes  :
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 02/11/2003
//----------------------------------------------------------------------------
bool CircuitContext::resolveStrings( Util::Expression & expression,
                                     std::vector<std::string> exceptionStrings)
{
  // Strings in the expression must be previously resolved parameters
  // that appear in paramList else there is an error.
  bool unresolvedStrings = false;
  if ( expression.get_num(XEXP_STRING) > 0 )
  {
    // Get the list of strings in the expression.
    std::vector<std::string> strings;
    expression.get_names(XEXP_STRING, strings);

    // If the expression is resolvable, each string in the current expression
    // must appear as a resolved parameter in netlistParameters. Get the value
    // if it appears there.
    ExtendedString parameterName("");
    int numStrings = strings.size();

    if (DEBUG_IO)
    {
      Xyce::dout() <<" CircuitContext::resolveStrings numStrings = " << numStrings << std::endl;
    }

    for (int i = 0; i < numStrings; ++i)
    {
      if (DEBUG_IO)
      {
        Xyce::dout() <<" CircuitContext::resolveStrings resolving " << strings[i] << std::endl;
      }

      // Skip the current string if it is in exception strings. This prevents
      // a function argument from being improperly resolved when there is
      // a parameter in a .param statement with the same name as the function
      // argument.
      if (!exceptionStrings.empty())
      {
        std::vector<std::string>::iterator stringIter = find(exceptionStrings.begin(),
                                                             exceptionStrings.end(),
                                                             strings[i]);
        if (stringIter != exceptionStrings.end())
        {
          if (DEBUG_IO)
          {
            Xyce::dout() <<" CircuitContext::resolveStrings skipping exception string " << strings[i] << std::endl;
          }

          continue;
        }
      }

      // Look for the string in netlistParameters.
      parameterName = strings[i];
      parameterName.toUpper();

      Util::Param expressionParameter(parameterName, "");
      bool parameterFound = getResolvedParameter(expressionParameter);
      if (parameterFound)
      {
        if (DEBUG_IO)
        {
          Xyce::dout() <<" CircuitContext::resolveStrings string " << strings[i] << " is a resolved parameter " << expressionParameter.uTag() << " with type "
                       << expressionParameter.getType() << " and value ";
          switch (expressionParameter.getType()) {
            case Xyce::Util::STR:
              Xyce::dout() << expressionParameter.stringValue();
              break;
            case Xyce::Util::DBLE:
              Xyce::dout() << expressionParameter.getImmutableValue<double>();
              break;
            case Xyce::Util::EXPR:
              Xyce::dout() << "EXPR("<<expressionParameter.getValue<Util::Expression>().get_expression()<< ")";
              break;
            default:
              Xyce::dout() << expressionParameter.stringValue();
          }
          Xyce::dout() << std::endl;
        }

        if ( expressionParameter.getType() == Xyce::Util::STR ||
             expressionParameter.getType() == Xyce::Util::DBLE )
        {
          if (!expression.make_constant(strings[i], expressionParameter.getImmutableValue<double>()))
          {
            Report::UserWarning0() << "Problem converting parameter " << parameterName << " to its value.";
          }
        }
        else if (expressionParameter.getType() == Xyce::Util::EXPR)
        {
          std::string expressionString=expression.get_expression();
          if (expression.replace_var(strings[i], expressionParameter.getValue<Util::Expression>()) != 0)
          {
            Report::UserWarning0() << "Problem inserting expression " << expressionParameter.getValue<Util::Expression>().get_expression()
                                   << " as substitute for " << parameterName << " in expression " << expressionString;
          }
        }
      }
      else
      {
        parameterFound = getResolvedGlobalParameter(expressionParameter);
        if (DEBUG_IO)
        {
          Xyce::dout() << "CircuitContext::resolveStrings attempting to resolve "
                       <<  " parameter " << expressionParameter.uTag() << std::endl;
          if (parameterFound)
          {
            Xyce::dout() << "Found it." << std::endl;
          }
          else
          {
            Xyce::dout() << " Did not find a resolved global parameter named "
                         << expressionParameter.uTag()	<< std::endl;
          }
        }

        if (parameterFound)
        {
          if (!expression.make_var(strings[i]))
          {
            Report::UserWarning0() << "Problem converting parameter " << parameterName <<" to its value";
          }
        }
        else
        {
          if (compare_nocase(strings[i].c_str(), "gmin") == 0
              || compare_nocase(strings[i].c_str(), "vt") == 0)  // This needs to be generalized, but for now just GMIN and VT
          {
            Util::Op::Operator *op = opBuilderManager_.createOp(strings[i]);
            if (op)
              expression.replace_var(strings[i], op);
          }
          else
          {
            if (Util::isBool(strings[i]))
            {
              bool stat = false;
              if (Util::Bval(strings[i]))
                stat = expression.make_constant(strings[i], static_cast<double>(1));
              else
                stat = expression.make_constant(strings[i], static_cast<double>(0));
              if (!stat)
              {
                Report::UserWarning0() << "Problem converting parameter " << parameterName << " to its value";
              }
            }
            else
            {
              unresolvedStrings = true;
            }
          }
        }
      }
    }
  }

  return !unresolvedStrings;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::resolveFunctions
// Purpose        : Determine if expression has any unresolved functions
//                  and resolve appropriately. Return true if all functions
//                  are resolved otherwise return false.
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 02/11/2003
//----------------------------------------------------------------------------
bool CircuitContext::resolveFunctions(Util::Expression & expression)
{
  // Functions in the expression must be previously defined functions
  // that appear in resolvedFunctions_ else there is an error.
  bool unresolvedFunctions = false;
  if ( expression.get_num(XEXP_FUNCTION) > 0)
  {
    // Get the list of strings in the expression.
    std::vector<std::string> functions;
    expression.get_names(XEXP_FUNCTION, functions);

    // If the expression is resolvable, each function in the current expression
    // must appear as a defined function in resolvedFunctions_.
    int numFunctions = functions.size();
    for (int i = 0; i < numFunctions; ++i)
    {
      // Look for the function in resolvedFunctions_.
      Util::Param functionParameter(functions[i], "");
      bool functionfound = getResolvedFunction(functionParameter);
      if (functionfound)
      {
        std::string functionPrototype(functionParameter.tag());
        std::string functionBody(functionParameter.stringValue());

        // The function prototype is defined here as a string whose
        // value is the  function name together with its parenthese
        // enclosed comma separated argument list. To resolve a
        // function, create an expression from the function prototype
        // and get its ordered list of arguments via get_names, then
        // create an expression from the function definition and
        // order its names from that list. Finally, replace the
        // function in the expression to be resolved.
        Util::Expression prototypeExression(functionPrototype);
        std::vector<std::string> arguments;
        prototypeExression.get_names(XEXP_STRING, arguments);
        Util::Expression functionExpression(functionBody);
        functionExpression.order_names(arguments);

        if (expression.replace_func(functions[i], functionExpression,
                                    static_cast<int>(arguments.size())) < 0)
        {
          Report::UserError0() << "Wrong number of arguments for user defined function " << functionPrototype << " in expression " << expression.get_expression();
        }

        // Set the expression value.
        if (DEBUG_IO)
        {
          Xyce::dout() << "CircuitContext::resolveFunctions: After all expression handling, get_expression returns "
                       << expression.get_expression() << std::endl;
        }
      }
      else
      {
        unresolvedFunctions = true;
      }
    }
  }

  return !unresolvedFunctions;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::getResolvedParameter
// Purpose        : Look for a parameter (set in the netlist by .param) 
//                  in the current context's set of resolved parameters. 
//                  Check the current context and recurively check parent 
//                  contexts.
// Special Notes  :
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 02/11/2003
//----------------------------------------------------------------------------
bool CircuitContext::getResolvedParameter(Util::Param & parameter)
{
  bool success = false;

  const Util::Param* parameterPtr = findParameter(currentContextPtr_->resolvedParams_.begin(), currentContextPtr_->resolvedParams_.end(), parameter.tag());
  if (parameterPtr != NULL)
  {
    // Found a parameter with given name, set the value
    // of parameter and return.
    parameter.setVal(*parameterPtr);
    success = true;
  }
  else if (currentContextPtr_->parentContextPtr_ != NULL)
  {
    setContext(currentContextPtr_->parentContextPtr_);
    success = getResolvedParameter(parameter);
    restorePreviousContext();
  }

  return success;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::getResolvedGlobalParameter
// Purpose        : Look for a parameter (set in the netlist by .global_param)
//                  in the current context's set of resolved parameters. 
//                  Check the current context and recurively check parent 
//                  contexts.
// Special Notes  :
// Scope          : public
// Creator        : Dave Shirley, PSSI
// Creation Date  : 11/17/2005
//----------------------------------------------------------------------------
bool CircuitContext::getResolvedGlobalParameter(Util::Param & parameter)
{
  bool success = false;

  const Util::Param* parameterPtr = findParameter(currentContextPtr_->resolvedGlobalParams_.begin(), currentContextPtr_->resolvedGlobalParams_.end(), parameter.tag());
  if (parameterPtr != NULL)
  {
    // Found a parameter with given name, set the value
    // of parameter and return.
    parameter.setVal(*parameterPtr);
    success = true;
  }
  else if (currentContextPtr_->parentContextPtr_ != NULL)
  {
    setContext(currentContextPtr_->parentContextPtr_);
    success = getResolvedGlobalParameter(parameter);
    restorePreviousContext();
  }

  return success;
}

//-----------------------------------------------------------------------------
// Function      : CircuitContext::getResolvedFunction
// Purpose       : Look for a function (set in the netlist by .func) 
//                 in resolvedFunctions_.  Check current context and recursively 
//                 check each parent context.
// Special Notes :
// Scope         : public
// Creator       : Lon Waters, SNL
// Creation Date : 12/27/2001
//-----------------------------------------------------------------------------
bool CircuitContext::getResolvedFunction(Util::Param & parameter)
{
  bool success = false;

  std::string functionToFind(parameter.uTag());

  if (currentContextPtr_->resolvedFunctions_.find(functionToFind) !=
      currentContextPtr_->resolvedFunctions_.end())
  {
    parameter = currentContextPtr_->resolvedFunctions_[functionToFind];
    success = true;
  }
  else if (currentContextPtr_->parentContextPtr_ != NULL)
  {
    setContext(currentContextPtr_->parentContextPtr_);
    success = getResolvedFunction(parameter);
    restorePreviousContext();
  }

  return success;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::findModel
// Purpose        : Search the models in the current context for the model of
//                  the given name. If it is not found, recursively
//                  search each parent context. Return a pointer to
//                  the parameterBlock for the model if it is found,
//                  otherwise return NULL. Also, if the model is found,
//                  construct the appropriate model prefix.
// Special Notes  :
// Scope          :
// Creator        : Lon Waters
// Creation Date  : 02/12/2003
//----------------------------------------------------------------------------
bool CircuitContext::findModel(
  const std::string &           modelName,
  ParameterBlock* &             modelPtr,
  std::string &                 modelPrefix) const
{
  bool success = false;

  ModelMap::const_iterator modelIter = currentContextPtr_->models_.find(modelName);
  if (modelIter != currentContextPtr_->models_.end())
  {
    modelPtr = modelIter->second;
    if (modelPtr->hasExpressionValuedParams())
    {
      modelPrefix = currentContextPtr_->getPrefix();
    }
    else
    {
      std::string prefix = currentContextPtr_->getCurrentContextName();
      if (prefix == "")
        modelPrefix = "";
      else
        modelPrefix = prefix + ":";
    }

    return true;
  }

  // The model was not found in the current circuit's model list,
  // recursively search the parent circuit's model list.
  if ( currentContextPtr_->parentContextPtr_ != NULL )
  {
    setContext(currentContextPtr_->parentContextPtr_);
    success = findModel( modelName, modelPtr, modelPrefix );
    restorePreviousContext();
  }

  return success;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::findModel
// Purpose        : Overloaded version of findModel for cases when the model
//                  prefix is not needed.
// Special Notes  :
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 02/13/2003
//----------------------------------------------------------------------------
bool CircuitContext::findModel(
  const std::string &           modelName,
  ParameterBlock* &        modelPtr) const
{
  bool success;
  std::string temp;
  success = findModel(modelName, modelPtr, temp);

  return success;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::hasSubcircuitParams
// Purpose        : Check whether a subcircuit context is dependent on
//                  subcircuit parameters. These are parameters on the
//                  .subckt line identified by "params:" keyword. The result
//                  should be true if either the current subcircuit context
//                  or any subcircuit context in the hierarchy containing the
//                  current subcircuit context has subcircuit parameters.
// Special Notes  :
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 07/15/2003
//----------------------------------------------------------------------------
bool CircuitContext::hasSubcircuitParams()
{
  bool foundSubcircuitParams = false;

  if (!currentContextPtr_->subcircuitParameters_.empty())
  {
    foundSubcircuitParams = true;
  }
  else if (currentContextPtr_->parentContextPtr_ != NULL)
  {
    setContext(currentContextPtr_->parentContextPtr_);
    foundSubcircuitParams = hasSubcircuitParams();
    restorePreviousContext();
  }

  return foundSubcircuitParams;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::getTotalDeviceCount
// Purpose        : Calculate the total number of devices starting at current
//                  context and including all subcircuit instances.
// Special Notes  : This function also accumulates the used context list and 
//                  device type count map.
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 02/13/2003
//----------------------------------------------------------------------------
int CircuitContext::getTotalDeviceCount()
{
  // Get device count for the current context.
  int count = currentContextPtr_->deviceCount_;
  int linearCount = currentContextPtr_->linearDeviceCount_;

  // Determine the device count associated with each subcircuit instance.
  if (!currentContextPtr_->deviceCountDone_)
  {
    DeviceCountMap tmpDCMap = currentContextPtr_->localDeviceCountMap_;

    std::vector<std::string>::iterator instanceIter;
    std::vector<std::string>::iterator start = currentContextPtr_->instanceList_.begin();
    std::vector<std::string>::iterator end = currentContextPtr_->instanceList_.end();

    for (instanceIter = start; instanceIter != end; ++instanceIter)
    {
      bool result = setContext(*instanceIter);
      if (result)
      {
        // MUST perform getTotalDeviceCount before getTotalLinearDeviceCount.
        count += getTotalDeviceCount();
        linearCount += currentContextPtr_->getTotalLinearDeviceCount();

        // Add in devices to map for counting.
        DeviceCountMap::iterator lDCMb = currentContextPtr_->localDeviceCountMap_.begin();
        DeviceCountMap::iterator lDCMe = currentContextPtr_->localDeviceCountMap_.end();
        DeviceCountMap::iterator lDCMit = lDCMb;
        for ( ; lDCMit != lDCMe; ++lDCMit) 
        {
          if ( tmpDCMap[lDCMit->first] )
          {
            tmpDCMap[lDCMit->first] += lDCMit->second;
          }
          else
          {
            tmpDCMap[lDCMit->first] = lDCMit->second;
          }
        }
      }
      
      restorePreviousContext();

      if (result)
      {
        // Just push back the used subcircuit, multiple entries will be filtered out later.
        (currentContextPtr_->usedContextList_).push_back( *instanceIter );
      }  
    }

    // Propagate this list up to the parent, eventually the top circuit context will have
    // all the names of the used subcircuits.
    if (currentContextPtr_->parentContextPtr_ != NULL)
    {
      std::vector<std::string>::iterator ucIter;
      std::vector<std::string>::iterator uc_start = (currentContextPtr_->usedContextList_).begin();
      std::vector<std::string>::iterator uc_end = (currentContextPtr_->usedContextList_).end();

      for (ucIter = uc_start; ucIter!=uc_end; ++ucIter)
      {
        // Just push back the used subcircuit, multiple entries will be filtered out later.
        ((currentContextPtr_->parentContextPtr_)->usedContextList_).push_back( *ucIter );
      }
    }

    currentContextPtr_->deviceCount_ = count;
    currentContextPtr_->linearDeviceCount_ = linearCount;
    currentContextPtr_->localDeviceCountMap_ = tmpDCMap;
    currentContextPtr_->deviceCountDone_ = true;
  }
    
  return count;
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::pruneContexts
// Purpose        : Remove any context that is not specified in the passed in 
//                : list.  These were obtained through the computation of the
//                : total number of devices.
// Special Notes  :
// Scope          : public
// Creator        : Heidi Thornquist
// Creation Date  : 06/20/2014
//----------------------------------------------------------------------------
void CircuitContext::pruneContexts( std::vector<std::string>& keepContexts )
{
  unordered_map< std::string, CircuitContext * >::iterator itsc2, itsc = circuitContextTable_.begin();

  // Delete each context object in the table that is not on the provided list.
  while ( itsc != circuitContextTable_.end() )
  {
    itsc->second->pruneContexts( keepContexts );

    // Safely remove any contexts that were not used.
    if ( !std::binary_search( keepContexts.begin(), keepContexts.end(), itsc->first ) )
    {
      // Delete the pointer to the object and then remove it from the map.
      // Make sure that itsc is a valid iterator after the entry is removed from the map.
      itsc2 = itsc;
      itsc++;
      delete itsc2->second;
      circuitContextTable_.erase( itsc2 );
    }
    else
    {
      itsc++;
    }
  }
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::totalMutualInductanceCount
// Purpose        : Calculate the total number of MIs starting at current
//                  context and including all subcircuit instances.
// Special Notes  :
// Scope          : public
// Creator        : Rob Hoekstra
// Creation Date  : 08/27/2004
//----------------------------------------------------------------------------
int CircuitContext::totalMutualInductanceCount()
{
  // Get device count for the current context.
  int count = currentContextPtr_->mutualInductances_.size();

  // Determine the device count associated with each subcircuit instance.
  std::vector<std::string>::iterator instanceIter;
  std::vector<std::string>::iterator start = currentContextPtr_->instanceList_.begin();
  std::vector<std::string>::iterator end = currentContextPtr_->instanceList_.end();

  for (instanceIter = start; instanceIter != end; ++instanceIter)
  {
    bool result = setContext(*instanceIter);
    if (result)
    {
      count += totalMutualInductanceCount();
    }

    restorePreviousContext();
  }

  return count;
}


//-----------------------------------------------------------------------------
// Function      : CircuitContext::MutualInductance::MutualInductance
// Purpose       : Constructor
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra
// Creation Date : 8/27/04
//-----------------------------------------------------------------------------
CircuitContext::MutualInductance::MutualInductance( DeviceBlock & device )
{
  int numParameters = device.getNumberOfInstanceParameters();
  Device::Param parameter;
  bool first = true;
  for ( int i = 0; i < numParameters; ++i )
  {
    parameter = device.getInstanceParameter(i);

    if ( parameter.tag() != "COUPLING" )
    {
      std::string inductorName (parameter.uTag());

      if( first )
      {
        firstInductor = inductorName;
        first = false;
      }

      inductors[inductorName] = 0.0;
    }
    else
      coupling = parameter.stringValue();
  }

  model = device.getModelName();

  name = device.getInstanceName().getEncodedName();
  sharedKey = 0;

}

//-----------------------------------------------------------------------------
// Function      : CircuitContext::bundleMIs
// Purpose       : Convert all MIs into tokenized device lines
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void CircuitContext::bundleMIs()
{
  std::string type;
  unsigned int i, j, mTableSize, mTableRowSize;
  StringToken field;
  std::vector< std::string > tmpInductorList, tmpCouplingList;
  TokenVector tmpLine;

  // retrieve number of K lines in this (sub)circuit
  mTableSize = currentContextPtr_->allIndexedMIs_.size();

  for( i = 0; i < mTableSize; ++i )
  {
    // reset lists, tables and indices
    tmpInductorList.clear();
    tmpCouplingList.clear();

    // reset tmp vars
    std::string tmpName("");
    std::string tmpModel("");
    tmpLine.clear();

    std::vector< std::set< std::string > > & tmpTable =
      currentContextPtr_->getSharedInductorTable();

    mTableRowSize = currentContextPtr_->allIndexedMIs_[i].size();

    for( j = 0; j < mTableRowSize; ++j )
    {
      // select current mutual inductance
      MutualInductance & mutind =
        currentContextPtr_->mutualInductances_[
          currentContextPtr_->allIndexedMIs_[i][j]];

      // collect name segments
      if ( tmpName == "" )
        tmpName = mutind.name;
      else
        tmpName += "_" + mutind.name;

      // flag nonlinear coupling
      if( mutind.model != "" )
      {
        tmpModel = mutind.model;
        type = "N";
      }

      // else linear coupling
      else
      {
        type = "L";
      }

      // assemble inductor list
      std::map< std::string, double >::iterator mIter = mutind.inductors.begin();
      for( ; mIter != mutind.inductors.end(); ++mIter )
      {
        // add inductor data if not already present
        if( ( type == "N" ) ||
            ( tmpTable[mutind.sharedKey].find( (*mIter).first ) !=
              tmpTable[mutind.sharedKey].end() ) )
        {
          // retrieve the inductor name
          tmpInductorList.push_back( (*mIter).first );

          //Do some error-checking to make sure that the inductor has been
          //defined elsewhere!

          std::map<std::string,std::vector<std::string> >::iterator inducIter;
          inducIter = mutind.terminals.find((*mIter).first);
          if (inducIter == mutind.terminals.end())
          {
            Report::UserError0() << "Undefined inductor " << (*mIter).first << " in mutual inductor " << mutind.name << " definition.";
          }
          else
          {
            // retrieve the inductor nodes
            tmpInductorList.push_back( mutind.terminals[(*mIter).first][0] );
            tmpInductorList.push_back( mutind.terminals[(*mIter).first][1] );
          }

          // retrieve the inductance value
          std::stringstream cnvtr;
          cnvtr << (*mIter).second;
          tmpInductorList.push_back( cnvtr.str() );

          // remove inductor from list to prevent duplicates
          if( type == "L" )
          {
            tmpTable[mutind.sharedKey].erase( (*mIter).first );
          }

          /*
          // manually check for nonlin/lin inductor overlap & exit
          else
          {
          // FIXME
          }
          */
        }

        // append to coupling list; names are a bit redundant, but easy to track
        tmpCouplingList.push_back( (*mIter).first );
      }

      // append coupling value to string
      std::stringstream cnvtr;
      cnvtr << mutind.coupling;
      tmpCouplingList.push_back( cnvtr.str() );
    }

    // fully contruct line from components and type
    if( tmpName != "" )
    {
      // append type
      field.string_ = "YMI" + type;
      tmpLine.push_back( field );

      // append name
      field.string_ = tmpName;
      tmpLine.push_back( field );

      // append list of (inductor, terminals, inductance)+
      for( j = 0; j < tmpInductorList.size(); ++j )
      {
        field.string_ = tmpInductorList[j];
        tmpLine.push_back( field );
      }

      // append coupling list
      for( j = 0; j < tmpCouplingList.size(); ++j )
      {
        field.string_ = tmpCouplingList[j];
        tmpLine.push_back( field );
      }

      // append model name
      if( type == "N" )
      {
        field.string_ = tmpModel;
        tmpLine.push_back( field );
      }

      // store in current context
      currentContextPtr_->kLines_.push_back( tmpLine );
    }
  }
}


//-----------------------------------------------------------------------------
// Function      : CircuitContext::getMILine
// Purpose       : Retrieve one tokenized device line
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
TokenVector &
CircuitContext::getMILine( int i )
{
  if( ( i < 0 ) || ( i > (int)currentContextPtr_->kLines_.size() ) )
  {
    // bounds checking error exit
    Report::UserError() << "Request exceeds number of mutual inductances in this subcircuit";
  }

  return currentContextPtr_->kLines_[i];
}

//----------------------------------------------------------------------------
// Function       : CircuitContext::augmentTotalDeviceCount
// Purpose        : Augment total device count after we process k-lines
//
// Special Notes  :
// Scope          : public
// Creator        : Keith Santarelli
// Creation Date  : 09/22/08
//----------------------------------------------------------------------------
void CircuitContext::augmentTotalDeviceCount(int kLineCount,
                                             int coupledICount,
                                             int YDeviceCount)
{
  // Get device count for the current context.
  int count = currentContextPtr_->deviceCount_;
  count += YDeviceCount - kLineCount - coupledICount;

  // Remove the inductors from the count
  if (coupledICount)
  {
    currentContextPtr_->localDeviceCountMap_["L"] -= coupledICount;
  }

  if (count < 0)
  {
    Report::DevelFatal() << "Augmented number of devices is less than 0.";
  }
  else
  {
    currentContextPtr_->deviceCount_ = count;
  }
}

} // namespace IO

//-----------------------------------------------------------------------------
// Function      : CircuitContext::packedByteCount
// Purpose       : Counts bytes needed to pack block.
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
template<>
int Pack<IO::CircuitContext>::packedByteCount(
  const IO::CircuitContext &    circuit_context)
{
  int byteCount = 0;
  int size, i;

  // count name
  byteCount += sizeof( int );
  byteCount += circuit_context.name_.length();

  // count device count
  byteCount += sizeof( int );

  // count models
  IO::ModelMap::const_iterator it_spbM = circuit_context.models_.begin();
  IO::ModelMap::const_iterator it_speM = circuit_context.models_.end();
  byteCount += sizeof( int );
  for( ; it_spbM != it_speM; ++it_spbM )
  {
    // ---- count key
    byteCount += sizeof( int );
    byteCount += it_spbM->first.length();

    // ---- count data
    byteCount += Xyce::packedByteCount(*it_spbM->second);
  }

  // count unresolved functions
  size = circuit_context.unresolvedFunctions_.size();
  byteCount += sizeof( int );
  for( i = 0; i < size; ++i )
  {
    byteCount += Xyce::packedByteCount(circuit_context.unresolvedFunctions_[i]);
  }

  // count instance list
  std::vector< std::string >::const_iterator it_stbL = circuit_context.instanceList_.begin();
  std::vector< std::string >::const_iterator it_steL = circuit_context.instanceList_.end();
  byteCount += sizeof( int );
  for( ; it_stbL != it_steL; ++it_stbL )
  {
    byteCount += sizeof( int );
    byteCount += it_stbL->length();
  }

  // count node list
  size = circuit_context.nodeList_.size();
  byteCount += sizeof( int );
  for( i = 0; i < size; ++i )
  {
    byteCount += sizeof( int );
    byteCount += circuit_context.nodeList_[ i ].length();
  }

  // count subcircuit parameters
  size = circuit_context.subcircuitParameters_.size();
  byteCount += sizeof( int );
  for (Util::ParamList::const_iterator it = circuit_context.subcircuitParameters_.begin(), end = circuit_context.subcircuitParameters_.end(); it != end; ++it)
  {
    byteCount += Pack<Util::Param>::packedByteCount(*it);
  }

  // count unresolved params
  size = circuit_context.unresolvedParams_.size();
  byteCount += sizeof( int );
  for (Util::ParamList::const_iterator it = circuit_context.unresolvedParams_.begin(), end = circuit_context.unresolvedParams_.end(); it != end; ++it)
  {
    byteCount += Pack<Util::Param>::packedByteCount(*it);
  }

  // pack global node names
  std::set<std::string>::const_iterator globalNodes_i, globalNodes_end;
  globalNodes_i = circuit_context.globalNodes_.begin();
  globalNodes_end = circuit_context.globalNodes_.end();
  byteCount += sizeof( int );
  for ( ; globalNodes_i != globalNodes_end ; ++globalNodes_i )
  {
    byteCount += sizeof( int );
    byteCount += globalNodes_i->length();
  }

  // count global params
  size = circuit_context.unresolvedGlobalParams_.size();
  byteCount += sizeof( int );
  for (Util::ParamList::const_iterator it = circuit_context.unresolvedGlobalParams_.begin(), end = circuit_context.unresolvedGlobalParams_.end(); it != end; ++it)
  {
    byteCount += Pack<Util::Param>::packedByteCount(*it);
  }

  //count mutual inductances
  size = circuit_context.mutualInductances_.size();
  byteCount += sizeof( int );
  for( i = 0; i < size; ++i )
  {
    byteCount += Xyce::packedByteCount(circuit_context.mutualInductances_[i]);
  }

  // count subcircuit contexts
  size = circuit_context.circuitContextTable_.size();
  byteCount += sizeof( int );
  unordered_map< std::string, IO::CircuitContext * >::const_iterator itsc;
  unordered_map< std::string, IO::CircuitContext * >::const_iterator itsc_end = circuit_context.circuitContextTable_.end();
  for ( itsc = circuit_context.circuitContextTable_.begin(); itsc != itsc_end; ++itsc )
  {
    byteCount += sizeof( int );
    byteCount += itsc->first.length();
    byteCount += Xyce::packedByteCount(*itsc->second);
  }

  return byteCount;
}

//-----------------------------------------------------------------------------
// Function      : CircuitContext::pack
// Purpose       : Packs circuit context into char buffer using MPI_PACK.
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
template<>
void
Pack<IO::CircuitContext>::pack(
  const IO::CircuitContext &    circuit_context,
  char *                        buf,
  int                           bsize,
  int &                         pos,
  N_PDS_Comm *                  comm )
{
  int size, length, i;
#ifdef Xyce_COUNT_PACKED_BYTES
  int predictedPos = pos+packedByteCount();
#endif

  // pack name
  length = circuit_context.name_.length();
  comm->pack( &length, 1, buf, bsize, pos );
  comm->pack( circuit_context.name_.c_str(), length, buf, bsize, pos );

  // pack device count
  comm->pack( &circuit_context.deviceCount_, 1, buf, bsize, pos );

  // pack models_
  IO::ModelMap::const_iterator it_spbM = circuit_context.models_.begin();
  IO::ModelMap::const_iterator it_speM = circuit_context.models_.end();
  size = circuit_context.models_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  for( ; it_spbM != it_speM; ++it_spbM )
  {
    // ---- pack key
    length = it_spbM->first.length();
    comm->pack( &length, 1, buf, bsize, pos );
    comm->pack( it_spbM->first.c_str(), length, buf, bsize, pos );

    // ---- pack data
    Xyce::pack(*it_spbM->second, buf, bsize, pos, comm );
  }

  // pack unresolved functions
  size = circuit_context.unresolvedFunctions_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  for( i = 0; i < size; ++i )
  {
    Xyce::pack(circuit_context.unresolvedFunctions_[i], buf, bsize, pos, comm );
  }

  // pack instance list
  std::vector< std::string >::const_iterator it_stbL = circuit_context.instanceList_.begin();
  std::vector< std::string >::const_iterator it_steL = circuit_context.instanceList_.end();
  size = circuit_context.instanceList_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  for( ; it_stbL != it_steL; ++it_stbL )
  {
    length = it_stbL->length();
    comm->pack( &length, 1, buf, bsize, pos );
    comm->pack( it_stbL->c_str(), length, buf, bsize, pos );
  }

  // pack node list
  size = circuit_context.nodeList_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  for( i = 0; i < size; ++i )
  {
    length = circuit_context.nodeList_[ i ].length();
    comm->pack( &length, 1, buf, bsize, pos );
    comm->pack( circuit_context.nodeList_[ i ].c_str(), length, buf, bsize, pos );
  }

  // pack subcircuit parameters
  size = circuit_context.subcircuitParameters_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  for (Util::ParamList::const_iterator it = circuit_context.subcircuitParameters_.begin(), end = circuit_context.subcircuitParameters_.end(); it != end; ++it)
  {
    Pack<Util::Param>::pack(*it, buf, bsize, pos, comm );
  }

  // pack unresolved params
  size = circuit_context.unresolvedParams_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  for (Util::ParamList::const_iterator it = circuit_context.unresolvedParams_.begin(), end = circuit_context.unresolvedParams_.end(); it != end; ++it)
  {
    Pack<Util::Param>::pack(*it, buf, bsize, pos, comm );
  }

  // pack global node names
  size = circuit_context.globalNodes_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  std::set<std::string>::const_iterator globalNodes_i, globalNodes_end;
  globalNodes_i = circuit_context.globalNodes_.begin();
  globalNodes_end = circuit_context.globalNodes_.end();
  for ( ; globalNodes_i != globalNodes_end ; ++globalNodes_i )
  {
    length = globalNodes_i->length();
    comm->pack( &length, 1, buf, bsize, pos );
    comm->pack( globalNodes_i->c_str(), length, buf, bsize, pos );
  }

  // pack global params
  size = circuit_context.unresolvedGlobalParams_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  for (Util::ParamList::const_iterator it = circuit_context.unresolvedGlobalParams_.begin(), end = circuit_context.unresolvedGlobalParams_.end(); it != end; ++it)
  {
    Pack<Util::Param>::pack(*it, buf, bsize, pos, comm );
  }

  // pack mutual inductances
  size = circuit_context.mutualInductances_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  for( i = 0; i < size; ++i )
  {
    Xyce::pack(circuit_context.mutualInductances_[i], buf, bsize, pos, comm );
  }

  // pack circuitContextTable_
  size = circuit_context.circuitContextTable_.size();
  comm->pack( &size, 1, buf, bsize, pos );
  unordered_map< std::string, IO::CircuitContext * >::const_iterator itsc = circuit_context.circuitContextTable_.begin();
  unordered_map< std::string, IO::CircuitContext * >::const_iterator itsc_end = circuit_context.circuitContextTable_.end();
  for ( ; itsc != itsc_end; ++itsc )
  {
    length = itsc->first.length();
    comm->pack( &length, 1, buf, bsize, pos );
    comm->pack( itsc->first.c_str(), length, buf, bsize, pos );
    Xyce::pack(*itsc->second, buf, bsize, pos, comm );
  }

#ifdef Xyce_COUNT_PACKED_BYTES
  if (pos != predictedPos)
  {
    DevelFatal(*this, "CircuitContext::pack") << "Predicted pos does not match actual pos";
  }
#endif
}


//-----------------------------------------------------------------------------
// Function      : CircuitContext::unpack
// Purpose       : Unpacks circuit context from char buffer using MPI_UNPACK.
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
template<>
void
Pack<IO::CircuitContext>::unpack(
  IO::CircuitContext &  circuit_context,
  char *                pB,
  int                   bsize,
  int &                 pos,
  N_PDS_Comm *          comm )
{
  int size = 0;
  int length = 0;

  // unpack name
  comm->unpack( pB, bsize, pos, &length, 1 );
  circuit_context.name_ = std::string( ( pB + pos ), length );
  pos += length;

  // unpack device count
  comm->unpack( pB, bsize, pos, &circuit_context.deviceCount_, 1 );

  // unpack models_
  comm->unpack( pB, bsize, pos, &size, 1 );

  for (int i = 0; i < size; ++i )
  {

    // ---- unpack key
    comm->unpack( pB, bsize, pos, &length, 1 );
    std::string aString(std::string( ( pB + pos ), length ));
    pos += length;

    // ---- unpack data
    circuit_context.models_[ aString ] = new IO::ParameterBlock();
    Xyce::unpack(*circuit_context.models_[aString], pB, bsize, pos, comm );
  }

  // unpack unresolved functions (vector of FunctionBlock)
  comm->unpack( pB, bsize, pos, &size, 1 );
  for (int i = 0; i < size; ++i )
  {
    IO::FunctionBlock aFuncBlk;
    Xyce::unpack(aFuncBlk, pB, bsize, pos, comm );
    circuit_context.unresolvedFunctions_.push_back( aFuncBlk );
  }

  // unpack instance list
  comm->unpack( pB, bsize, pos, &size, 1 );
  for (int i = 0;  i < size; ++i )
  {
    comm->unpack( pB, bsize, pos, &length, 1 );
    circuit_context.instanceList_.push_back( std::string( ( pB + pos ), length ) );
    pos += length;
  }

  // unpack node list
  comm->unpack( pB, bsize, pos, &size, 1 );
  for (int i = 0; i < size; ++i)
  {
    comm->unpack( pB, bsize, pos, &length, 1 );
    circuit_context.nodeList_.push_back( std::string( ( pB + pos ), length ) );
    pos += length;
  }

  // unpack subcircuit parameters
  comm->unpack( pB, bsize, pos, &size, 1 );
  for (int i = 0; i < size; ++i)
  {
    Util::Param aParam;
    Pack<Util::Param>::unpack(aParam, pB, bsize, pos, comm );
    circuit_context.subcircuitParameters_.push_back( aParam );
  }

  // unpack unresovled params vector
  comm->unpack( pB, bsize, pos, &size, 1 );
  for (int i = 0; i < size; ++i)
  {
    Util::Param aParam;
    Pack<Util::Param>::unpack(aParam, pB, bsize, pos, comm );
    circuit_context.unresolvedParams_.push_back( aParam );
  }

  // unpack global node names
  comm->unpack( pB, bsize, pos, &size, 1 );
  for (int i = 0;  i < size; ++i )
  {
    comm->unpack( pB, bsize, pos, &length, 1 );
    circuit_context.globalNodes_.insert( std::string( ( pB + pos ), length ) );
    pos += length;
  }

  // unpack unresovled params vector
  comm->unpack( pB, bsize, pos, &size, 1 );
  for (int i = 0; i < size; ++i)
  {
    Util::Param aParam;
    Pack<Util::Param>::unpack(aParam, pB, bsize, pos, comm );
    circuit_context.unresolvedGlobalParams_.push_back( aParam );
  }

  // unpack mutual inductances vector
  comm->unpack( pB, bsize, pos, &size, 1 );
  for (int i = 0; i < size; ++i)
  {
    IO::CircuitContext::MutualInductance MI;
    Xyce::unpack(MI, pB, bsize, pos, comm );
    circuit_context.mutualInductances_.push_back( MI );
  }

  // unpack circuitContextTable_
  comm->unpack( pB, bsize, pos, &size, 1 );
  for (int i = 0; i < size; ++i)
  {
    comm->unpack( pB, bsize, pos, &length, 1 );
    std::string tmp( ( pB + pos ), length );
    pos += length;
    std::pair< unordered_map< std::string, IO::CircuitContext *>::iterator, bool > p =
      circuit_context.circuitContextTable_.insert(
        std::pair< std::string, IO::CircuitContext *>(
          tmp,
          new IO::CircuitContext( circuit_context.opBuilderManager_, circuit_context.contextList_, circuit_context.currentContextPtr_ ) ) );

    // set the parent context of my children to me
    p.first->second->setParentContextPtr( &circuit_context );
    Xyce::unpack(*p.first->second, pB, bsize, pos, comm );
  }

  circuit_context.currentContextPtr_ = &circuit_context;
}

//-----------------------------------------------------------------------------
// Function      : CircuitContext::MutualInductance::packedByteCount
// Purpose       : Counts bytes needed to pack block.
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra
// Creation Date : 8/27/04
//-----------------------------------------------------------------------------
template<>
int
Pack<IO::CircuitContext::MutualInductance>::packedByteCount(
  const IO::CircuitContext::MutualInductance &  mutual_inductance)
{
  int byteCount = 0;

  // coupling value
  byteCount += sizeof(int);
  byteCount += mutual_inductance.coupling.length();

  // model name
  byteCount += sizeof(int);
  byteCount += mutual_inductance.model.length();

  // first inductor name
  byteCount += sizeof(int);
  byteCount += mutual_inductance.firstInductor.length();

  // inductor info
  byteCount += sizeof(int);
  std::map<std::string,double>::const_iterator iterI = mutual_inductance.inductors.begin();
  std::map<std::string,double>::const_iterator  endI = mutual_inductance.inductors.end();
  for( ; iterI != endI; ++iterI )
  {
    byteCount += sizeof(int);
    byteCount += iterI->first.length();
    byteCount += sizeof(double);
  }

  return byteCount;
}

//-----------------------------------------------------------------------------
// Function      : CircuitContext::MutualInductance::pack
// Purpose       : Packs MI into char buffer using MPI_PACK.
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra
// Creation Date : 08/27/04
//-----------------------------------------------------------------------------
template<>
void
Pack<IO::CircuitContext::MutualInductance>::pack(
  const IO::CircuitContext::MutualInductance &  mutual_inductance,
  char *                                        buf,
  int                                           bsize,
  int &                                         pos,
  N_PDS_Comm *                                  comm ) 
{
  int size, length;
#ifdef Xyce_COUNT_PACKED_BYTES
  int predictedPos = pos+packedByteCount();
#endif

  // coupling value
  length = mutual_inductance.coupling.length();
  comm->pack( &length, 1, buf, bsize, pos );
  if( length ) comm->pack( mutual_inductance.coupling.c_str(), length, buf, bsize, pos );

  // model name
  length = mutual_inductance.model.length();
  comm->pack( &length, 1, buf, bsize, pos );
  if( length ) comm->pack( mutual_inductance.model.c_str(), length, buf, bsize, pos );

  // first inductor name
  length = mutual_inductance.firstInductor.length();
  comm->pack( &length, 1, buf, bsize, pos );
  if( length ) comm->pack( mutual_inductance.firstInductor.c_str(), length, buf, bsize, pos );

  // pack inductors
  size = mutual_inductance.inductors.size();
  comm->pack( &size, 1, buf, bsize, pos );
  std::map<std::string,double>::const_iterator iterI = mutual_inductance.inductors.begin();
  std::map<std::string,double>::const_iterator  endI = mutual_inductance.inductors.end();
  for( ; iterI != endI; ++iterI )
  {
    length = iterI->first.length();
    comm->pack( &length, 1, buf, bsize, pos );
    comm->pack( iterI->first.c_str(), length, buf, bsize, pos );
    comm->pack( &(iterI->second), 1, buf, bsize, pos );
  }
#ifdef Xyce_COUNT_PACKED_BYTES
  if (pos != predictedPos)
  {
    DevelFatal(*this, "CircuitContext::MutualInductance::pack") << "Predicted pos does not match actual pos";
  }
#endif
}

//-----------------------------------------------------------------------------
// Function      : CircuitContext::MutualInductance::unpack
// Purpose       : Unpacks MI from char buffer using MPI_UNPACK.
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra
// Creation Date : 08/27/04
//-----------------------------------------------------------------------------
template<>
void
Pack<IO::CircuitContext::MutualInductance>::unpack(
  IO::CircuitContext::MutualInductance &        mutual_inductance,
  char *                                        pB,
  int                                           bsize,
  int &                                         pos,
  N_PDS_Comm *                                  comm )
{
  int size, length, i;
  double val;

  // coupling value
  comm->unpack( pB, bsize, pos, &length, 1 );
  if( length )
  {
    mutual_inductance.coupling = std::string( ( pB + pos ), length );
    pos += length;
  }

  // model name
  comm->unpack( pB, bsize, pos, &length, 1 );
  if( length )
  {
    mutual_inductance.model = std::string( ( pB + pos ), length );
    pos += length;
  }

  // first inductor name
  comm->unpack( pB, bsize, pos, &length, 1 );
  if( length )
  {
    mutual_inductance.firstInductor = std::string( ( pB + pos ), length );
    pos += length;
  }

  // unpack inductors
  mutual_inductance.inductors.clear();
  comm->unpack( pB, bsize, pos, &size, 1 );
  for( i = 0; i < size; ++i )
  {
    comm->unpack( pB, bsize, pos, &length, 1 );
    std::string name(std::string( ( pB + pos ), length ));
    pos += length;
    comm->unpack( pB, bsize, pos, &val, 1 );
    mutual_inductance.inductors[name] = val;
  }
}

} // namespace Xyce

//-----------------------------------------------------------------------------
// Function      : packCircuitContext
// Purpose       : send circuit context to all procs
// Special Notes :
// Scope         : non-member function
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
int Xyce::IO::packCircuitContext(Xyce::IO::CircuitContext* circuit_contexts, char* char_buffer, 
                                 int char_buffer_size, Xyce::Parallel::Communicator* pds_comm_ptr )
{
  int bsize = 0;

  if (Parallel::is_parallel_run(pds_comm_ptr->comm()))
  {
    int pos = 0;

    // pack circuit contexts
    Xyce::pack(*circuit_contexts, char_buffer, char_buffer_size, pos, pds_comm_ptr );

    bsize=pos;
  }

  return bsize;
}


//-----------------------------------------------------------------------------
// Function      : unpackCircuitContext
// Purpose       : receive circuit context from proc 0
// Special Notes :
// Scope         : non-member function
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool Xyce::IO::unpackCircuitContext(Xyce::IO::CircuitContext* circuit_contexts, char* char_buffer, 
                                    int bsize, Xyce::Parallel::Communicator* pds_comm_ptr )
{
  if (Parallel::is_parallel_run(pds_comm_ptr->comm()))
  {
    int pos = 0;

    circuit_contexts->setParentContextPtr( NULL );

    Xyce::unpack(*circuit_contexts, char_buffer, bsize, pos, pds_comm_ptr);
  }

  return true;
}

