// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_IO_DistToolDefault.C,v $
//
// Purpose        : Defines the DistToolDefault class.  Distribution tool
//                  buffers and distributes circuit blocks (and related data
//                  such as option blocks, metadata, etc) for/during parsing.
// Special Notes  :
//
// Creator        : Eric Rankin, SNL
//
// Creation Date  : 03/12/2003
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.10 $
//
// Revision Date  : $Date: 2015/11/10 21:02:33 $
//
// Current Owner  : $Author: hkthorn $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iostream>

#include <N_IO_fwd.h>
#include <N_DEV_DeviceMgr.h>
#include <N_ERH_ErrorMgr.h>
#include <N_ERH_Message.h>
#include <N_IO_CircuitBlock.h>
#include <N_IO_CmdParse.h>
#include <N_IO_DistToolDefault.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_ParsingHelpers.h>
#include <N_PDS_Comm.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_Stats.h>

namespace Xyce {
namespace IO {

//-----------------------------------------------------------------------------
// Function      : DistToolDefault::DistToolDefault
// Purpose       : ctor
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
DistToolDefault::DistToolDefault(
  Parallel::Communicator *                 pdsCommPtr,
  PkgOptionsMgr &                          options_manager,
  CircuitBlock &                           circuit_block,
  std::map<std::string,FileSSFPair>      & ssfMap,
  std::map<std::string, IncludeFileInfo> & iflMap)
  : pdsCommPtr_(pdsCommPtr),
    numProcs_(1),
    currProc_(0),
    optionsManager_(options_manager),
    circuitBlock_(circuit_block),
    circuitContext_(circuit_block.getCircuitContextPtr()),
    options_(circuit_block.getOptionsTable()),
    ssfMap_(ssfMap),
    iflMap_(iflMap),
    netlistFilename_(),
    device_(*circuit_block.getCircuitContextPtr(), circuit_block.getMetadata()),
    mainCircuitPtr_(&circuit_block), 
    parentCircuitPtr_(NULL),
    currentCircuitPtr_(&circuit_block),
    preprocessFilter_(PreprocessType::NUM_PREPROCESS, false),
    ssfPtr_(0),
    procDeviceCount_(0),
    devices_(0),
    deviceLinesSent_(0),
    charBufferSize_(250000),
    charBufferPos_(0),
    bufs_(),
    bufSize_(),
    charBuffer_(0),
    remove_any_redundant_(false),
    circuitContextReady_(false),
    circuitOptionsReady_(false),
    subcircuitNames_(),
    subcircuitNodes_(),
    subcircuitPrefixes_(),
    subcircuitParams_()
{
  numProcs_ = pdsCommPtr_->numProc();

  // Set the circuit context and options table.
  setCircuitContext();
  setCircuitOptions();

  // determine how many devices to send to each far proc
  devices_ = (circuit_block.getCircuitContextPtr())->getTotalDeviceCount();
  procDeviceCount_ = devices_ / numProcs_;

  // procID == 0 may not be true if we're running in a hierarchical parallel context
  // so check if numProc() == 1 too.
  if (pdsCommPtr_->procID() == 0)
  {
    currProc_ = numProcs_ == 1 ? 0 : 1;
  }
}


//-----------------------------------------------------------------------------
// Function      : DistToolDefault::~DistToolDefault
// Purpose       : dtor
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
DistToolDefault::~DistToolDefault()
{
  delete[] charBuffer_;
}

//-----------------------------------------------------------------------------
// Function      : DistToolDefault::setCircuitContext
// Purpose       : check buffer size wrt the circuit context size
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void
DistToolDefault::setCircuitContext()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if (pdsCommPtr_->procID() == 0)
    {
      // resize buffer if necessary
      int tmpSize = Xyce::packedByteCount(*circuitContext_) + sizeof( int );

      if ( tmpSize > charBufferSize_ )
      {
        charBufferSize_ = tmpSize;
      }

      // set flag to tx global data
      circuitContextReady_ = true;
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : DistToolDefault::setCircuitOptions
// Purpose       : stage options for tx
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void
DistToolDefault::setCircuitOptions()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if (pdsCommPtr_->procID() == 0)
    {
      // resize buffer if necessary
      int tmpSize = sizeof(int);
      for (std::list<Util::OptionBlock>::const_iterator it = options_.begin(), end = options_.end(); it != end; ++it)
        tmpSize += Xyce::packedByteCount(*it);

      if ( tmpSize > charBufferSize_ )
        charBufferSize_ = tmpSize;

      // set flag to tx global data
      circuitOptionsReady_ = true;
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : DistToolDefault::circuitDeviceLine
// Purpose       : Send a circuit device line to current proc
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool DistToolDefault::circuitDeviceLine(TokenVector & deviceLine )
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if( currProc_ != 0 )
    {
      int size = 0;

      // count line type
      size += sizeof(char);

      // count line size
      size += sizeof(int);

      // count device line
      int deviceLineSize = deviceLine.size();
      for (int i = 0; i < deviceLineSize; ++i)
      {
        size += Xyce::packedByteCount(deviceLine[i]);
      }

      // flush buffer as needed
      send(size);

      // pack line type; "d"evice line
      char lineType = 'd';
      pdsCommPtr_->pack( &lineType, 1, charBuffer_, charBufferSize_, charBufferPos_ );

      // pack device line size
      pdsCommPtr_->pack( &deviceLineSize, 1, charBuffer_, charBufferSize_, charBufferPos_ );

      // pack device line
      for (int i = 0; i < deviceLineSize; ++i)
      {
        Xyce::pack(deviceLine[i],  charBuffer_, charBufferSize_, charBufferPos_, pdsCommPtr_ );
      }

      // increment device line counter
      deviceLinesSent_++;

      // manage proc change if n/p reached
      if (deviceLinesSent_ >= procDeviceCount_)
      {
        int minus1 = -1;

        // flush buffer
        send();

        // allow this proc to exit receive loop and begin processing
        pdsCommPtr_->send( &minus1, 1, currProc_ );

        // reset device line counter
        deviceLinesSent_ = 0;

        // move to next processor
        currProc_++;

        // switch to proc 0 if nearly complete
        if (currProc_ == numProcs_)
        {
          currProc_ = 0;
        }

        // otherwise, prepare currProc for next set of device lines
        else
        {
          int length = netlistFilename_.size();
          lineType = 'f';

          // flush buffer as needed
          send(sizeof(char) + sizeof(int) + length);

          // pack the filename
          pdsCommPtr_->pack( &lineType, 1, charBuffer_, charBufferSize_, charBufferPos_ );
          pdsCommPtr_->pack( &length, 1, charBuffer_, charBufferSize_, charBufferPos_ );
          pdsCommPtr_->pack( netlistFilename_.c_str(), length, charBuffer_, charBufferSize_, charBufferPos_ );

          // pack the subcircuit names
          int subcircuitNamesSize = subcircuitNames_.size();
          for (int i = 0; i < subcircuitNamesSize; ++i)
          {
            if( currProc_ != 0 )
            {
              packSubcircuitData(subcircuitNames_[i], subcircuitNodes_[i],
                                 subcircuitPrefixes_[i], subcircuitParams_[i]);
            }
          }
        }
      }

      return true;
    }
    else
      return false;
  }

  return false;
}

//-----------------------------------------------------------------------------
// Function      : DistToolDefault::endDeviceLines()
// Purpose       : Make sure other processors have been told that device processing
//                 is ended.  This can be a problem with small device count circuits
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 03/06/06
//-----------------------------------------------------------------------------
void DistToolDefault::endDeviceLines()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if (currProc_ > 0)
    {
      int minus1 = -1;

      // flush buffer
      send();

      // end parsing on current node
      pdsCommPtr_->send( &minus1, 1, currProc_ );
      ++currProc_;

      // end parsing on remaining nodes
      for ( ; currProc_ < numProcs_ ; ++currProc_)
      {
        pdsCommPtr_->send( &minus1, 1, currProc_ );
      }

      // complete any remaining parser action on node 0
      currProc_ = 0;
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : DistToolDefault::circuitStart
// Purpose       : change current subcircuit context after a new subcircuit is
//               : started
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool DistToolDefault::circuitStart( std::string const & subcircuitName,
                                     std::vector<std::string> const & nodes,
                                     std::string const & prefix,
                                     std::vector<Device::Param> const & params )
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if( currProc_ != 0 )
    {
      // save new context data on stacks
      subcircuitNames_.push_back( subcircuitName );
      subcircuitPrefixes_.push_back( prefix );
      subcircuitNodes_.push_back( nodes );
      subcircuitParams_.push_back( params );

      // pack data into buffer
      packSubcircuitData( subcircuitName, nodes, prefix, params );
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DistToolDefault::circuitEnd
// Purpose       : change current subcircuit context to previous context
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool DistToolDefault::circuitEnd()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if( currProc_ != 0 )
    {
      // remove latest context data on stacks
      subcircuitNames_.pop_back();
      subcircuitPrefixes_.pop_back();
      subcircuitNodes_.pop_back();
      subcircuitParams_.pop_back();

      // flush buffer as needed
      send(sizeof(char));

      // pack line type; subcircuit "e"nd
      char lineType = 'e';
      pdsCommPtr_->pack( &lineType, 1, charBuffer_, charBufferSize_, charBufferPos_ );
    }
  }

  return true;
}


//-----------------------------------------------------------------------------
// Function      : DistToolDefault::setFileName
// Purpose       : Change name of netlist file
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 03/10/06
//-----------------------------------------------------------------------------
void DistToolDefault::setFileName(std::string const & fileNameIn)
{
  netlistFilename_ = fileNameIn;
  circuitBlock_.setFileName( fileNameIn );

  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    char lineType = 'f';
    int length = netlistFilename_.size();

    // flush buffer as needed
    send(sizeof(char) + sizeof(int) + length);

    pdsCommPtr_->pack( &lineType, 1, charBuffer_, charBufferSize_, charBufferPos_ );
    pdsCommPtr_->pack( &length, 1, charBuffer_, charBufferSize_, charBufferPos_ );
    pdsCommPtr_->pack( netlistFilename_.c_str(), length, charBuffer_, charBufferSize_, charBufferPos_ );
  }
}


//-----------------------------------------------------------------------------
// Function      : DistToolDefault::distributeDevices()
// Purpose       : Distribute the devices according to the strategy.
// Special Notes : The strategy is first-come-first-served, flattening the circuit
//               : as you go through the netlist files.
// Scope         : public
// Creator       : Heidi Thornquist, SNL
// Creation Date : 08/28/2014
//-----------------------------------------------------------------------------
void DistToolDefault::distributeDevices()
{
  Xyce::Parallel::Machine comm = pdsCommPtr_->comm();

  if (DEBUG_IO) {
    Xyce::dout() << "Pass 2 parsing of netlist file and distributing devices: " <<
      circuitBlock_.getNetlistFilename() << std::endl;
  }

  if (Parallel::rank(comm) == 0)
  {
    setFileName(circuitBlock_.getNetlistFilename());

    ssfPtr_ = ssfMap_[netlistFilename_].second;
    ssfPtr_->setLocation( circuitBlock_.getStartPosition() );
    ssfPtr_->setLineNumber( circuitBlock_.getLineStartPosition() );

    // If this is the main circuit, skip over the title line and continue.
    std::ifstream* netlistIn = ssfMap_[netlistFilename_].first;
    std::string title("");
    netlistIn->clear();
    netlistIn->seekg(0, std::ios::beg);
    Xyce::IO::readLine( *netlistIn, title ); 
    ssfPtr_->changeCursorLineNumber( 1 );

    // Instantiate all devices in the current context.
    std::string libSelect, libInside;
    TokenVector line;

    while (getLine(line, libSelect, libInside))
    {
      if (!line.empty() && compare_nocase(line[0].string_.c_str(), ".ends") != 0)
      {
        // parse locally if distool does not distribute
        if( !circuitDeviceLine( line ) )
        {
          handleDeviceLine( line, libSelect, libInside );
        }
      }
    }

    // send MIs if present in top circuit
    if( circuitContext_->haveMutualInductances() )
    {
      // Distribute each YMI?|name device line, end of circuit
      int n = circuitContext_->getNumMILines();
      for( int i = 0; i < n; ++i )
      {
        // parse locally if not distributed to another processor
        if( !circuitDeviceLine( circuitContext_->getMILine( i ) ) )
        {
          handleDeviceLine( circuitContext_->getMILine( i ), libSelect, libInside );
        }
      }
    }

    endDeviceLines();

  }
  else
  {
    // Wait to receive the circuit data from processor 0.
    receiveCircuitData();
  }

  if (DEBUG_IO)
    Xyce::dout() << "Done with pass 2 netlist file parsing and device distribution" << std::endl;

  // Just in case an error was reported in parsing the netlist.
  N_ERH_ErrorMgr::safeBarrier(comm);
}


//-----------------------------------------------------------------------------
// Function      : DistToolDefault::packSubcircuitData
// Purpose       : pack subcircuit data into buffer; helper for circuitStart()
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool DistToolDefault::packSubcircuitData( std::string const & subcircuitName,
                                           std::vector<std::string> const & nodes,
                                           std::string const & prefix,
                                           std::vector<Device::Param> const & params )
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    int size = 0;

    // count line type, name size, and name chars
    size += sizeof(char) + sizeof(int) + subcircuitName.length();

    // count # of nodes,
    size += sizeof(int);

    // count node sizes and node chars
    std::vector<std::string>::const_iterator it_sbL = nodes.begin();
    std::vector<std::string>::const_iterator it_seL = nodes.end();
    for( ; it_sbL != it_seL; ++it_sbL  )
    {
      size += sizeof(int) + it_sbL->length();
    }

    // count prefix size and prefix chars
    size += sizeof(int) + prefix.length();

    // count # of params
    size += sizeof(int);

    // count params
    std::vector<Device::Param>::const_iterator it_pbL = params.begin();
    std::vector<Device::Param>::const_iterator it_peL = params.end();
    for( ; it_pbL != it_peL; ++it_pbL  )
    {
      size += Pack<Device::Param>::packedByteCount(*it_pbL);
    }

    // flush buffer as needed
    send(size);


    // pack line type; subcircuit "s"tart
    char lineType = 's';
    pdsCommPtr_->pack( &lineType, 1, charBuffer_, charBufferSize_, charBufferPos_ );

    // pack name size and name chars
    int length = subcircuitName.length();
    pdsCommPtr_->pack( &length, 1, charBuffer_, charBufferSize_, charBufferPos_ );
    pdsCommPtr_->pack( subcircuitName.c_str(), length, charBuffer_, charBufferSize_, charBufferPos_ );

    // pack # of nodes
    size = nodes.size();
    pdsCommPtr_->pack( &size, 1, charBuffer_, charBufferSize_, charBufferPos_ );

    // pack nodes name sizes and chars
    it_sbL = nodes.begin();
    for( ; it_sbL != it_seL; ++it_sbL  )
    {
      length = it_sbL->length();
      pdsCommPtr_->pack( &length, 1, charBuffer_, charBufferSize_, charBufferPos_ );
      pdsCommPtr_->pack( it_sbL->c_str(), length, charBuffer_, charBufferSize_, charBufferPos_ );
    }

    // pack prefix size and prefix chars
    length = prefix.length();
    pdsCommPtr_->pack( &length, 1, charBuffer_, charBufferSize_, charBufferPos_ );
    pdsCommPtr_->pack( prefix.c_str(), length, charBuffer_, charBufferSize_, charBufferPos_ );

    // pack # of params
    size = params.size();
    pdsCommPtr_->pack( &size, 1, charBuffer_, charBufferSize_, charBufferPos_ );

    // pack params
    it_pbL = params.begin();
    for( ; it_pbL != it_peL; ++it_pbL  )
    {
      Pack<Device::Param>::pack(*it_pbL, charBuffer_, charBufferSize_, charBufferPos_, pdsCommPtr_ );
    }
  }

  return true;
}


//-----------------------------------------------------------------------------
// Function      : DistToolDefault::receiveCircuitData
// Purpose       : receive all data from proc 0 and store it in device buffer
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool DistToolDefault::receiveCircuitData()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    int bsize;
    char *currBuffer;

    // receive the device lines first
    while (true)
    {
      pdsCommPtr_->recv( &bsize, 1, 0 );

      // check for halt request
      if (bsize < 0)
      {
        break;
      }

      currBuffer = new char[bsize];
      bufs_.push_back(currBuffer);
      bufSize_.push_back(bsize);

      pdsCommPtr_->recv(currBuffer, bsize, 0);
    }

    return processDeviceBuffer();
  }
  else
    return true;
}


//-----------------------------------------------------------------------------
// Function      : DistToolDefault::send
// Purpose       : send buffer from proc 0 and adjust buffer limits
// Special Notes : size == -1, and no args send(), will force a transmission
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void
DistToolDefault::send(
  int                   size)
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    // transmit buffer if next set of object will fill it completely, or forced
    if ((charBufferPos_ + size >= charBufferSize_) || (size == -1))
    {
      if (DEBUG_DISTRIBUTION)
        dout() << "node " << pdsCommPtr_->procID() << " packed "
               << charBufferPos_ << " bytes into " << charBufferSize_ << " byte buffer"
               << " [data size:  " << size << "]" << std::endl
               << "node " << currProc_ << " was sent "
               << deviceLinesSent_ << " of " << procDeviceCount_ << " devices "
               << (float)deviceLinesSent_ / procDeviceCount_ * 100.0
               << ( size == -1 ? "% complete  [flushed]" : "% complete"  ) << std::endl;

      // tx buffer size
      pdsCommPtr_->send( &charBufferPos_, 1, currProc_ );

      // tx buffer contents
      pdsCommPtr_->send( charBuffer_, charBufferPos_, currProc_ );

      // reset counters
      charBufferPos_ = 0;

      // adjust buffer size if objects larger than existing buffer
      if (size > charBufferSize_)
      {

        // free memory
        // delete [] charBuffer_;

        // set new buffer size
        charBufferSize_ = size;

        // reserve memory; including offset
        charBuffer_ = new char[charBufferSize_ + sizeof(char) + sizeof(int)];

        if (DEBUG_DISTRIBUTION)
          dout() << "node " << pdsCommPtr_->procID()
                 << " resized buffer to " << charBufferSize_
                 << " with offset " << sizeof(char) + sizeof(int) << std::endl;
      }
    }
  }
}


//-----------------------------------------------------------------------------
// Function      : DistToolDefault::processDeviceBuffer
// Purpose       : process all data in the buffer
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool
DistToolDefault::processDeviceBuffer()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    int size, bsize, length, pos, i;

    char *currBuffer;
    unsigned int ib;

    CircuitContext* circuit_context = circuitBlock_.getCircuitContextPtr();

    // process received device lines, contexts info, and exit calls
    for (ib=0 ; ib<bufs_.size() ; ++ib)
    {
      currBuffer = bufs_[ib];
      bsize = bufSize_[ib];

      // get ready to process buffer
      pos = 0;

      // break apart buffered lines and parse
      while( pos < bsize )
      {
        char lineType = '\0';
        // get the linetype marker that indicates incoming line
        pdsCommPtr_->unpack( currBuffer, bsize, pos, &lineType, 1 );

        // process line according to type
        switch( lineType )
        {
          case 'd': // process a device line
          {
            // unpack the device line
            pdsCommPtr_->unpack( currBuffer, bsize, pos, &size, 1 );
            TokenVector deviceLine( size );

            for( i = 0; i < size; ++i )
            {
              Xyce::unpack(deviceLine[i], currBuffer, bsize, pos, pdsCommPtr_ );
            }

            // hand to circuit block for processing
            std::string libSelect(""), libInside(""); 
            handleDeviceLine( deviceLine, libSelect, libInside );

            break;
          }

          case 's': // subcircuit start found
          {
            // get the subcircuit name
            pdsCommPtr_->unpack( currBuffer, bsize, pos, &length, 1 );
            std::string subcircuitName(std::string( ( currBuffer + pos ), length ));
            pos += length;

            // get the nodes for this subcircuit call
            std::vector<std::string> nodes;
            pdsCommPtr_->unpack( currBuffer, bsize, pos, &size, 1 );
            for( i = 0; i < size; ++i )
            {
              pdsCommPtr_->unpack( currBuffer, bsize, pos, &length, 1 );
              nodes.push_back( std::string( ( currBuffer + pos ), length ) );
              pos += length;
            }

            // get the prefix
            pdsCommPtr_->unpack( currBuffer, bsize, pos, &length, 1 );
            std::string prefix(std::string( ( currBuffer + pos ), length ));
            pos += length;

            // get params
            std::vector<Device::Param> params;
            pdsCommPtr_->unpack( currBuffer, bsize, pos, &size, 1 );
            for( i = 0; i < size; ++i )
            {
              Device::Param param;
              Pack<Device::Param>::unpack(param, currBuffer, bsize, pos, pdsCommPtr_ );
              params.push_back( param );
            }

            // send to cktblk
            circuit_context->setContext( subcircuitName, prefix, nodes );
            circuit_context->resolve( params );

            break;
          }

          case 'e': // subcircuit end found
          {
            // adjust the circuit context pointer
            circuit_context->restorePreviousContext();

            break;
          }

          case 'f': // change netlist file name
          {
            pdsCommPtr_->unpack( currBuffer, bsize, pos, &length, 1 );
            netlistFilename_ = std::string( ( currBuffer + pos ), length );
            pos += length;
            setFileName(netlistFilename_);

            break;
          }

          default:  // something went wrong
          {
            Report::DevelFatal().in("DistToolDefault::processDeviceBuffer")
              << "Node " << pdsCommPtr_->procID() << " received invalid message type \"" << lineType;
          }
        }
      }
      delete [] bufs_[ib];
    }

    bufs_.clear();
    bufSize_.clear();
  }

  return true;
}


//-----------------------------------------------------------------------------
// Function      : DistToolDefault::broadcastGlobalData
// Purpose       : send bufsize, options, and context to procs
// Special Notes :
// Scope         : private
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool
DistToolDefault::broadcastGlobalData()
{
  // Get the preprocessor filter before broadcasting.
  preprocessFilter_ = circuitBlock_.getPreprocessFilter();

  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    // tx buffer size
    pdsCommPtr_->bcast( &charBufferSize_, 1, 0 );

    // check for halt request
    if (charBufferSize_ < 0)
    {
      return false;
    }

    // reserve memory; including offset
    charBuffer_ = new char[charBufferSize_ + sizeof(char) + sizeof(int)];

    // make sure data is ready to tx
    if (numProcs_ > 1)
    {
      int bsize = 0;

      // tx global data
      if (circuitOptionsReady_)
      {
        bsize = Xyce::IO::packCircuitOptions(options_, charBuffer_, charBufferSize_, pdsCommPtr_);
      }

      // broadcast options
      pdsCommPtr_->bcast( &bsize, 1, 0 );
      pdsCommPtr_->bcast( charBuffer_, bsize, 0 );

      // receive global data
      if (!circuitOptionsReady_)
      {
        Xyce::IO::unpackCircuitOptions(options_, charBuffer_, bsize, pdsCommPtr_);
      }

      // tx global data
      if (circuitContextReady_)
      {
        bsize = Xyce::IO::packCircuitContext(circuitContext_, charBuffer_, charBufferSize_, pdsCommPtr_);
      } 
      // broadcast context
      pdsCommPtr_->bcast( &bsize, 1, 0 );
      pdsCommPtr_->bcast( charBuffer_, bsize, 0 );

      if (!circuitContextReady_)
      {
        Xyce::IO::unpackCircuitContext(circuitContext_, charBuffer_, bsize, pdsCommPtr_);
        circuitBlock_.receiveCircuitContext( *circuitContext_ );
      }
    }

    HangingResistor &hanging_resistor = circuitBlock_.getHangingResistor();
  
    // transform booleans into integers and transmit them
    bool netlistcopy = hanging_resistor.getNetlistCopy();
    bool oneterm = hanging_resistor.getOneTerm();
    bool nodcpath = hanging_resistor.getNoDCPath();

    int nlcopyint = 0;
    int otint = 0;
    int nodcint = 0;

    if (netlistcopy)
      nlcopyint = 1;
    if (oneterm)
      otint = 1;
    if (nodcpath)
      nodcint = 1;

    // tx global data
    pdsCommPtr_->bcast(&nlcopyint,1,0);
    pdsCommPtr_->bcast(&otint,1,0);
    pdsCommPtr_->bcast(&nodcint,1,0);

    std::string onetermres(hanging_resistor.getOneTermRes());
    std::string nodcpathres(hanging_resistor.getNoDCPathRes());
    int otreslength = onetermres.size();
    int nodcreslength = nodcpathres.size();

    // tx global data
    pdsCommPtr_->bcast(&otreslength,1,0);
    onetermres.resize(otreslength);
    pdsCommPtr_->bcast(&onetermres[0],otreslength,0);
    pdsCommPtr_->bcast(&nodcreslength,1,0);
    nodcpathres.resize(nodcreslength);
    pdsCommPtr_->bcast(&nodcpathres[0],nodcreslength,0);

    if (nlcopyint == 1)
      hanging_resistor.setNetlistCopy(true);
    if (otint == 1)
      hanging_resistor.setOneTerm(true);
    if (nodcint == 1)
      hanging_resistor.setNoDCPath(true);

    hanging_resistor.setOneTermRes(onetermres);
    hanging_resistor.setNoDCPathRes(nodcpathres);

    N_ERH_ErrorMgr::safeBarrier(pdsCommPtr_->comm());

    // tx preprocess filter
    for (int i=0; i<PreprocessType::NUM_PREPROCESS; i++)
    {
      int preprocint = 0;
      if ( preprocessFilter_[i] )
        preprocint = 1;

      pdsCommPtr_->bcast(&preprocint,1,0);

      if ( preprocint == 1 )
        preprocessFilter_[i] = true;
    }
  }

  if ( preprocessFilter_[PreprocessType::REDUNDANT_C]
       || preprocessFilter_[PreprocessType::REDUNDANT_D]
       || preprocessFilter_[PreprocessType::REDUNDANT_I]
       || preprocessFilter_[PreprocessType::REDUNDANT_L]
       || preprocessFilter_[PreprocessType::REDUNDANT_M]
       || preprocessFilter_[PreprocessType::REDUNDANT_Q]
       || preprocessFilter_[PreprocessType::REDUNDANT_R]
       || preprocessFilter_[PreprocessType::REDUNDANT_V] )
  {
    remove_any_redundant_ = true;
  }

  return true;
}

//----------------------------------------------------------------------------
// Function       : DistToolDefault::getLine
// Purpose        :
// Special Notes  :
// Scope          : private
// Creator        : Lon Waters
// Creation Date  : 08/01/2003
//----------------------------------------------------------------------------
bool DistToolDefault::getLine( TokenVector &line,
  const std::string &libSelect, std::string &libInside)
{
  int eof = 0;

  while (!eof)
  {
    {
      //Stats::StatTop _getLineStat("Tokenize Line");
      //Stats::TimeBlock _getLineTimer(_getLineStat);

      eof = !ssfPtr_->getLine(line,preprocessFilter_[PreprocessType::REPLACE_GROUND]); // Breaks the line into fields.
    }

    if (DEBUG_IO) {
      Xyce::dout() << "Pass 2 read netlist line:  "<< std::endl;
      for (unsigned int i = 0; i < line.size(); ++i)
      {
        Xyce::dout() << line[i].string_ << " ";
      }
      Xyce::dout() << std::endl;

      if (eof)
      {
        Xyce::dout() << "We are at the end of file, size of line is " << line.size()
                     << std::endl;
      }
    }

    // better not try to do anything if getLine returned an empty line!
    if ( line.empty() )
      continue;

    // Determine what to do with the parsed line.
    ExtendedString ES1(line[0].string_);
    ES1.toUpper();
    if ( !(libSelect != libInside && libInside != "" && ES1 != ".ENDL"))
    {
      char lineType;
      lineType = ES1[0];
      bool removecomponent=false;

      if (lineType >= 'A' && lineType <= 'Z')
      {

        if ( lineType == 'C' || lineType == 'D' || lineType == 'I' ||
             lineType == 'L' || lineType == 'R' || lineType == 'V')
        {
          if (remove_any_redundant_ && line.size() > 2) //make sure that there are two nodes to check!
          {
            ExtendedString node1 ( line[1].string_ );
            ExtendedString node2 ( line[2].string_ );
            node1.toUpper();
            node2.toUpper();

            removecomponent = IO::removeTwoTerminalDevice(preprocessFilter_,
                                                          lineType, node1, node2);
          }
        }
        else if (lineType == 'M' || lineType == 'Q')
        {
          if (remove_any_redundant_ && line.size() > 3) //make sure that there are three nodes to check!
          {
            ExtendedString node1 ( line[1].string_ );
            ExtendedString node2 ( line[2].string_ );
            ExtendedString node3 ( line[3].string_ );
            node1.toUpper();
            node2.toUpper();
            node3.toUpper();

            removecomponent = IO::removeThreeTerminalDevice(preprocessFilter_,
                                                            lineType, node1,
                                                            node2, node3);
          }
        }
        if (!removecomponent)
        {
          if (lineType != 'X' && lineType != 'K' && lineType != 'L')
          {
            // check for .INITCOND values
            std::string tmpName(circuitContext_->getPrefix());
            if (tmpName == "")
              tmpName = ES1;
            else
              tmpName = tmpName + ":" + ES1;

            if( mainCircuitPtr_->initCondIndex.find( tmpName ) !=
                mainCircuitPtr_->initCondIndex.end() )
            {
              // append the IC=val1...valN list to device line
              line.insert( line.end(),
                           ( mainCircuitPtr_->initCondIndex[tmpName] ).begin(),
                           ( mainCircuitPtr_->initCondIndex[tmpName] ).end() );
            }

            return true;
          }
          else if (lineType == 'L')
          {
            std::set< std::string > & cTable =
              circuitContext_->getAllCoupledInductors();

            if( cTable.find( ES1 ) == cTable.end() )
            {
              // treat as a normal device line if inductor is not coupled
              return true;
            }
            continue;
          }
          else if (lineType == 'X')
          {
            handleDeviceLine(line, libSelect, libInside);
          }
        }
        else
        {
          // This device is being removed, clear the line.
          line.clear();

          if (DEBUG_IO)
          {
            Xyce::dout() << "Netlist Parse 2:  ";
            Xyce::dout() << "removing component " << ES1 << ".  All nodes on the device";
            Xyce::dout() << " are the same."  << std::endl;
          }

          return true;
        }
      }
      else if (lineType == '.')
      {
        if (ES1 == ".SUBCKT")
        {
          // Jump to the end of the subcircuit.
          // Find the subcircuit corresponding to this instance.
          CircuitBlock* subcircuitPtr =
            currentCircuitPtr_->findSubcircuit(ExtendedString(line[1].string_).toUpper());

          ssfPtr_->setLocation(subcircuitPtr->getEndPosition());
          ssfPtr_->setLineNumber(subcircuitPtr->getLineEndPosition());

        }
        else if (ES1 == ".INCLUDE" || ES1 == ".INC" || ES1 == ".LIB")
        {
          std::string includeFile, libSelect_child;
          libSelect_child = libSelect;
          Xyce::IO::handleIncludeLine( (ssfMap_.size()==1), netlistFilename_, line,
                                       ES1, includeFile, libSelect_child, libInside );
          if (includeFile != "")
          {
            int result = parseIncludeFile(includeFile, libSelect_child);
            if (!result)
              return result;
          }
        }
        else if (ES1 == ".ENDL")
        {
          Xyce::IO::handleEndlLine ( netlistFilename_, line, libSelect, libInside );
        }
        else if (ES1 == ".ENDS")
        {
          return parentCircuitPtr_ == 0;
        }
        else if (ES1 == ".END")
        {
          return false;
        }
      }
      else
      {
        // Do nothing for other types of lines
      }
    }
  }
  // Only get here if end of file.

  return false;
}

//----------------------------------------------------------------------------
// Function       : DistToolDefault::handleDeviceLine
// Purpose        : Processor zero / serial processing of devices
// Special Notes  :
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 02/21/2003
//----------------------------------------------------------------------------
bool DistToolDefault::handleDeviceLine( TokenVector const& deviceLine,
    const std::string &libSelect, std::string &libInside)
{
  bool result;

  device_.clear();

  // Fully parse the device line and extract the relevant data.
  {
    //Stats::StatTop _extractData("Extract Device Data");
    //Stats::TimeBlock _extractDataTimer(_extractData);

    result = device_.extractData(netlistFilename_, deviceLine);
    if (!result)
      return result;
  }

  unordered_map<std::string, std::string> * nodeMapPtr = circuitContext_->getNodeMapPtr();
  result = instantiateDevice(device_, circuitContext_->getPrefix(), *nodeMapPtr, libSelect, libInside);

  return result;
}

//--------------------------------------------------------------------------
// Function      : DistToolDefault::parseIncludeFile
// Purpose       : Jump to include file for 2nd pass
// Special Notes :
// Creator       : Rob Hoekstra
// Creation Date : 08/02/2004
//--------------------------------------------------------------------------
bool DistToolDefault::parseIncludeFile(std::string const& includeFile,
             const std::string &libSelect)
{
  // Save current ssfPtr_ and netlistFilename_.
  SpiceSeparatedFieldTool* oldssfPtr = ssfPtr_;
  std::string old_netlistFilename(netlistFilename_);
  setFileName(includeFile);

  if (DEBUG_IO)
    Xyce::dout() << "Parsing include file Pass 2: " << includeFile << std::endl;

  // Find SSF for file
  if( !ssfMap_.count( includeFile ) )
  {
    Report::UserError() << "Could not find include file SSF " << includeFile;
    return false;
  }
  ssfPtr_ = ssfMap_[includeFile].second;

  //Save current position and move to beginning
  ssfPtr_->setLocation(0);
  ssfPtr_->setLineNumber( 1 );

  TokenVector line;
  std::string empty("");
  while (getLine(line, libSelect, empty)) 
  {
    if (!line.empty() && compare_nocase(line[0].string_.c_str(), ".ends") != 0)
    {
      // parse locally if distool does not distribute
      if( !circuitDeviceLine( line ) )
      {
        handleDeviceLine( line, libSelect, empty );
      }
    }
  }

  // Restore old ssfPtr_ and netlistFilename_.
  ssfPtr_ = oldssfPtr;
  setFileName(old_netlistFilename);

  if (DEBUG_IO)
    Xyce::dout() << "Done with include file Pass 2: " << includeFile << std::endl;

  return true; // Only get here on success.
}


//--------------------------------------------------------------------------
// Function      : DistToolDefault::expandSubcircuitInstance
// Purpose       : Expand a subcircuit instance by adding the devices and
//                 device models that compose the subcircuit to the main
//                 (top level) circuit.
// Special Notes :
// Creator       : Lon Waters
// Creation Date : 12/28/2001
//--------------------------------------------------------------------------
bool DistToolDefault::expandSubcircuitInstance(
  IO::DeviceBlock &     subcircuitInstance,
  const std::string &   libSelect,
  std::string &         libInside)
{

  // Set subcircuitPrefix.
  std::string subcircuitPrefix;
  if ( circuitContext_->getPrefix() != "" )
  {
    subcircuitPrefix = circuitContext_->getPrefix() +
      ":" + subcircuitInstance.getInstanceName().getEncodedName();
  }
  else
  {
    subcircuitPrefix = subcircuitInstance.getInstanceName().getEncodedName();
  }

  // Find the subcircuit corresponding to this instance.
  CircuitBlock* subcircuitPtr = currentCircuitPtr_->findSubcircuit(subcircuitInstance.getModelName());
  if ( subcircuitPtr == NULL )
  {
    Report::UserError0().at(netlistFilename_, subcircuitInstance.getLineNumber())
      << "Subcircuit " << subcircuitInstance.getModelName()
      << " has not been defined for instance " << subcircuitInstance.getInstanceName();
    return false;
  }

  // For hierarchical circuits, set the pointers accordingly. 
  parentCircuitPtr_ = currentCircuitPtr_;
  currentCircuitPtr_ = subcircuitPtr;
  CircuitBlock* oldParentCircuitPtr = parentCircuitPtr_;

  // get the list of X nodes
  const std::vector< std::string >& subcircuitInstanceNodes =
    subcircuitInstance.getNodeValues();

  // Set the context for this subcircuit instance.
  std::string subcircuitName(subcircuitInstance.getModelName());
  bool cresult = circuitContext_->setContext(subcircuitName, subcircuitPrefix, subcircuitInstanceNodes);
  if (!cresult)
  {
    Report::UserError0().at(netlistFilename_, subcircuitInstance.getLineNumber())
      << "Error invoking subcircuit " << subcircuitInstance.getModelName()
      << " instance " << subcircuitInstance.getInstanceName();
    return false;
  }

  // get the list of .SUBCKT nodes
  std::vector<std::string> subcircuitNodes = circuitContext_->getNodeList();

  // Make sure the subcircuit instance and subcircuit definition agree on the
  // number of nodes.
  if ( subcircuitInstanceNodes.size() != subcircuitNodes.size() )
  {
    Report::UserError0() << "Number of nodes for subcircuit instance " << subcircuitInstance.getInstanceName()
                         << " does not agree with number of nodes in subcircuit "
                         << circuitContext_->getCurrentContextName();
    return false;
  }

  // these iterators loop over the nodes in this context (i.e. the real node
  // connected to the interface nodes)
  std::vector<std::string>::const_iterator circuitInstanceNodeIt = subcircuitInstanceNodes.begin();
  std::vector<std::string>::const_iterator endCircuitInstanceNodeIt = subcircuitInstanceNodes.end();

  // these iterators loop over the subcircuits interface nodes
  std::vector<std::string>::iterator subcircuitNodeInterfaceIt = subcircuitNodes.begin();
  std::vector<std::string>::iterator endSubcircuitNodeInterfaceIt = subcircuitNodes.end();

  // add the interface nodes of this subcircuit to the aliasNodeMap so that
  // we can look up the aliases later if needed
  while( ( circuitInstanceNodeIt != endCircuitInstanceNodeIt ) &&
         ( subcircuitNodeInterfaceIt != endSubcircuitNodeInterfaceIt ) )
  {
    std::string key( subcircuitPrefix + ":" + *subcircuitNodeInterfaceIt );
  
    if ( ( mainCircuitPtr_->getAliasNodeMapHelper() ).find( key ) !=
         ( mainCircuitPtr_->getAliasNodeMapHelper() ).end() )
    {
      (mainCircuitPtr_->getAliasNodeMap())[key] = *circuitInstanceNodeIt;

      if (DEBUG_IO)
        Xyce::dout() << "Found node alias:  " << key << " ==> " << *circuitInstanceNodeIt << std::endl;
    }

    circuitInstanceNodeIt++;
    subcircuitNodeInterfaceIt++;
  }

  // Save the current file pointer and netlist filename for later. 
  SpiceSeparatedFieldTool * oldssf = ssfPtr_;
  std::string oldNetlistFilename = netlistFilename_;

  // Locate the subcircuit in the netlist file. It can either be in
  // the file currently being read, or in a separate include file.
  SpiceSeparatedFieldTool * newssf = ssfPtr_;

  if (subcircuitPtr->getNetlistFilename() != netlistFilename_)
  { // The subcircuit is in an include file.
    // Get SSF from Pass 1's ssf map
    if( ssfMap_.count( subcircuitPtr->getNetlistFilename()) )
      newssf = ssfMap_[subcircuitPtr->getNetlistFilename()].second;
    else
    {
      Report::UserError().at(netlistFilename_, subcircuitInstance.getLineNumber())
        << "Can't find include file " << subcircuitPtr->getNetlistFilename();
    }

    netlistFilename_ = subcircuitPtr->getNetlistFilename();
  }

  // Store old positions in the parent netlist file.
  std::streampos oldLoc = oldssf->getFilePosition();
  int oldLine = oldssf->getLineNumber();

  // Set file pointer to location of subcircuit.
  // Set the position of the subcircuit in its file.
  ssfPtr_ = newssf;
  ssfPtr_->setLocation( subcircuitPtr->getStartPosition() );
  ssfPtr_->setLineNumber( subcircuitPtr->getLineStartPosition() );

  // Resolve parameters and functions in the current context.
  bool result;
  std::vector<Device::Param> subcircuitInstanceParams;
  subcircuitInstance.getInstanceParameters(subcircuitInstanceParams);

  result = circuitContext_->resolve(subcircuitInstanceParams);
  if (!result)
    return result;

  // Tell the distribution tool that the context has changed.
  circuitStart( subcircuitName, subcircuitInstanceNodes, subcircuitPrefix,
                subcircuitInstanceParams );

  // Instantiate the devices in this subcircuit instance.
  TokenVector line;

  while (getLine(line, libSelect, libInside))
  {
    if (!line.empty() && compare_nocase(line[0].string_.c_str(), ".ends") != 0)
    {
      // parse locally if distool does not distribute
      if( !circuitDeviceLine( line ) )
      {
        handleDeviceLine( line, libSelect, libInside );
      }
    }
  }

  // send MIs if present
  if( circuitContext_->haveMutualInductances() )
  {
    // Distribute each YMI?!name device line, end of circuit
    int n = circuitContext_->getNumMILines();
    for( int i = 0; i < n; ++i )
    {
      // parse locally if distool does not distribute; normally the
      // DistToolDefault::instantiateDevices() performs this step
      if( !circuitDeviceLine( circuitContext_->getMILine( i ) ) )
      {
        handleDeviceLine( circuitContext_->getMILine( i ), libSelect, libInside );
      }
    }
  }

  // Return to previous context and line numbers.
  circuitContext_->restorePreviousContext();

  // Return file pointer to parent file pointers and location.
  ssfPtr_ = oldssf;
  ssfPtr_->setLocation( oldLoc );
  ssfPtr_->setLineNumber( oldLine );
  netlistFilename_ = oldNetlistFilename;

  // For hierarchical circuits, reset the pointers accordingly. 
  currentCircuitPtr_ = parentCircuitPtr_;
  parentCircuitPtr_ = oldParentCircuitPtr;

  // Tell the distribution tool that the current context has ended and
  // to switch back to the previous context.
  circuitEnd();

  // Only get here on success.
  return true;
}

//----------------------------------------------------------------------------
// Function       : DistToolDefault::instantiateDevice
// Purpose        : Extract data from tokenized device and place all data about
//                  the device in Topology.
// Special Notes  : This method is misnamed as it does not actually instantiate
//                  the device.  That is done after partitioning (if parallel)
//                  and is called from Topology.
//                  This actually fills the DeviceBlock which is stored in
//                  Topology.
// Scope          : private
// Creator        : Lon Waters
// Creation Date  : 07/28/2003
//----------------------------------------------------------------------------
bool DistToolDefault::instantiateDevice(
  DeviceBlock &                                 device,
  const std::string &                           prefix,
  const unordered_map<std::string,std::string> &nodeMap,
  const std::string &                           libSelect,
  std::string &                                 libInside)
{
  bool result = true;

  unordered_map<std::string,std::string>::const_iterator nodeMapIter;
  unordered_map<std::string,std::string>::const_iterator nodeMapEnd = nodeMap.end();

  // Global params were discovered by proc 0 in pass 1 and stored in root level of
  // circuitContext_->  Then circuitContext was transmitted to all procs.

  // Now, before device parsing, move global params from circuitContext_ to deviceMgr

  // This is done here because the global parameters are needed for the parsing of
  // device lines, but are not known until after pass 1.  This point is immediately
  // before device line parsing on all processors.

  // Map device nodes.
  if (!nodeMap.empty())
  {
    //Stats::StatTop _instantiateStat("Determine Device Nodes");
    //Stats::TimeBlock _instantiateTimer(_instantiateStat);

    std::vector<std::string>::iterator nodeIter = device.getNodeValues().begin();
    std::vector<std::string>::iterator nodeIterEnd = device.getNodeValues().end();
    for (; nodeIter != nodeIterEnd; ++nodeIter)
    {
      nodeMapIter = nodeMap.find( *nodeIter );
      if (nodeMapIter != nodeMap.end())
      {
        // The device node is an external node, map it using nodeMap
        *nodeIter = nodeMapIter->second;
      }
      else if ( *nodeIter != "0" && nodeIter->substr(0,2) != "$G" &&
               !circuitContext_->globalNode( *nodeIter ) )
      {
        // The node is internal, prepend subcircuitPrefix. Note: the
        // ground node and global nodes (0 and $G*) are unchanged.
        *nodeIter = std::string( prefix + ":" + *nodeIter );
      }
    }
  }

  // If the device is not a subcircuit instance add its data to the
  // tables, otherwise, expand the instance.
  if (!device.isSubcircuitInstance())
  {
    //Stats::StatTop _instantiateStat("Process Device Data");
    //Stats::TimeBlock _instantiateTimer(_instantiateStat);

    // Once last check to verify that this is device should be instantiated, 
    // now that the node values have been resolved.
    // NOTE:  It is easier to NOT instantiate the device, than remove it from topology later!
    if (remove_any_redundant_)
    {
      if ( IO::removeDevice(preprocessFilter_, device) )
      {
        return result;
      }
    }

    // Map device name.
    if (prefix != "")
      device.setName(prefix + ":" + device.getInstanceName().getEncodedName());

    // If the device has a model, find it and instantiate it (if it has not
    // already been instantiated). Prepend the model name in the device and
    // the model with the model prefix.  Add the model to the circuit.
    if (device.getModelName() != "")
    {
      //Stats::StatTop _modelStat("Map Model Name");
      //Stats::TimeBlock _modelTimer(_modelStat);

      std::string modelPrefix;
      ParameterBlock* modelPtr;

      bool modelFound = circuitContext_->findModel(device.getModelName(), modelPtr, modelPrefix);
      if (modelFound)
      {
        // Add the model prefix to the device's model name.
        if (modelPrefix != "")
        {
          device.setModelName(modelPrefix + ":" + device.getModelName());
        }
      }
      else
      {
        Report::UserError0() << "Unable to find model " << device.getModelName()
                             << " for device " << device.getInstanceName();
        return false;
      }

      // Add the model to the circuit.
      {
        //Stats::StatTop _modelStat2("Add Model Circuit Block");
        //Stats::TimeBlock _modelTimer2(_modelStat2);
        mainCircuitPtr_->addModel(modelPtr, modelPrefix);
      }
    }

    // If the device is a Bsource, the nodes and sources appearing in
    // the expression must be appropriately prefixed.

    // THIS IS AN AWFUL HACK!  It means that *ANY* device that has
    // "SOLN_DEP" parameters must be hacked in here or they can't be used
    // in a subcircuit!

    if ((device.getNetlistDeviceType() == "B" || device.getNetlistDeviceType() == "S" || device.getNetlistDeviceType() == "C")&& (!nodeMap.empty()))
    {
      Device::Param* givenParameter = 0;
      if (device.getNetlistDeviceType() == "B")
      {
        Device::Param* I_parameter = device.findInstanceParameter( "I" );
        Device::Param* V_parameter = device.findInstanceParameter( "V" );
        if ( I_parameter->given() )
          givenParameter = I_parameter;
        else if ( V_parameter->given() )
          givenParameter = V_parameter;
      }
      else
      {
        Device::Param* C_parameter = 0;
        if (device.getNetlistDeviceType() == "S")
        {
            C_parameter = device.findInstanceParameter( "CONTROL" );
        }
        else if (device.getNetlistDeviceType() == "C")
        {
            C_parameter = device.findInstanceParameter( "C" );
        }

        // Note!  The capacitor could be a semiconductor capacitor, and
        // therefore might not even HAVE a C parameter.  Must guard against
        // that.
        if ( C_parameter->given() )
          givenParameter = C_parameter;
        else
          givenParameter = 0;
      }

      if (givenParameter && givenParameter->getType() == Xyce::Util::EXPR)
      {
        Util::Expression &expression = givenParameter->getValue<Util::Expression>();
        if ( (expression.get_num( XEXP_NODE ) > 0) ||
            (expression.get_num( XEXP_INSTANCE ) > 0) ||
            (expression.get_num( XEXP_LEAD ) > 0) )
        {
          // If the expression has nodes or voltage source instances, get
          // the nodes and map them appropriately or add the subcircuit
          // prefix to them.
          std::vector<std::string> names;
          std::vector<std::string> nodes;
          std::vector<std::string> instances;
          std::vector<std::string> leads;
          if ( expression.get_num( XEXP_NODE ) > 0 )
          {
            expression.get_names( XEXP_NODE, nodes );
            names.insert( names.end(), nodes.begin(), nodes.end() );
          }
          if ( expression.get_num( XEXP_INSTANCE ) > 0 )
          {
            expression.get_names( XEXP_INSTANCE, instances);
            names.insert( names.end(), instances.begin(), instances.end() );
          }
          if ( expression.get_num( XEXP_LEAD ) > 0 )
          {
            expression.get_names( XEXP_LEAD, leads);
            names.insert( names.end(), leads.begin(), leads.end() );
          }

          std::vector<std::string> actualName;
          std::string newName, tmpName;
          unsigned int i;

          nodeMapEnd = nodeMap.end();

          for (i = 0; i < names.size(); ++i )
          {
            nodeMapIter = nodeMap.find(names[i]);
            if (nodeMapIter != nodeMapEnd)
            {
              newName = nodeMapIter->second;
            }
            else if (!(i < nodes.size() && ( names[i].substr(0,2) == "$G" ||
                     circuitContext_->globalNode(names[i]))))
            {
              newName = prefix + ":" + names[i];
            }
            else
            {
              // which leaves the one case, where we're a global node $G...,
              // in which case we just keep the original name.
              newName = names[i];
            }

            // Replace the old node name with the new node name
            // in the expression.
            if (names[i] != newName)
            {
              if (DEBUG_IO)
                Xyce::dout() << "Replacing " << names[i] << " with " << newName << std::endl;

              actualName.push_back(newName);
              tmpName = ";" + newName;
              expression.replace_name( names[i], tmpName );
              if (DEBUG_EXPRESSION) {
                Xyce::dout() << "DistToolDefault::instantiateDevice:  After this replacement, get_expression returns "
                             << expression.get_expression() << std::endl
                             << " Parse Tree: " << std::endl;
                expression.dumpParseTree();
              }
            }
          }
          for (i=0 ; i<actualName.size() ; ++i)
          {
            if (DEBUG_IO)
              Xyce::dout() << "Replacing ;"<<actualName[i]<<" with " << actualName[i] << std::endl;

            newName = actualName[i];
            tmpName = ";" + newName;
            expression.replace_name( tmpName, newName);

            if (DEBUG_EXPRESSION) {
              Xyce::dout() << "DistToolDefault::instantiateDevice:  After this replacement get_expression returns "
                           << expression.get_expression() << std::endl;
              expression.dumpParseTree();
            }
          }

          // Reset Bsource's expression.
          if (DEBUG_IO)
            Xyce::dout() << "DistToolDefault::instantiateDevice:  After all expression handling, get_expression returns "
                         << expression.get_expression() << std::endl;
        }
      }
    }

    if (device.getNetlistDeviceType() == "L" && circuitContext_->haveMutualInductances() )
    {
      handleMutualInductance(device);
    }

    // Add device info to the main circuit device and node tables.
    {
      //Stats::StatTop _addTableStat("Add Table Data");
      //Stats::TimeBlock _addTableTimer(_addTableStat);

      circuitBlock_.addTableData(device);
    }
  }
  else
  {
    result = expandSubcircuitInstance(device, libSelect, libInside);
  }

  if (DEBUG_IO)
    device.print();

  return result;
}


//----------------------------------------------------------------------------
// Function       : DistToolDefault::handleMutualInductance
// Purpose        : Post-process the mutual inductors in the current circuit,
//                  each inductor must get the list of inductors it is coupled
//                  to, the coupling coeefficient, and model information.
// Special Notes  :
// Scope          :
// Creator        : Rob Hoekstra
// Creation Date  : 08/28/04
//----------------------------------------------------------------------------
bool DistToolDefault::handleMutualInductance( DeviceBlock & device )
{
  std::string subcircuitPrefix = circuitContext_->getPrefix();

  //Check for mutual inductance assoc. with this inductor
  std::vector<CircuitContext::MutualInductance> & MIs = circuitContext_->getMutualInductances();

  std::string name = device.getInstanceName().getEncodedName();
  std::string::size_type pos = name.find_last_of(":");
  if (pos != std::string::npos)
    name = name.substr( pos + 1, name.length() - (pos + 1));

  for( unsigned int i = 0; i < MIs.size(); ++i)
  {
    if( MIs[i].inductors.count( name ) )
    {
      // add mutual inductance info to this inductor
      if( MIs[i].inductors.begin()->first == name )
      {
        Device::Param param;

        param.setTag( "FIRSTINDUCTOR" );
        param.setGiven(true);
        param.setVal( 1 );
        device.addInstanceParameter( param );
      }

      if( MIs[i].model != "" )
      {
        std::string modelName(MIs[i].model);

        Device::Param param;

        param.setTag( "NONLINEARCOUPLING" );
        param.setVal( 1 );
        param.setGiven(true);
        device.addInstanceParameter( param );

        std::string modelPrefix;
        ParameterBlock* MI_modelPtr;
        bool modelFound = circuitContext_->
              findModel( modelName, MI_modelPtr, modelPrefix );

        if(modelFound)
        {
          // Check whether the current context is dependent on subcircuit
          // parameters (via a "params:" list on the .subckt line. This
          // will affect the model prefixing.
          if (circuitContext_->hasSubcircuitParams() &&
                MI_modelPtr->hasExpressionValuedParams())
          {
            // If there is a subcircuit parameter dependency, we will
            // assume the worst and add the subcircuit instance prefix
            // to the model prefix. This will insure that model parameters
            // that might depend on subcircuit parameters are assigned to
            // a unique model for each subcircuit instance.
            if( subcircuitPrefix != "" )
              modelPrefix = subcircuitPrefix + ":" + modelPrefix;
          }

          // Add the model prefix to the device's model name.
          if( modelPrefix != "" ) modelName = modelPrefix + ":" + modelName;

          device.setModelName( modelName );
        }
        else
        {
          Report::UserError0() << "Unable to find mutual inductor model " << modelName;
          return false;
        }

        // Add the model to the circuit.
        mainCircuitPtr_->addModel(MI_modelPtr, modelPrefix);
      }
      {
        Device::Param param;

        param.setTag( "COUPLING" );
        param.setVal( MIs[i].coupling );
        param.setGiven(true);
        device.addInstanceParameter( param );
      }

      // add the set of inductors to which it is coupled
      std::map<std::string,double>::iterator iterI = MIs[i].inductors.begin();
      std::map<std::string,double>::iterator  endI = MIs[i].inductors.end();
      for( ; iterI != endI; ++iterI )
      {
        std::string ci = iterI->first;

        if( name != ci )
        {
          Device::Param param;

          param.setTag( "COUPLEDINDUCTOR" );
          if( subcircuitPrefix != "" )
          {
            ci = subcircuitPrefix + ":" + ci;
          }

          param.setVal( ci );
          param.setGiven (true);
          device.addInstanceParameter( param );

          param.setTag("COUPLEDINDUCTANCE");

          if (DEBUG_IO)
            Xyce::dout() << "coupledinductance value from " << name << " to " << ci
                         << " is " << iterI->second << std::endl;

          param.setVal( iterI->second);
          param.setGiven (true);
          device.addInstanceParameter( param );
        }
      }
    }
  }

  return true;
}


} // namespace IO
} // namespace Xyce
