//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright(C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename       : $RCSfile: N_IO_OutputterHBPrn.C,v $
//
// Purpose        :
//
// Special Notes  :
//
// Creator        : Dave Baur
//
// Creation Date  :
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.20 $
//
// Revision Date  : $Date: 2015/10/12 22:52:40 $
//
// Current Owner  : $Author: erkeite $
//-------------------------------------------------------------------------

#include <Xyce_config.h>

#include <N_IO_OutputterHBPrn.h>
#include <N_IO_OutputMgr.h>
#include <N_IO_Tecplot.h>
#include <N_IO_Op.h>
#include <N_UTL_DeleteList.h>
#include <N_LAS_BlockVector.h>

namespace Xyce {
namespace IO {
namespace Outputter {

//-----------------------------------------------------------------------------
// Function      : HBPrn::HBPrn
// Purpose       : Outputter for HB runs, standard (PRN) output
//                 format
// Special Notes :
// Creator       : David Baur, Raytheon
// Creation Date : 6/7/2013
//-----------------------------------------------------------------------------
HBPrn::HBPrn(Parallel::Machine comm, OutputMgr &output_manager, const PrintParameters &freq_print_parameters, const PrintParameters &time_print_parameters)
  : outputManager_(output_manager),
    freqPrintParameters_(freq_print_parameters),
    timePrintParameters_(time_print_parameters),
    timeFilename_(),
    freqFilename_(),
    tos_(0),
    fos_(0),
    index_(0),
    currentStep_(0),
    numberOfSteps_(0)
{
  if (timePrintParameters_.defaultExtension_.empty())
    timePrintParameters_.defaultExtension_ = ".HB.TD.prn";

  if (freqPrintParameters_.defaultExtension_.empty())
    freqPrintParameters_.defaultExtension_ = ".HB.FD.prn";

  fixupColumns(comm, outputManager_.getOpBuilderManager(), timePrintParameters_, timeOpList_);
  fixupColumns(comm, outputManager_.getOpBuilderManager(), freqPrintParameters_, freqOpList_);
}

//-----------------------------------------------------------------------------
// Function      : HBPrn::~HBPrn
// Purpose       :
// Special Notes :
// Scope         :
// Creator       : David Baur, Raytheon
// Creation Date :
//-----------------------------------------------------------------------------
HBPrn::~HBPrn()
{
  outputManager_.closeFile(tos_);
  outputManager_.closeFile(fos_);

  deleteList(timeOpList_.begin(), timeOpList_.end());
  deleteList(freqOpList_.begin(), freqOpList_.end());
}

//-----------------------------------------------------------------------------
// Function      : HBPrn::doOutputHB
// Purpose       :
// Special Notes :
// Scope         :
// Creator       : David Baur, Raytheon
// Creation Date :
//-----------------------------------------------------------------------------
void HBPrn::doOutputHB(
  Parallel::Machine             comm,
  const std::vector<double> &   timePoints,
  const std::vector<double> &   freqPoints,
  const Linear::BlockVector &     timeDomainSolutionVec,
  const Linear::BlockVector &     freqDomainSolutionVecReal,
  const Linear::BlockVector &     freqDomainSolutionVecImaginary,
  const Linear::BlockVector &     timeDomainStoreVec,
  const Linear::BlockVector &     freqDomainStoreVecReal,
  const Linear::BlockVector &     freqDomainStoreVecImaginary,
  const Linear::BlockVector &     timeDomainLeadCurrentVec,
  const Linear::BlockVector &     freqDomainLeadCurrentVecReal,
  const Linear::BlockVector &     freqDomainLeadCurrentVecImaginary,
  const Linear::BlockVector &     timeDomainJunctionVoltageVec,
  const Linear::BlockVector &     freqDomainJunctionVoltageVecReal,
  const Linear::BlockVector &     freqDomainJunctionVoltageVecImaginary)
{
  int blockCount = timeDomainSolutionVec.blockCount();

  if (Parallel::rank(comm) == 0 && !fos_ && !tos_)
  {
    timeFilename_ = outputFilename(timePrintParameters_.filename_, timePrintParameters_.defaultExtension_, timePrintParameters_.suffix_, outputManager_.getNetlistFilename());
    freqFilename_ = outputFilename(freqPrintParameters_.filename_, freqPrintParameters_.defaultExtension_, freqPrintParameters_.suffix_, outputManager_.getNetlistFilename());

    tos_ = outputManager_.openFile(timeFilename_);
    fos_ = outputManager_.openFile(freqFilename_);

    if (outputManager_.getPrintHeader())
    {
      printHeader(*tos_, timePrintParameters_);
      printHeader(*fos_, freqPrintParameters_);
    }
  }

  // Loop over the time points of the Linear::BlockVecor:
  for (int iblock = 0; iblock < blockCount; ++iblock)
  {
    outputManager_.setCircuitTime(timePoints[iblock]);
    Linear::Vector * time_solution_vector = &timeDomainSolutionVec.block(iblock);


    Linear::Vector * time_store_vector = &timeDomainStoreVec.block(iblock);
    Linear::Vector * time_lead_current_vector = &timeDomainLeadCurrentVec.block(iblock);

    Linear::Vector * time_junction_voltage_vector = &timeDomainJunctionVoltageVec.block(iblock);

    { // periodic time-domain steady-state output
      std::vector<complex> result_list;
      getValues(comm, timeOpList_, Util::Op::OpData(index_, time_solution_vector, 0, 0, time_store_vector, 0, time_lead_current_vector, 0, time_junction_voltage_vector), result_list);

      for (int i = 0; i < result_list.size(); ++i)
        if (tos_)
          printValue(*tos_, timePrintParameters_.table_.columnList_[i], timePrintParameters_.delimiter_, i, result_list[i].real());
    }

    if (tos_)
      *tos_ << std::endl;

    ++index_;
  }

  int fdblockCount = freqDomainSolutionVecReal.blockCount();
  index_ = 0;
  for (int iblock = 0; iblock < fdblockCount; ++iblock)
  {
    outputManager_.setCircuitFrequency(freqPoints[iblock]);

    Linear::Vector * real_solution_vector = &freqDomainSolutionVecReal.block(iblock);
    Linear::Vector * imaginary_solution_vector = &freqDomainSolutionVecImaginary.block(iblock);
    Linear::Vector * real_store_vector = &freqDomainStoreVecReal.block(iblock);
    Linear::Vector * imaginary_store_vector = &freqDomainStoreVecImaginary.block(iblock);
    Linear::Vector * real_lead_current_vector = &freqDomainLeadCurrentVecReal.block(iblock);
    Linear::Vector * imaginary_lead_current_vector = &freqDomainLeadCurrentVecImaginary.block(iblock);
    Linear::Vector * real_junction_voltage_vector = &freqDomainJunctionVoltageVecReal.block(iblock);
    Linear::Vector * imaginary_junction_voltage_vector = &freqDomainJunctionVoltageVecImaginary.block(iblock);

    { // Fourier coefficient output
      std::vector<complex> result_list;
      // state and store vec are not available in this context, but we must
      // pass in both the real and imaginary vectors
      getValues(comm, freqOpList_, Util::Op::OpData(index_, real_solution_vector, imaginary_solution_vector, 0, real_store_vector, imaginary_store_vector, real_lead_current_vector, imaginary_lead_current_vector, real_junction_voltage_vector, imaginary_junction_voltage_vector ), result_list);

      for (int i = 0; i < result_list.size(); ++i)
        if (fos_)
          printValue(*fos_, freqPrintParameters_.table_.columnList_[i], freqPrintParameters_.delimiter_, i, result_list[i].real());
    }

    if (fos_)
      *fos_ << std::endl;

    ++index_;
  }
}

//-----------------------------------------------------------------------------
// Function      : HBPrn::doFinishOutput
// Purpose       :
// Special Notes :
// Scope         :
// Creator       : David Baur, Raytheon
// Creation Date :
//-----------------------------------------------------------------------------
void HBPrn::doFinishOutput()
{
  if (fos_)
  {
    if (numberOfSteps_ == 0)
    {
      if (outputManager_.getPrintFooter())
      {
        *fos_ << "End of Xyce(TM) Simulation" << std::endl;
      }

      outputManager_.closeFile(fos_);
      fos_ = 0;
    }
  }
  if (tos_)
  {
    if (numberOfSteps_ == 0)
    {
      if (outputManager_.getPrintFooter())
      {
        *tos_ << "End of Xyce(TM) Simulation" << std::endl;
      }

      outputManager_.closeFile(tos_);
      tos_ = 0;
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : HBPrn::doStartStep
// Purpose       :
// Special Notes :
// Scope         :
// Creator       : David Baur, Raytheon
// Creation Date :
//-----------------------------------------------------------------------------
void
HBPrn::doStartStep(
  int                           current_step,
  int                           number_of_steps)
{
  index_ = 0;
  currentStep_ = current_step;
  numberOfSteps_ = number_of_steps;
}

//-----------------------------------------------------------------------------
// Function      : HBPrn::doStartStep
// Purpose       :
// Special Notes :
// Scope         :
// Creator       : David Baur, Raytheon
// Creation Date :
//-----------------------------------------------------------------------------
void
HBPrn::doResetIndex()
{
  index_ = 0;
}

//-----------------------------------------------------------------------------
// Function      : HBPrn::doSteppingComplete
// Purpose       :
// Special Notes :
// Scope         :
// Creator       : David Baur, Raytheon
// Creation Date :
//-----------------------------------------------------------------------------
void HBPrn::doSteppingComplete()
{ 
 if (outputManager_.getPrintFooter()) 
 {
    if (tos_)
      (*tos_) << "End of Xyce(TM) Parameter Sweep" << std::endl;
    if (fos_)
      (*fos_) << "End of Xyce(TM) Parameter Sweep" << std::endl;
  }

  outputManager_.closeFile(tos_);
  tos_ = 0;
  outputManager_.closeFile(fos_);
  fos_ = 0;
}

} // namespace Outputter
} // namespace IO
} // namespace Xyce
