//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_LAS_Builder.h,v $
//
// Purpose        : Builder for LAS objects, hides parallel map stuff
//
// Special Notes  :
//
// Creator        : Robert Hoekstra, SNL, Parallel Computational Sciences
//
// Creation Date  : 01/22/01
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.35 $
//
// Revision Date  : $Date: 2015/05/28 00:05:47 $
//
// Current Owner  : $Author: hkthorn $
//-----------------------------------------------------------------------------

#ifndef  Xyce_LAS_Builder_h
#define  Xyce_LAS_Builder_h


#include <Teuchos_RCP.hpp>
#include <Epetra_Map.h>

#include <N_LAS_fwd.h>

// to eliminate RCP warnings, putting N_PDS_Manager header here.
#include <N_PDS_Manager.h>


class N_PDS_ParMap;

class Epetra_MapColoring;
class Epetra_CrsGraph;

using Teuchos::RCP;
using Teuchos::rcp;

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Class         : Builder
// Purpose       :
// Special Notes :
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 1/22/02
//-----------------------------------------------------------------------------
class Builder
{

public:

  // Default Constructor
  Builder()
  : pdsMgr_(0),
    lasQueryUtil_(0)
  {}

  // Destructor
  virtual ~Builder();

  // Registration methods for necessary utilities
  bool registerPDSManager(N_PDS_Manager * PDS_Manager)
  {
    return (pdsMgr_ = PDS_Manager);
  }

  bool registerQueryUtil(QueryUtil * LAS_QUtil)
  {
    return (lasQueryUtil_ = LAS_QUtil);
  }

  // Vector and Matrix creators which use QueryUtil and N_PDS_ParMap
  // attributes to transparently construct proper objects for this linear
  // system

  // Multivector factory with num vectors and initial value
  virtual MultiVector * createMultiVector( const int numVectors = 1, const double value = 0.0 ) const;
  // State Multivector factory with num vectors and initial value
  virtual MultiVector * createStateMultiVector( const int numVectors = 1, const double value = 0.0 ) const;
  // Store Multivector factory with num vectors and initial value
  virtual MultiVector * createStoreMultiVector( const int numVectors = 1, const double value = 0.0 ) const;
  // Vector factory with initial value
  virtual Vector * createVector( const double value = 0.0 ) const;
  // State-vector factory
  virtual Vector * createStateVector( const double value = 0.0 ) const;
  // Store-vector factory
  virtual Vector * createStoreVector( const double value = 0.0 ) const;
  // LeadCurrent-vector factory
  virtual Vector * createLeadCurrentVector( const double value = 0.0 ) const;

  // Matrix factory
  virtual Matrix * createMatrix( const double initialValue = 0.0 ) const;

  //Coloring Assoc with Variable Types in Solution Vector
  virtual Epetra_MapColoring * createSolnColoring() const;

  //Coloring needed for imposing .IC and .NODESET
  virtual Epetra_MapColoring * createInitialConditionColoring() const;

  virtual bool generateParMaps();

  virtual bool generateGraphs();

  // This is an on-demand capability for the builder to analyze the
  // current solution maps and Jacobian graphs and separate them based
  // on linear vs. nonlinear GID information from topology.
  virtual bool setupSeparatedLSObjects();
  virtual void getSeparatedSolnMap( RCP<N_PDS_ParMap>& linear_map, 
                                    RCP<N_PDS_ParMap>& nonlin_map
                                  ) const;
  virtual void getSeparatedGraph( RCP<Epetra_CrsGraph>& linear_graph,
                                  RCP<Epetra_CrsGraph>& nonlin_graph
                                ) const;

  virtual RCP<const Epetra_Map> getSolutionMap() const;

  virtual void getSolutionMaps(
      RCP<N_PDS_ParMap>& soln_map,
      RCP<N_PDS_ParMap>& soln_ognd_map
      ) const;
  virtual void getStateMap(
      RCP<N_PDS_ParMap>& state_map
      ) const;
  virtual void getStoreMap(
      RCP<N_PDS_ParMap>& store_map
      ) const;
  virtual void getdQdxGraphs(
      RCP<Epetra_CrsGraph>& dQdx_graph,
      RCP<Epetra_CrsGraph>& dQdx_ognd_graph
      ) const;
  virtual void getdFdxGraphs(
      RCP<Epetra_CrsGraph>& dFdx_graph,
      RCP<Epetra_CrsGraph>& dFdx_ognd_graph
      ) const;

private:
  N_PDS_Manager *       pdsMgr_;
  QueryUtil *           lasQueryUtil_;
};

} // namespace Linear
} // namespace Xyce

#endif // Xyce_LAS_Builder_h
