//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_LAS_HBBuilder.C,v $
// Purpose       : 
// Special Notes :
// Creator       : Robert Hoekstra, 9233, Computational Sciences
// Creation Date : 03/12/04
//
// Revision Information:
// ---------------------
// Revision Number: $Revision: 1.21 $
// Revision Date  : $Date: 2015/04/08 19:18:28 $
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <N_LAS_HBBuilder.h>
#include <N_MPDE_Discretization.h>

#include <N_LAS_Vector.h>
#include <N_LAS_BlockVector.h>
#include <N_LAS_BlockMatrix.h>
#include <N_LAS_BlockSystemHelpers.h>

#include <N_PDS_ParMap.h>

#include <N_ERH_ErrorMgr.h>
#include <N_UTL_fwd.h>
#include <N_UTL_FeatureTest.h>

#include <Epetra_Comm.h>
#include <Epetra_Map.h>
#include <Epetra_CrsGraph.h>
#include <Teuchos_OrdinalTraits.hpp>
#include <Teuchos_Utils.hpp>

using Teuchos::rcp;
using Teuchos::RCP;

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Function      : HBBuilder::HBBuilder
// Purpose       : Default constructor 
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, 9233, Computational Sciences
// Creation Date : 03/12/04
//-----------------------------------------------------------------------------
HBBuilder::HBBuilder( const int Size,
                            Teuchos::RCP<const N_MPDE_Discretization> Disc )
: numHarmonics_(Size),
  numSolVariables_(0),
  numStateVariables_(0),
  numStoreVariables_(0),
  blockPattern_(Size),
  Disc_(Disc),
  offset_(0),
  stateOffset_(0),
  storeOffset_(0)
{
  // Create the block graph pattern.
  int Start = Disc_->Start();
  int Width = Disc_->Width();

  for( int i = 0; i < numHarmonics_; ++i )
  {
    blockPattern_[i].resize(Width);
    for( int j = 0; j < Width; ++j )
    {
      int Loc = i+(j+Start);
      if( Loc < 0 )           Loc += numHarmonics_;
      else if( Loc >= numHarmonics_ ) Loc -= numHarmonics_;
      blockPattern_[i][j] = Loc;
    }
    std::sort(blockPattern_[i].begin(),blockPattern_[i].end());
  } 
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, 9233, Computational Sciences
// Creation Date : 03/12/04
//-----------------------------------------------------------------------------
Vector * HBBuilder::createVector( double initialValue ) const
{
  RCP<Vector> vector =
    createExpandedRealFormTransposeBlockVector(); 
  vector.release(); // Release ownership of the object.
  return(&*vector);
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createTimeDomainBlockVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey 1414, Ting Mei 1437
// Creation Date : 9/9/08
//-----------------------------------------------------------------------------
RCP<BlockVector> HBBuilder::createTimeDomainBlockVector( ) const
{
  RCP<BlockVector> vector = rcp(
        new BlockVector( numHarmonics_, HBMap_, BaseMap_ ) 
        );
  return(vector);
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createTimeDomainStateBlockVector( ) 
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey 1414, Ting Mei 1437
// Creation Date : 9/9/08
//-----------------------------------------------------------------------------
RCP<BlockVector> HBBuilder::createTimeDomainStateBlockVector( ) const
{
  RCP<BlockVector> vector = rcp(
        new BlockVector( numHarmonics_, HBStateMap_, BaseStateMap_ ) 
        );
  return(vector);
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createTimeDomainStoreBlockVector( ) 
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 
//-----------------------------------------------------------------------------
RCP<BlockVector> HBBuilder::createTimeDomainStoreBlockVector( ) const
{
  RCP<BlockVector> vector = rcp(
        new BlockVector( numHarmonics_, HBStoreMap_, BaseStoreMap_ ) 
        );
  return(vector);
}


//-----------------------------------------------------------------------------
// Function      : HBBuilder::createTimeDomainLeadCurrentBlockVector( ) 
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 
//-----------------------------------------------------------------------------
RCP<BlockVector> HBBuilder::createTimeDomainLeadCurrentBlockVector() const
{
  RCP<BlockVector> vector = rcp(
        new BlockVector( numHarmonics_, HBLeadCurrentMap_, BaseLeadCurrentMap_) 
        );
  return(vector);
}


//-----------------------------------------------------------------------------
// Function      : HBBuilder::createExpandedRealFormBlockVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey 1414, Rich Schiek 1437
// Creation Date : 9/9/08
//-----------------------------------------------------------------------------
RCP<BlockVector> HBBuilder::createExpandedRealFormBlockVector() const
{
  RCP<BlockVector> vec = rcp(
        new BlockVector( 2*numHarmonics_, HBExpandedRealFormBVMap_, BaseMap_ )
        );
  return(vec);
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createExpandedRealFormTransposeBlockVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey 1414, Rich Schiek 1437
// Creation Date : 9/9/08
//-----------------------------------------------------------------------------
RCP<BlockVector> HBBuilder::createExpandedRealFormTransposeBlockVector() const
{
  RCP<BlockVector> vec = rcp(
        new BlockVector( 2*numHarmonics_, HBExpandedRealFormBVMap_ )
        );
  return(vec);
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createExpandedRealFormTransposeStateBlockVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey 1414, Rich Schiek 1437
// Creation Date : 9/9/08
//-----------------------------------------------------------------------------
RCP<BlockVector> HBBuilder::createExpandedRealFormTransposeStateBlockVector() const
{
  RCP<BlockVector> vec = rcp(
        new BlockVector( 2*numHarmonics_, HBExpandedRealFormStateBVMap_ )
        );
  return(vec);
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createExpandedRealFormTransposeStoreBlockVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 
//-----------------------------------------------------------------------------
RCP<BlockVector> HBBuilder::createExpandedRealFormTransposeStoreBlockVector() const
{
  RCP<BlockVector> vec = rcp(
        new BlockVector( 2*numHarmonics_, HBExpandedRealFormStoreBVMap_ )
        );
  return(vec);
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createExpandedRealFormTransposeLeadCurrentBlockVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 
//-----------------------------------------------------------------------------
RCP<BlockVector> HBBuilder::createExpandedRealFormTransposeLeadCurrentBlockVector() const
{
  RCP<BlockVector> vec = rcp(
        new BlockVector( 2*numHarmonics_, HBExpandedRealFormLeadCurrentBVMap_ )
        );
  return(vec);
}


//-----------------------------------------------------------------------------
// Function      : HBBuilder::createStateVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey, 1414
// Creation Date : 01/17/07
//-----------------------------------------------------------------------------
Vector * HBBuilder::createStateVector( double initialValue ) const
{
  return dynamic_cast<Vector*>(
        new BlockVector( numHarmonics_, HBStateMap_, BaseStateMap_ ) );
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createStoreVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 
//-----------------------------------------------------------------------------
Vector * HBBuilder::createStoreVector( double initialValue ) const
{
  return dynamic_cast<Vector*>(
        new BlockVector( numHarmonics_, HBStoreMap_, BaseStoreMap_ ) );
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createLeadCurrentVector
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 
//-----------------------------------------------------------------------------
Vector * HBBuilder::createLeadCurrentVector( double initialValue ) const
{
  return dynamic_cast<Vector*>(
        new BlockVector( numHarmonics_, HBLeadCurrentMap_, BaseLeadCurrentMap_ ) );
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createDAEdQdxMatrix
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, 9233, Computational Sciences
// Creation Date : 03/12/04
//-----------------------------------------------------------------------------
Matrix * HBBuilder::createDAEdQdxMatrix( double initialValue ) const
{
  
  /*  NOTE:  The dQdx matrix is created with the same graph as the dFdx so
             that the construction of the Jacobian is faster */
  return createDAEdFdxMatrix( initialValue );
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createDAEdFdxMatrix
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, 9233, Computational Sciences
// Creation Date : 03/12/04
//-----------------------------------------------------------------------------
Matrix * HBBuilder::createDAEdFdxMatrix( double initialValue ) const
{
  return dynamic_cast<Matrix*>(
        new BlockMatrix( numHarmonics_,
                               offset_,
                               blockPattern_,
                               *HBFullGraph_,
                               *BaseFullGraph_ ) );
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::createDAEFullMatrix
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, 9233, Computational Sciences
// Creation Date : 03/12/04
//-----------------------------------------------------------------------------
Matrix * HBBuilder::createDAEFullMatrix( double initialValue ) const
{
  return createDAEdFdxMatrix( initialValue );
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::generateMaps
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, 9233, Computational Sciences
// Creation Date : 03/12/04
//-----------------------------------------------------------------------------
bool HBBuilder::generateMaps( const RCP<N_PDS_ParMap>& BaseMap, 
                                    const RCP<N_PDS_ParMap>& oBaseMap )
{
  //Save copy of base map
  BaseMap_ = BaseMap;
  //Xyce::dout() << "BaseMap" << std::endl;
  //BaseMap->petraMap()->Print(Xyce::dout());
  oBaseMap_ = oBaseMap;
  //Xyce::dout() << "oBaseMap" << std::endl;
  //oBaseMap->petraMap()->Print(Xyce::dout());

  //Determine block offset
  offset_ = generateOffset( *BaseMap );

  //Check to see if a graph can be generated using 32-bit integers for the 
  //specified number of harmonics.
  int MaxGID = BaseMap->maxGlobalEntity();
  long int hbMaxGID = MaxGID + offset_*(numHarmonics_-1);
  if ( hbMaxGID > Teuchos::OrdinalTraits<int>::max() )
  {
      int allowableHarmonics = (Teuchos::OrdinalTraits<int>::max() - MaxGID + offset_)/offset_ - 1;
      std::string msg = "HBBuilder::generateMaps():  Harmonic Balance map cannot be constructed, reduce number of harmonics to " 
                      + Teuchos::Utils::toString( allowableHarmonics ) + ".";
      N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::USR_ERROR_0, msg);
  }                          

  // Use the block linear system helper to create the block parallel maps
  std::vector<RCP<N_PDS_ParMap> > blockMaps = createBlockParMaps(numHarmonics_, *BaseMap, *oBaseMap);

  HBMap_ = blockMaps[0];
  oHBMap_ = blockMaps[1];
  
  //Xyce::dout() << "HBMap_" << std::endl;
  //Xyce::dout() << *HBMap_ << std::endl; 
  //Xyce::dout() << "oHBMap_" << std::endl;
  //Xyce::dout() << *oHBMap_ << std::endl;
  //Save copy of base map

  // Helpful names for various sizes (subtract 1 for ground node):
  numSolVariables_ = oBaseMap_->numLocalEntities()-1;

  // Expanded Real Form for Block Vector map:
  HBExpandedRealFormBVMap_ = createBlockFreqERFParMap( numHarmonics_, *BaseMap );

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::generateStateMaps
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey, 1414
// Creation Date : 01/17/07
//-----------------------------------------------------------------------------
bool HBBuilder::generateStateMaps( const RCP<N_PDS_ParMap>& BaseStateMap )
{
  //Save copy of base map
  BaseStateMap_ = BaseStateMap;

  //determine block offset
  stateOffset_ = generateOffset( *BaseStateMap );

  // Use the block linear system helper to create the block parallel maps
  // NOTE:  At this time augmented parallel maps are not supported.
  HBStateMap_ = createBlockParMap(numHarmonics_, *BaseStateMap);

  // Helpful names for various sizes:
  numStateVariables_ = BaseStateMap_->numGlobalEntities();

  // Expanded Real Form for Block Vector map:
  HBExpandedRealFormStateBVMap_ = createBlockFreqERFParMap( numHarmonics_, *BaseStateMap );

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::generateStoreMaps
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 
//-----------------------------------------------------------------------------
bool HBBuilder::generateStoreMaps( const RCP<N_PDS_ParMap>& BaseStoreMap )
{
  //Save copy of base map
  BaseStoreMap_ = BaseStoreMap;

  //determine block offset
  storeOffset_ = generateOffset( *BaseStoreMap );

  // Use the block linear system helper to create the block parallel maps
  // NOTE:  At this time augmented parallel maps are not supported.
  HBStoreMap_ = createBlockParMap(numHarmonics_, *BaseStoreMap);

  // Helpful names for various sizes:
  numStoreVariables_ = BaseStoreMap_->numGlobalEntities();

  // Expanded Real Form for Block Vector map:
  HBExpandedRealFormStoreBVMap_ = createBlockFreqERFParMap( numHarmonics_, *BaseStoreMap );

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::generateLeadCurrentMaps
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 
//-----------------------------------------------------------------------------
bool HBBuilder::generateLeadCurrentMaps( const RCP<N_PDS_ParMap>& BaseLeadCurrentMap )
{
  //Save copy of base map
  BaseLeadCurrentMap_ = BaseLeadCurrentMap;

  //determine block offset
  leadCurrentOffset_ = generateOffset( *BaseLeadCurrentMap );

  // Use the block linear system helper to create the block parallel maps
  // NOTE:  At this time augmented parallel maps are not supported.
  HBLeadCurrentMap_ = createBlockParMap(numHarmonics_, *BaseLeadCurrentMap);

  // Helpful names for various sizes:
  numLeadCurrentVariables_ = BaseLeadCurrentMap_->numGlobalEntities();

  // Expanded Real Form for Block Vector map:
  HBExpandedRealFormLeadCurrentBVMap_ = createBlockFreqERFParMap( numHarmonics_, *BaseLeadCurrentMap );

  return true;
}
//-----------------------------------------------------------------------------
// Function      : HBBuilder::generateGraphs
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, 9233, Computational Sciences
// Creation Date : 03/12/04
//-----------------------------------------------------------------------------
bool HBBuilder::generateGraphs( const Epetra_CrsGraph & BaseFullGraph )
{
  if( Teuchos::is_null(BaseMap_) )
    Xyce::Report::DevelFatal0().in("HBBuilder::generateGraphs")
      << "Need to setup Maps first";

  //Copies of base graphs
  BaseFullGraph_ = rcp(new Epetra_CrsGraph( BaseFullGraph ));

  int BlockSize = BaseMap_->numLocalEntities();

  //Check to see if a graph can be generated using 32-bit integers for the
  //specified number of harmonics.
  int MaxGID = BaseMap_->maxGlobalEntity();
  long int hbMaxGID = MaxGID + offset_*(numHarmonics_-1);
  if ( hbMaxGID > Teuchos::OrdinalTraits<int>::max() )
  {
      int allowableHarmonics = (Teuchos::OrdinalTraits<int>::max() - MaxGID + offset_)/offset_ - 1;
      std::string msg = "HBBuilder::generateGraphs():  Harmonic Balance graph cannot be constructed, reduce number of harmonics to "
                      + Teuchos::Utils::toString( allowableHarmonics ) + ".";
      N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::USR_ERROR_0, msg);
  }

  //Construct HB Graph [All graphs are the same, so only one needs to be made]
  HBFullGraph_ = rcp(new Epetra_CrsGraph( Copy,
                                          dynamic_cast<Epetra_BlockMap&>(*(HBMap_->petraMap())),
                                          0 ));
  
  int MaxIndices = BaseFullGraph_->MaxNumIndices();
  std::vector<int> Indices(MaxIndices);
  int NumIndices=0, BaseRow=0, HBRow=0;

  for( int i = 0; i < numHarmonics_; ++i )
  {
    for( int j = 0; j < BlockSize; ++j )
    {
      BaseRow = BaseMap_->petraMap()->GID(j);
      BaseFullGraph.ExtractGlobalRowCopy( BaseRow, MaxIndices, NumIndices, &Indices[0] );
      for( int k = 0; k < NumIndices; ++k ) Indices[k] += offset_*i;
      //Diagonal Block
      HBRow = BaseRow + offset_*i;
      // Insert in F, as its the default graph for the full Jacobian matrix
      HBFullGraph_->InsertGlobalIndices( HBRow, NumIndices, &Indices[0] );
    }
  }
  
  if (DEBUG_HB)
  {
    Xyce::dout() << "HBBuilder::generateGraphs():  HB full graph after inserting diagonal blocks:" << std::endl;
    HBFullGraph_->Print(Xyce::dout());
  }
 
  MaxIndices = BaseFullGraph_->MaxNumIndices();
  Indices.resize(MaxIndices);
  int Shift=0, Index=0;
  int DiscWidth = Disc_->Width();
  std::vector<int> NewIndices(MaxIndices*DiscWidth);
  
  for( int j = 0; j < BlockSize; ++j )
  {
    // Extract the base entries from the base row.
    BaseRow = BaseMap_->petraMap()->GID(j);
    BaseFullGraph.ExtractGlobalRowCopy( BaseRow, MaxIndices, NumIndices, &Indices[0] );
  
    for( int i = 0; i < numHarmonics_; ++i )
    {
      // For this harmonic, which row will be inserted.
      HBRow = BaseRow + offset_*i;
      
      // Find all entries from a row before inserting it.
      for( int k = 0; k < DiscWidth; ++k )
      {
        // Find which block column to start at.
        Shift = blockPattern_[i][k]*offset_;  // Actual column index.
        Index = k*NumIndices;  // Pointer to next block of column indices.
        for( int kk = 0; kk < NumIndices; ++kk ) NewIndices[Index+kk] = Indices[kk] + Shift;
      }

      // Insert entire row for all blocks.
      HBFullGraph_->InsertGlobalIndices( HBRow, DiscWidth*NumIndices, &NewIndices[0] );
    }
  }
  HBFullGraph_->FillComplete();

  if (DEBUG_HB)
  {
    Xyce::dout() << "HBBuilder::generateGraphs():  Final HB Graph = " << std::endl;
    HBFullGraph_->Print(Xyce::dout());
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBuilder::getSolutionMap
// Purpose       : 
// Special Notes : This is specifically for blockAnalysis types (like MPDE & HB)
// so we can get a valid map from the builder.
// Scope         : Public
// Creator       : Todd Coffey, 1414
// Creation Date : 09/05/08
//-----------------------------------------------------------------------------
RCP<const Epetra_Map> HBBuilder::getSolutionMap() const
{
    return(rcp(HBExpandedRealFormBVMap_->petraMap(),false));
}

} // namespace Linear
} // namespace Xyce
