//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_LOA_CktLoader.h,v $
//
// Purpose        : This file contains class definitions for the loader
//                  services package.
//
// Special Notes  :
//
// Creator        : Eric R. Keiter, SNL, Parallel Computational Sciences
//
// Creation Date  : 05/26/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.127 $
//
// Revision Date  : $Date: 2015/07/11 23:40:29 $
//
// Current Owner  : $Author: erkeite $
//-----------------------------------------------------------------------------

#ifndef Xyce_LOA_CktLoader_H
#define Xyce_LOA_CktLoader_H

#include <N_LAS_fwd.h>
#include <N_PDS_fwd.h>
#include <N_UTL_fwd.h>

#include <N_LOA_Loader.h>

namespace Xyce {
namespace Loader {

//-----------------------------------------------------------------------------
// Class         : CktLoader
// Purpose       :
// Special Notes :
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 5/03/00
//-----------------------------------------------------------------------------
class CktLoader : public Loader
{
public:
  CktLoader(
    Device::DeviceMgr & device_manager);

  virtual ~CktLoader();

  // Method which is called to load the new-DAE contributions to the nonlinear Jacobian matrix.
  bool loadDAEMatrices(
    Linear::Vector *    tmpSolVectorPtr,
    Linear::Vector *    tmpStaVectorPtr,
    Linear::Vector *    tmpStaDerivVectorPtr,
    Linear::Vector *    tmpStoVectorPtr,
    Linear::Matrix *    tmpdQdxMatrixPtr,
    Linear::Matrix *    tmpdFdxMatrixPtr);

  // Method which is called to load the new-DAE vectors, which contribute
  // to the nonlinear residual (RHS) vector.
  bool loadDAEVectors(
    Linear::Vector * nextSolVectorPtr,
    Linear::Vector * currSolVectorPtr,
    Linear::Vector * lastSolVectorPtr,
    Linear::Vector * nextStaVectorPtr,
    Linear::Vector * currStaVectorPtr,
    Linear::Vector * lastStaVectorPtr,
    Linear::Vector * StaDerivVectorPtr,
    Linear::Vector * nextStoVectorPtr,
    Linear::Vector * currStoVectorPtr,
    Linear::Vector * lastStoVectorPtr,
    Linear::Vector * stoLeadCurrQVectorPtr,
    Linear::Vector * nextLeadFVectorPtr,
    Linear::Vector * currLeadFVectorPtr,
    Linear::Vector * lastLeadFVectorPtr,
    Linear::Vector * nextLeadQVectorPtr,
    Linear::Vector * nextJunctionVVectorPtr,
    Linear::Vector * currJunctionVVectorPtr,
    Linear::Vector * lastJunctionVVectorPtr,
    Linear::Vector * QVectorPtr,
    Linear::Vector * FVectorPtr,
    Linear::Vector * BVectorPtr,
    Linear::Vector * dFdxdVpVectorPtr,
    Linear::Vector * dQdxdVpVectorPtr);

  // Method is called to load the mask to be used in calculating error norms.
  bool loadDeviceErrorWeightMask(Linear::Vector * deviceMask) const;

  // Initializes the nonlinear problem.
  bool initializeProblem(
    Linear::Vector * nextSolVectorPtr,
    Linear::Vector * currSolVectorPtr,
    Linear::Vector * lastSolVectorPtr,
    Linear::Vector * nextStaVectorPtr,
    Linear::Vector * currStaVectorPtr,
    Linear::Vector * lastStaVectorPtr,
    Linear::Vector * StateDerivVectorPtr,
    Linear::Vector * nextStoVectorPtr,
    Linear::Vector * currStoVectorPtr,
    Linear::Vector * lastStoVectorPtr,
    Linear::Vector * QVectorPtr,
    Linear::Vector * FVectorPtr,
    Linear::Vector * BVectorPtr,
    Linear::Vector * dFdxdVpVectorPtr,
    Linear::Vector * dQdxdVpVectorPtr) const;

  bool updateState(
    Linear::Vector * nextSolVectorPtr,
    Linear::Vector * currSolVectorPtr,
    Linear::Vector * lastSolVectorPtr,
    Linear::Vector * nextStaVectorPtr,
    Linear::Vector * currStaVectorPtr,
    Linear::Vector * lastStaVectorPtr,
    Linear::Vector * nextStoVectorPtr,
    Linear::Vector * currStoVectorPtr,
    Linear::Vector * lastStoVectorPtr);

  bool loadBVectorsforAC (Linear::Vector * bVecRealPtr,
                          Linear::Vector * bVecImagPtr);

  int getNumNoiseSources();

  int getNumNoiseDevices();

  void setupNoiseSources(std::vector<Xyce::Analysis::NoiseData*> & noiseDataVec);

  void getNoiseSources(std::vector<Xyce::Analysis::NoiseData*> & noiseDataVec);

  bool getBMatrixEntriesforMOR(std::vector<int>& bMatEntriesVec, std::vector<int>& bMatPosEntriesVec);

  // Function for setting the initial guess.
  bool setInitialGuess(Linear::Vector * solVectorPtr);

  // Function for setting a single parameter value.
  bool setParam(std::string & name, double val, bool overrideOriginal=false);

  // Function for getting a single parameter value.
  virtual double getParamAndReduce(Parallel::Machine comm, const std::string & name) const;

  // Method which is called to update the sources.
  bool updateSources();

  // Get the voltage limiter flag:
  bool getLimiterFlag ();

  // Gets the double DC Operating Point flag - used for PDE devices.
  bool isPDESystem() const;
  bool outputPlotFiles() const;
  bool finishOutput() const;

  // two-level newton functions:
  int  enablePDEContinuation();
  bool disablePDEContinuation ();

  void getNumInterfaceNodes (std::vector<int> & numINodes);
  bool loadCouplingRHS(int iSubProblem, int iCouple, Linear::Vector * dfdvPtr);
  bool calcCouplingTerms (int iSubProblem, int iCouple, const Linear::Vector * dxdvPtr);
  // Gets the time integration required breakpoint times (in a vector).
  bool getBreakPoints(std::vector< Util::BreakPoint > & breakPointTimes) const;

  // Accessor which returns the maximum time step size (in seconds).
  double getMaxTimeStepSize();

  // Get block size from device options for block gainscale homotopy
  int getHomotopyBlockSize() const;

  // Get convergence info from devices
  bool allDevicesConverged(Parallel::Machine comm);

  // Get convergence info from inner-solves
  bool innerDevicesConverged(Parallel::Machine comm);

  void stepSuccess(Xyce::Analysis::TwoLevelMode analysis);
  void stepFailure(Xyce::Analysis::TwoLevelMode analysis);

  void acceptStep();

  virtual bool getInitialQnorm (std::vector<TimeIntg::TwoLevelError> & tleVec );

  virtual bool getInnerLoopErrorSums (std::vector<TimeIntg::TwoLevelError> & tleVec) const;

  bool updateStateArrays ();
  bool startTimeStep(
    bool                          beginIntegrationFlag,
    double                        nextTimeStep,
    double                        nextTime,
    int                           currentOrder);
  void setExternalSolverState(bool external_initJctFlag);

    // Function for determining if an analytic sensitivity (df/dp or dqdp) is available.
  virtual bool analyticSensitivitiesAvailable(std::string & name);

  virtual void getAnalyticSensitivities(
    std::string & name, 
    std::vector<double> & dfdpVec, 
    std::vector<double> & dqdpVec,
    std::vector<double> & dbdpVec,
    std::vector<int> & FindicesVec,
    std::vector<int> & QindicesVec,
    std::vector<int> & BindicesVec) const;

  // voltage limiter toggle functions
  virtual bool getVoltageLimiterStatus();
  virtual void setVoltageLimiterStatus(bool voltageLimterStatus);

public:
  Device::DeviceMgr &   deviceManager_;         ///< Device manager
};

} // namespace Loader
} // namespace Xyce

#endif
