//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_LOA_HBLoader.h,v $
//
// Purpose        : HB Specific Loader
//
// Special Notes  :
//
// Creator        : Todd Coffey, Ting Mei
//
// Creation Date  : 07/28/08
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.47 $
//
// Revision Date  : $Date: 2015/08/31 20:12:23 $
//
// Current Owner  : $Author: tmei $
//-----------------------------------------------------------------------------

#ifndef Xyce_LOA_HBLoader_H
#define Xyce_LOA_HBLoader_H

#include <vector>

#include <Teuchos_RCP.hpp>

#include <N_DEV_fwd.h>
#include <N_LAS_fwd.h>
#include <N_PDS_fwd.h>

#include <N_UTL_DFTInterfaceDecl.hpp>
#include <N_LOA_CktLoader.h>

// ---------- Forward declarations --------

class N_MPDE_Discretization;

namespace Xyce {
namespace Loader {

//-----------------------------------------------------------------------------
// Class         : HBLoader
// Purpose       : HB specific CktLoader interface
// Special Notes :
// Creator       : Todd Coffey, Ting Mei, Rich Schiek
// Creation Date : 07/29/08
//-----------------------------------------------------------------------------
class HBLoader : public CktLoader
{
public:
  HBLoader(
    const Teuchos::RCP<const N_MPDE_Discretization> discPtr,
    Device::DeviceMgr &                 device_manager,
    Linear::Builder &                   builder);

  ~HBLoader()
  {}

  // Method which is called to load the new-DAE contributions 
  bool loadDAEMatrices( Linear::Vector * X,
                        Linear::Vector * S,
                        Linear::Vector * dSdt,
                        Linear::Vector * Store,
                        Linear::Matrix * dQdx,
                        Linear::Matrix * dFdx );

  // Method which is called to load the time-dependent HB Matrices
  bool loadTimeDepDAEMatrices( Linear::Vector * X,
                        Linear::Vector * S,
                        Linear::Vector * dSdt,
                        Linear::Vector * Store,
                        Linear::Matrix * dQdx,
                        Linear::Matrix * dFdx );

  // Method for matrix-free 
  bool applyDAEMatrices( Linear::Vector * X,
                         Linear::Vector * S,
                         Linear::Vector * dSdt,
                         Linear::Vector * Store,
                         const Linear::Vector & V,
                         Linear::Vector * dQdxV,
                         Linear::Vector * dFdxV );

  // Method is called to load the mask to be used in calculating error norms.
  bool loadDeviceErrorWeightMask(Linear::Vector * deviceMask) const;

  // Method which is called to load the new-DAE vectors
  bool loadDAEVectors( Linear::Vector * X,
                       Linear::Vector * currX,
                       Linear::Vector * lastX,
                       Linear::Vector * S,
                       Linear::Vector * currS,
                       Linear::Vector * lastS,
                       Linear::Vector * dSdt,
                       Linear::Vector * Store,
                       Linear::Vector * currStore,
                       Linear::Vector * lastStore,
                       Linear::Vector * storeLeadCurrQ,
                       Linear::Vector * nextLeadFVectorPtr,
                       Linear::Vector * currLeadFVectorPtr,
                       Linear::Vector * lastLeadFVectorPtr,
                       Linear::Vector * nextLeadQVectorPtr,
                       Linear::Vector * nextJunctionVVectorPtr,
                       Linear::Vector * currJunctionVVectorPtr,
                       Linear::Vector * lastJunctionVVectorPtr,
                       Linear::Vector * Q,
                       Linear::Vector * F,
                       Linear::Vector * B,
                       Linear::Vector * dFdxdVp,
                       Linear::Vector * dQdxdVp );

  bool updateState(    Linear::Vector * nextSolVectorPtr,
                       Linear::Vector * currSolVectorPtr,
                       Linear::Vector * lastSolVectorPtr,
                       Linear::Vector * nextStaVectorPtr,
                       Linear::Vector * currStaVectorPtr,
                       Linear::Vector * lastStaVectorPtr,
                       Linear::Vector * nextStoVectorPtr,
                       Linear::Vector * currStoVectorPtr,
                       Linear::Vector * lastStoVectorPtr
                       );

    // Virtual method which initializes the nonlinear problem.
  virtual bool initializeProblem( Linear::Vector * nextSolVectorPtr,
                          Linear::Vector * currSolVectorPtr,
                          Linear::Vector * lastSolVectorPtr,
                          Linear::Vector * nextStaVectorPtr,
                          Linear::Vector * currStaVectorPtr,
                          Linear::Vector * lastStaVectorPtr,
                          Linear::Vector * StateDerivVectorPtr,
                          Linear::Vector * nextStoVectorPtr,
                          Linear::Vector * currStoVectorPtr,
                          Linear::Vector * lastStoVectorPtr,
                          Linear::Vector * QVectorPtr,
                          Linear::Vector * FVectorPtr,
                          Linear::Vector * BVectorPtr,
                          Linear::Vector * dFdxdVpVectorPtr,
                          Linear::Vector * dQdxdVpVectorPtr) const
  {
    return false;
  }

  // Get the voltage limiter flag:
  bool getLimiterFlag () { return HBLoader::appLoaderPtr_->getLimiterFlag (); }

  // Get the stored time-domain Jacobians from the HB loader.
  Teuchos::RCP<Linear::Matrix>& getStoreLindQdx() { return linAppdQdxPtr_; }
  Teuchos::RCP<Linear::Matrix>& getStoreLindFdx() { return linAppdFdxPtr_; }
  std::vector<Teuchos::RCP<Linear::Matrix> >& getStoreNLdQdx() { return vecNLAppdQdxPtr_; }
  std::vector<Teuchos::RCP<Linear::Matrix> >& getStoreNLdFdx() { return vecNLAppdFdxPtr_; }

  bool applyLinearMatrices( const Linear::BlockVector & bVf,
                            Linear::BlockVector & permlindQdxV,
                            Linear::BlockVector & permlindFdxV );

  // New functions for LOA_HBLoader
  // Assumption:  xt is in the block format:
  // xt.block(i) = { x_n(t_i) }_{n=0..N} N = number of solution components, t_i = time point i
  // P takes xt's block format and converts it to: 
  // (P*xt).block(n) = { x_n(t_i) }_{i=0..T} T = number of time points in block vector.

   
  void setFastTimes( const std::vector<double> & times );

  void setHBFreqs( const std::vector<double> & freqs); 
  Teuchos::RCP<Linear::BlockVector> & getStoreVecFreqPtr()  { return bStoreVecFreqPtr_;} 
  Teuchos::RCP<Linear::BlockVector> & getLeadCurrentVecFreqPtr()  { return bLeadCurrentVecFreqPtr_;} 

  // Is the problem being applied as a matrix free operator.
  void setMatrixFreeFlag(bool matrixFreeFlag)
  {
    matrixFreeFlag_ = matrixFreeFlag;
  }
 
  // xf = D*P*xt, xf has the same block format as (P*xt).
  void permutedFFT(const Linear::BlockVector & xt, Linear::BlockVector * xf); // Reference for input, pointer for output
  // xt = P^{-1}D^{-1}*xf
  void permutedIFT(const Linear::BlockVector & xf, Linear::BlockVector * xt, int numTimePts_= 0); 

  // Registration method for the device packaage
  void registerAppLoader( Teuchos::RCP<Loader> appLoaderPtr )
  { appLoaderPtr_ = appLoaderPtr; }

  void registerHBBuilder(Teuchos::RCP<Linear::HBBuilder> hbBuilderPtr);

  void registerDFTInterface( const Teuchos::RCP<N_UTL_DFTInterfaceDecl<std::vector<double> > >& dftInterface )
  { dftInterface_ = dftInterface; }

  virtual bool analyticSensitivitiesAvailable (std::string & name) { return false; }
  virtual void getAnalyticSensitivities(
      std::string & name, 
      std::vector<double> & dfdpVec, 
      std::vector<double> & dqdpVec,
      std::vector<double> & dbdpVec,
      std::vector<int> & FindicesVec,
      std::vector<int> & QindicesVec,
      std::vector<int> & BindicesVec) const
  {}

  virtual bool setParam (std::string & name, double val, bool overrideOriginal=false) { return false; }
  virtual double getParamAndReduce(Parallel::Machine comm, const std::string & name) const { return 0.0; }

  // voltage limiter toggle functions
  virtual bool getVoltageLimiterStatus();
  virtual void setVoltageLimiterStatus(bool voltageLimterStatus);

private:
  // discretization
  Teuchos::RCP<const N_MPDE_Discretization> fastTimeDiscPtr_;

  //Fast Time Scale Points
  std::vector<double> times_;
  int periodicTimesOffset_;
  std::vector<double> periodicTimes_;
  double period_;
 
  // Matrix free flag, operator is being applied not loaded
  bool matrixFreeFlag_;

  std::vector<double> freqs_ ;
  // Base Application loader
  Teuchos::RCP<Loader>          appLoaderPtr_;          ///< Actually a CktLoader
  Device::DeviceMgr &           deviceManager_;

  // Application Linear Objects
  Teuchos::RCP<Linear::Vector> appVecPtr_;
  Teuchos::RCP<Linear::Vector> appNextStaVecPtr_;
  Teuchos::RCP<Linear::Vector> appCurrStaVecPtr_;
  Teuchos::RCP<Linear::Vector> appLastStaVecPtr_;

  Teuchos::RCP<Linear::Matrix> appdQdxPtr_;
  Teuchos::RCP<Linear::Matrix> appdFdxPtr_;

  Teuchos::RCP<Linear::Matrix> linAppdQdxPtr_;
  std::vector<Teuchos::RCP<Linear::Matrix> > vecNLAppdQdxPtr_;
  Teuchos::RCP<Linear::Matrix> linAppdFdxPtr_;
  std::vector<Teuchos::RCP<Linear::Matrix> > vecNLAppdFdxPtr_;
  
  Teuchos::RCP<Linear::Vector> appNextStoVecPtr_;
  Teuchos::RCP<Linear::Vector> appCurrStoVecPtr_;
  Teuchos::RCP<Linear::Vector> appLastStoVecPtr_;
  Teuchos::RCP<Linear::Vector> appStoLeadCurrQVecPtr_;
  
  Teuchos::RCP<Linear::Vector> appNextLeadFVecPtr_;
  Teuchos::RCP<Linear::Vector> appCurrLeadFVecPtr_;
  Teuchos::RCP<Linear::Vector> appLastLeadFVecPtr_;
  Teuchos::RCP<Linear::Vector> appLeadQVecPtr_;
  Teuchos::RCP<Linear::Vector> appNextJunctionVVecPtr_;
  Teuchos::RCP<Linear::Vector> appCurrJunctionVVecPtr_;
  Teuchos::RCP<Linear::Vector> appLastJunctionVVecPtr_;

  // Tmp storage block matrices 
  Teuchos::RCP<Linear::BlockMatrix> bmdQdxPtr_;
  Teuchos::RCP<Linear::BlockMatrix> bmdFdxPtr_;

  // HB Builder:  (needed to convert AztecOO created Linear::Vectors into Linear::BlockVectors
  Teuchos::RCP<Linear::HBBuilder> hbBuilderPtr_;

  // App Builder:  (needed to load time domain vectors and matrices)
  Linear::Builder &             builder_;

  Teuchos::RCP<Linear::BlockVector> bXtPtr_;
  Teuchos::RCP<Linear::BlockVector> bVtPtr_;

  Teuchos::RCP<Linear::BlockVector> bStoreVecFreqPtr_;
  Teuchos::RCP<Linear::BlockVector> bStoreLeadCurrQVecFreqPtr_; 
  Teuchos::RCP<Linear::BlockVector> bLeadCurrentVecFreqPtr_; 
  Teuchos::RCP<Linear::BlockVector> bLeadCurrentQVecFreqPtr_; 

  // DFT Interface
  Teuchos::RCP<N_UTL_DFTInterfaceDecl<std::vector<double> > > dftInterface_;
};

} // namespace Loader
} // namespace Xyce

#endif // Xyce_LOA_HBLoader_H
