//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename       : $RCSfile: N_NLS_Sensitivity.C,v $
//
// Purpose        : Body for the sensitivity class.
//
// Special Notes  :
//
// Creator        : Eric R. Keiter, SNL, Parallel Computational Sciences
//
// Creation Date  : 10/30/02
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.141 $
//
// Revision Date  : $Date $
//
// Current Owner  : $Author: erkeite $
//-------------------------------------------------------------------------

#include <Xyce_config.h>


// ----------   Standard Includes   ----------

#include <algorithm>
#include <sstream>
#include <stdexcept>

// ----------   Xyce Includes   ----------

#include <N_ANP_AnalysisManager.h>
#include <N_ERH_ErrorMgr.h>
#include <N_IO_CircuitBlock.h>
#include <N_IO_OptionBlock.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_SpiceSeparatedFieldTool.h>
#include <N_LAS_Builder.h>
#include <N_LAS_Matrix.h>
#include <N_LAS_Problem.h>
#include <N_LAS_Solver.h>
#include <N_LAS_System.h>
#include <N_LAS_Vector.h>
#include <N_LOA_NonlinearEquationLoader.h>
#include <N_NLS_Manager.h>
#include <N_NLS_Sensitivity.h>
#include <N_PDS_Comm.h>
#include <N_PDS_MPI.h>
#include <N_PDS_Manager.h>
#include <N_PDS_Serial.h>
#include <N_TIA_DataStore.h>
#include <N_TOP_Topology.h>
#include <N_UTL_Algorithm.h>
#include <N_UTL_Diagnostic.h>
#include <N_UTL_Expression.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_OptionBlock.h>

// ----------   Static Declarations ----------

namespace Xyce {
namespace Nonlinear {

//-----------------------------------------------------------------------------
// Function      : Sensitivity::Sensitivity
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/30/02
//-----------------------------------------------------------------------------
Sensitivity::Sensitivity (
  NonLinearSolver &     nls,
  Topo::Topology &      topTmp,
  const IO::CmdParse &  cp)
  : NonLinearSolver(cp),
    debugLevel_(1),
    solutionSize_(0),
    solveDirectFlag_(false),
    solveAdjointFlag_(true),
    outputScaledFlag_(false),
    outputUnscaledFlag_(true),
    maxParamStringSize_(0),
    stdOutputFlag_(true),
    fileOutputFlag_(false),
    dakotaFileOutputFlag_(false),
    forceFD_(false),
    numSolves_(0),
    objFuncGiven_(false),
    objFuncGIDsetup_(false),
    difference(SENS_FWD),
    sqrtEta_(1.0e-8),
    sqrtEtaGiven_(false),
    reuseFactors_(true),
    lambdaVectorPtr_(0),
    savedRHSVectorPtr_(0),
    savedNewtonVectorPtr_(0),

    nls_(nls),
    top_(topTmp),
    numSensParams_(0),
    numObjectives_(0)
{
  // if the base nonlinear solver class had a copy constructor, I could use
  // that here... maybe I'll set that up later. ERK 11/15/02.
  lasSysPtr_    = nls_.lasSysPtr_;

  registerAnalysisManager(&nls_.getAnalysisManager());
  nonlinearEquationLoader_    = nls_.nonlinearEquationLoader_;
  rhsVectorPtr_ = nls_.rhsVectorPtr_;

  NewtonVectorPtr_     = nls_.NewtonVectorPtr_;
  lasSolverPtr_        = nls_.lasSolverPtr_;
  jacobianMatrixPtr_   = nls_.jacobianMatrixPtr_;
  nextSolVectorPtrPtr_ = nls_.nextSolVectorPtrPtr_;
  currSolVectorPtrPtr_ = nls_.currSolVectorPtrPtr_;

  savedRHSVectorPtr_    = lasSysPtr_->builder().createVector();
  savedNewtonVectorPtr_ = lasSysPtr_->builder().createVector();

  lambdaVectorPtr_ = lasSysPtr_->builder().createVector();

  solutionSize_ = lasSysPtr_->getSolutionSize();
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::~Sensitivity
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/30/02
//-----------------------------------------------------------------------------
Sensitivity::~Sensitivity()
{
  delete savedRHSVectorPtr_;
  savedRHSVectorPtr_ = 0;

  delete savedNewtonVectorPtr_;
  savedNewtonVectorPtr_ = 0;

  delete lambdaVectorPtr_;
  lambdaVectorPtr_ = 0;

  for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
  {
    delete objFuncDataVec_[iobj]->dOdXVectorPtr_;
    objFuncDataVec_[iobj]->dOdXVectorPtr_ = 0;

    delete objFuncDataVec_[iobj]->expPtr_;
    objFuncDataVec_[iobj]->expPtr_ = 0;

    delete objFuncDataVec_[iobj];
    objFuncDataVec_[iobj] = 0;
  } 

  // For all the stuff that is to be deleted in the nonlinear solver
  // base class destructor, just set those pointers to zero because
  // they will have been deleted already.
  //
  // This is the consequence of having the sensitivity class derived
  // off of the nonlinear solver base class, but having it use all
  // the same linear solver objects, etc., as the nonlinear solver used
  // to solve the problem.
  NewtonVectorPtr_     = 0;
  gradVectorPtr_       = 0;
  solWtVectorPtr_      = 0;
  linsolOptionBlockPtr_ = 0;
  lasSolverPtr_        = 0;
}



//-----------------------------------------------------------------------------
// Function      : Sensitivity::stdOutput
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 2/18/2014
//-----------------------------------------------------------------------------
void Sensitivity::stdOutput (
       std::string idString,
       std::vector<double> & paramVals,
       std::vector<double> & sensitivities,
       std::vector<double> & scaled_sensitivities)
{
  // Send the sensitivity information to the screen:
#ifdef Xyce_PARALLEL_MPI
  N_PDS_Comm *pdsCommPtr = pdsMgrPtr_->getPDSComm();
  int myPID = pdsCommPtr->procID();
  if (myPID==0)
#endif
  {
    for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
    {
      Xyce::dout() << "\n"<<idString << " Sensitivities of objective function:";
      Xyce::dout()
        << objFuncDataVec_[iobj]->objFuncString_ << " = " 
        << objFuncDataVec_[iobj]->objFuncEval_ << std::endl;


      Xyce::dout() << std::setw(maxParamStringSize_)<<"Name";
      Xyce::dout() << "\t"<<std::setw(13)<<"Value";
      Xyce::dout() << "\t"<<std::setw(13)<<"Sensitivity";
      Xyce::dout() << "\t"<<std::setw(13)<<"Normalized"<<std::endl;

      for (int iparam=0; iparam< numSensParams_; ++iparam)
      {
        Xyce::dout() << std::setw(maxParamStringSize_)<<paramNameVec_[iparam];

        Xyce::dout() << "\t" << std::setw(13)<< std::scientific<< std::setprecision(4)
          << paramVals[iparam];

        int index= iobj*numSensParams_ +iparam;
        Xyce::dout() << "\t" << std::setw(13)<< std::scientific<< std::setprecision(4)
          << sensitivities[index];

        Xyce::dout() << "\t" << std::setw(13)<< std::scientific<< std::setprecision(4)
          << scaled_sensitivities[index] << std::endl;
      }
    }
  }
#ifdef Xyce_PARALLEL_MPI
  pdsCommPtr->barrier();
#endif
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::fileOutput
// Purpose       : Dump sensitivity information out to a file.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 2/18/2014
//-----------------------------------------------------------------------------
void Sensitivity::fileOutput (
       std::string idString,
       std::vector<double> & paramVals,
       std::vector<double> & sensitivities,
       std::vector<double> & scaled_sensitivities)
{
#ifdef Xyce_PARALLEL_MPI
  N_PDS_Comm *pdsCommPtr = pdsMgrPtr_->getPDSComm();
  int myPID = pdsCommPtr->procID();
  if (myPID==0)
#endif
  {
    std::ostringstream numSolvesOStr;
    numSolvesOStr << numSolves_;
    std::string dodpFileName = netlistFilename_ + numSolvesOStr.str() + "_dodp" + idString +".txt";
    FILE *fp = fopen(dodpFileName.c_str(),"w");
    for (int iparam=0;iparam< numSensParams_; ++iparam)
    {
      fprintf(fp,"\t%16.8e\n", sensitivities[iparam]);
    }
    fclose(fp);
  }
#ifdef Xyce_PARALLEL_MPI
  pdsCommPtr->barrier();
#endif
}

#if 0
//-----------------------------------------------------------------------------
// Function      : Sensitivity::dakOutput
// Purpose       : Dump sensitivity information out to a dakota-style file.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 2/19/2014
//-----------------------------------------------------------------------------
void Sensitivity::dakOutput (
       std::string idString,
       std::vector<double> & paramVals,
       std::vector<double> & sensitivities,
       std::vector<double> & scaled_sensitivities)
{
#ifdef Xyce_PARALLEL_MPI
  N_PDS_Comm *pdsCommPtr = pdsMgrPtr_->getPDSComm();
  int myPID = pdsCommPtr->procID();
  if (myPID==0)
#endif
  {
    // write a file format that can be used by dakota :
    // Note that currently, this will simply overwrite the same 
    // file every time this function is called.
    std::string dakotaFileName = netlistFilename_ + "_dodp" + idString + "_all.txt";
    FILE *fp2 = fopen(dakotaFileName.c_str(),"w");
    fprintf(fp2,"%16.8e", objFuncEval_ );
    fprintf(fp2,"%s","\n[\n");
    for (int iparam=0;iparam< numSensParams_; ++iparam)
    {
      fprintf(fp2,"\t%16.8e\n", sensitivities[iparam]);
    }
    fprintf(fp2,"%s","]\n");
    fclose(fp2);
  }
#ifdef Xyce_PARALLEL_MPI
  pdsCommPtr->barrier();
#endif
}
#endif


//-----------------------------------------------------------------------------
// Function      : Sensitivity::icSensitivity
// Purpose       : This function is called when there is a NOOP or UIC.  The
//                 dqdp vector needs to be set up for the history to be correct.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 
//-----------------------------------------------------------------------------
bool Sensitivity::icSensitivity ( 
     std::vector<double> & objectiveVec,
     std::vector<double> & dOdpVec, 
     std::vector<double> & dOdpAdjVec,
     std::vector<double> & scaled_dOdpVec, 
     std::vector<double> & scaled_dOdpAdjVec)
{
  if (!solveDirectFlag_ && !solveAdjointFlag_) return 1;

  TimeIntg::DataStore & ds = *dsPtr_;

  ds.dOdpVec_.clear();
  ds.dOdpAdjVec_.clear();

  ds.scaled_dOdpVec_.clear();
  ds.scaled_dOdpAdjVec_.clear();

  // first get the derivatives of the RHS vector w.r.t. the
  // user-specified optimization parameters.
  loadSensitivityResiduals (difference, forceFD_, sqrtEta_, netlistFilename_, 
      *dsPtr_, *nonlinearEquationLoader_, paramNameVec_, getAnalysisManager());

  calcObjFuncDerivs ();

  for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
  {
    objectiveVec.clear();
    objectiveVec.push_back(objFuncDataVec_[iobj]->objFuncEval_);
    ds.objectiveVec_.clear();
    ds.objectiveVec_.push_back(objFuncDataVec_[iobj]->objFuncEval_);
  }

  if (solveDirectFlag_) 
  {
    ds.dOdpVec_.resize(numSensParams_*numObjectives_ ,0.0);
    ds.scaled_dOdpVec_.resize(numSensParams_*numObjectives_,0.0);

    if (outputUnscaledFlag_)
    {
      dOdpVec = ds.dOdpVec_;
    }

    if (outputScaledFlag_)
    {
      scaled_dOdpVec = ds.scaled_dOdpVec_;
    }

    if (stdOutputFlag_)
    {
      stdOutput(std::string("Direct"), ds.paramOrigVals_, ds.dOdpVec_, ds.scaled_dOdpVec_);
    }
  }

  if (solveAdjointFlag_) 
  {
    ds.dOdpAdjVec_.resize(numSensParams_*numObjectives_,0.0);
    ds.scaled_dOdpAdjVec_.resize(numSensParams_*numObjectives_,0.0);

    if (outputUnscaledFlag_)
    {
      dOdpAdjVec = ds.dOdpAdjVec_;
    }

    if (outputScaledFlag_)
    {
      scaled_dOdpAdjVec = ds.scaled_dOdpAdjVec_;
    }
    if (stdOutputFlag_)
    {
      stdOutput(std::string("Adjoint"), ds.paramOrigVals_, ds.dOdpAdjVec_, ds.scaled_dOdpAdjVec_);
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::solve
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 11/21/02
//-----------------------------------------------------------------------------
int Sensitivity::solve( 
     std::vector<double> & objectiveVec,
     std::vector<double> & dOdpVec, 
     std::vector<double> & dOdpAdjVec,
     std::vector<double> & scaled_dOdpVec, 
     std::vector<double> & scaled_dOdpAdjVec)
{
  Stats::StatTop _solveStat("Sensistivity Solve");
  Stats::TimeBlock _solveTimer(_solveStat);

  if (!solveDirectFlag_ && !solveAdjointFlag_) return 1;

  TimeIntg::DataStore & ds = *dsPtr_;

  ds.dOdpVec_.clear();
  ds.dOdpAdjVec_.clear();

  ds.scaled_dOdpVec_.clear();
  ds.scaled_dOdpAdjVec_.clear();

  // It may now be neccessary to re-load the jacobian and rhs vectors.
  // It is necccessary, for example, if the two-level Newton solver is the
  // solver being used.
  nls_.enableSensitivity ();

  // first get the derivatives of the RHS vector w.r.t. the
  // user-specified optimization parameters.
  loadSensitivityResiduals (difference, forceFD_, sqrtEta_, netlistFilename_, 
      *dsPtr_, *nonlinearEquationLoader_, paramNameVec_, getAnalysisManager());

  calcObjFuncDerivs ();

  objectiveVec.clear();
  ds.objectiveVec_.clear();
  for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
  {
    objectiveVec.push_back(objFuncDataVec_[iobj]->objFuncEval_);
    ds.objectiveVec_.push_back(objFuncDataVec_[iobj]->objFuncEval_);
  }

  if (solveDirectFlag_) 
  {
    solveDirect ();

    if (outputUnscaledFlag_)
    {
      dOdpVec = ds.dOdpVec_;
    }

    if (outputScaledFlag_)
    {
      scaled_dOdpVec = ds.scaled_dOdpVec_;
    }
  }

  if (solveAdjointFlag_) 
  {
    solveAdjoint ();
    if (outputUnscaledFlag_)
    {
      dOdpAdjVec = ds.dOdpAdjVec_;
    }

    if (outputScaledFlag_)
    {
      scaled_dOdpAdjVec = ds.scaled_dOdpAdjVec_;
    }
  }

  numSolves_++;

  return 1;
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::solveDirect
//
// Purpose       : This function calculates the direct sensitivities for
//                 the user specified parameters.
//
//                 The ultimate goal of this function is to obtain dO/dp,
//                 where O is the objective function, and p is a
//                 user-defined optimization parameter.
//
//                 This is a bit confusing because the DAKOTA folks use
//                 a different naming convention than I have tended to
//                 use.  In Dakota's documentation, dO/dp would be referred
//                 to as df/du.  In Xyce, f is already considered to be the
//                 right-hand-side residual vector.  For clarity, this is
//                 the key between my notation and DAKOTA's:
//                        DAK     Xyce
//                         f       O
//                         y       x
//                         u       p
//                         c       f
//
//                 To obtain dOdp, the device manager is first called, and
//                 told to calculate dfdp, which is the derivative of the
//                 residual vector w.r.t. user-defined params.  It does
//                 this by numerical differentiation.  Then, after that,
//                 this function solves the equation:
//
//                 J dx/dp = df/dp    ->   dx/dp = J^-1 df/dp
//
//                 for each p.  
//
//                 J is the jacobian matrix, so J=df/dx.  (dc/dy)
//
//                 After performing these linear solves, dO/dp is to be
//                 obtained using the chain rule by:
//
//                 dO/dp = - dO/dx * dx/dp  + dO/dp
//
//                 The O in the dO/dp on the left side of this equation
//                 should have a hat over it "^", to indicate that it is
//                 different than the O on the right hand side.
//
//                 Note, this method is best if you have lots of objective
//                 functions, and a small number of parameters.  For adjoint
//                 calculations it is the other way around.
//
//                 11/19/02.
//                 It is assumed (for now) that dO/dp on the right hand side
//                 is zero, i.e., there is no direct
//                 dependence on p by O.  So, for example, if a user
//                 defined parameter to be used is the length of a MOSFET,
//                 the MOSFET length will NOT appear in the analytical
//                 expression for O.
//
//
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/31/02
//-----------------------------------------------------------------------------
int Sensitivity::solveDirect ()
{
  Stats::StatTop _solveDirectStat("Solve Direct");
  Stats::TimeBlock _solveDirectTimer(_solveDirectStat);

  if (DEBUG_NONLINEAR && isActive(Diag::SENS_SOLVER))
  {
    Xyce::dout() << std::endl
                 << "In Sensitivity::solveDirect" << std::endl;
  }

  TimeIntg::DataStore & ds = *dsPtr_;

  int iparam;
  std::vector<Linear::Vector *> & dXdpPtrVector_ = ds.nextDXdpPtrVector;
  std::vector<Linear::Vector *> & currdXdpPtrVector_ = ds.currDXdpPtrVector;
  std::vector<Linear::Vector *> & sensRHSPtrVector = ds.sensRHSPtrVector;

  std::vector<Linear::Vector *> & DQdxDXdpPtrVector_ = ds.nextDQdxDXdpPtrVector;

  // first save a copy of the rhs and newton vectors, in case we want them later.
  savedRHSVectorPtr_->update(1.0, *(rhsVectorPtr_), 0.0);
  savedNewtonVectorPtr_->update(1.0, *(NewtonVectorPtr_), 0.0);

  // via the loader, setup all the sensitivity residuals.
  nonlinearEquationLoader_->loadSensitivityResiduals();

  if (DEBUG_NONLINEAR && isActive(Diag::SENS_SOLVER))
  {
    std::string matrixFile = netlistFilename_ + "_directMatrix.txt";
    jacobianMatrixPtr_->writeToFile(const_cast<char *>(matrixFile.c_str()));
  }

  // Now solve the series of linear systems to get dXdp.
  for (iparam=0; iparam< numSensParams_; ++iparam)
  {
    // copy the sensitivity residual into the rhs vector location,
    // as that is the RHS vector that the linear solver expects.
    rhsVectorPtr_->update(1.0, *(sensRHSPtrVector[iparam]), 0.0);

    lasSolverPtr_->solve(reuseFactors_);

    // copy the result of the linear solve into the dxdp data structure.
    (dXdpPtrVector_[iparam])->update(1.0, *(NewtonVectorPtr_), 0.0);

    // do debug output.
    if (DEBUG_NONLINEAR && isActive(Diag::SENS_SOLVER))
    {
      Xyce::dout() << "iparam="<<iparam << "\t" << paramNameVec_[iparam] <<std::endl;
      for (int k = 0; k < solutionSize_; ++k)
      {
        Xyce::dout() << "dXdp[" << std::setw(3) << k << "] = "<< std::setw(15)<< std::scientific
          << std::setprecision(8)<< (*(dXdpPtrVector_[iparam]))[k]
          <<std::endl;
      }

      std::ostringstream filename; 
      filename << netlistFilename_ << "_dxdp";
      filename << std::setw(3) << std::setfill('0') << iparam;
      filename << ".txt";
      dXdpPtrVector_[iparam]->writeToFile(const_cast<char *>(filename.str().c_str()));
    }

    // Now store the DQdx*dXdp matvec
    Linear::Matrix & dQdx = *(ds.dQdxMatrixPtr);
    bool Transpose = false;
    (DQdxDXdpPtrVector_[iparam])->putScalar(0.0);
    dQdx.matvec( Transpose , *(dXdpPtrVector_[iparam]), *(DQdxDXdpPtrVector_[iparam]) );

  }// end of param for loop

  // Now update the time derivative of the matvec (this is needed for trap).
  nonlinearEquationLoader_->loadFinalSensitivityDerivatives ();

  // Restore the RHS and Newton vectors.
  rhsVectorPtr_->update(1.0, *(savedRHSVectorPtr_),0.0);
  NewtonVectorPtr_->update(1.0, *(savedNewtonVectorPtr_),0.0);

  // Now get the final dOdp's (one for each objective/param combination).
  for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
  {
    for (iparam=0; iparam< numSensParams_; ++iparam)
    {
      double tmp = objFuncDataVec_[iobj]->dOdXVectorPtr_->dotProduct( (*(dXdpPtrVector_[iparam]))  );
      tmp += objFuncDataVec_[iobj]->dOdp_;

      ds.dOdpVec_.push_back(tmp);
   
      // get scaled value.  dO/dp*(p/100)
      double normalize = ds.paramOrigVals_[iparam]/100.0;
      tmp *= normalize;
      ds.scaled_dOdpVec_.push_back(tmp);
    }
  }

  if (stdOutputFlag_)
  {
    stdOutput(std::string("Direct"), ds.paramOrigVals_, ds.dOdpVec_, ds.scaled_dOdpVec_);
  }

  if (fileOutputFlag_)
  {
    fileOutput(std::string("Direct"), ds.paramOrigVals_, ds.dOdpVec_, ds.scaled_dOdpVec_);
  }

#if 0
  if (dakotaFileOutputFlag_)
  {
    dakOutput(std::string("Direct"), ds.paramOrigVals_, ds.dOdpVec_, ds.scaled_dOdpVec_);
  }
#endif

  return 1;
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::calcObjFuncDerivs
//
// Purpose       : This function assumes an objective function, O, and
//                 uses it to calculate other stuff, including dO'/du,
//                 dO/dy, dO/du.
//
//                 The objective function simply assumes that you are
//                 trying to match a solution variable with a
//                 user-specified number.   (objective=1)
//
//                 Or that you are trying to match your final solution
//                 with a specified solution. (objective=2)
//
// Special Notes : This is mostly for testing purposes.  Usually, the
//                 objective function would be set and managed from
//                 the DAKOTA side.
//
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 11/15/02
//-----------------------------------------------------------------------------
bool Sensitivity::calcObjFuncDerivs ()
{
  bool bsuccess = true;
  int i;

  for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
  {
    objFuncDataVec_[iobj]->dOdXVectorPtr_->putScalar(0.0);
  }

  // first check if the user specified a function via the expressions
  // class:
  int found(0);
  int found2(0);
  bool foundLocal(false);
  bool foundLocal2(false);

  N_PDS_Comm & comm = *(pdsMgrPtr_->getPDSComm());
  int myPID = comm.procID();

  if (objFuncGiven_)
  {
    if (!objFuncGIDsetup_)
    {
      for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
      {
        // set up the gid's:
        objFuncDataVec_[iobj]->expVarGIDs_.resize( objFuncDataVec_[iobj]->numExpVars_, -1 );
        for (i = 0; i < objFuncDataVec_[iobj]->numExpVars_; ++i)
        {
          std::vector<int> svGIDList1, dummyList;
          char type1;
          foundLocal = top_.getNodeSVarGIDs(NodeID(objFuncDataVec_[iobj]->expVarNames_[i], Xyce::_VNODE), svGIDList1, dummyList, type1);
          found = static_cast<int>(foundLocal);
          Xyce::Parallel::AllReduce(comm.comm(), MPI_LOR, &found, 1);

          foundLocal2 = false;
          if (!found)// if looking for this as a voltage node failed, try a "device" (i.e. current) node.
          {
            foundLocal2 = top_.getNodeSVarGIDs(NodeID(objFuncDataVec_[iobj]->expVarNames_[i], Xyce::_DNODE), svGIDList1, dummyList, type1);
          }
          found2 = static_cast<int>(foundLocal2);
          Xyce::Parallel::AllReduce(comm.comm(), MPI_LOR, &found2, 1);

          if (!found && !found2)
          {
            static std::string tmp = "objective function variable not found!\n";
            N_ERH_ErrorMgr::report( N_ERH_ErrorMgr::USR_FATAL, tmp);
          }

          if (found || found2)
          {
            int tmpGID=-1;
            if(svGIDList1.size()==1)
            {
              tmpGID = svGIDList1.front();
            }
            objFuncDataVec_[iobj]->expVarGIDs_[i] = tmpGID;
          }
        }
      }
      objFuncGIDsetup_ = true;
    }

    // obtain the expression variable values.  It will only grab this value if it is owned on this processor.
    for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
    {
      objFuncDataVec_[iobj]->expVarVals_.resize (objFuncDataVec_[iobj]->numExpVars_, 0.0);
      objFuncDataVec_[iobj]->expVarDerivs_.resize (objFuncDataVec_[iobj]->numExpVars_, 0.0);

      comm.barrier();
      for (i = 0; i < objFuncDataVec_[iobj]->numExpVars_; ++i)
      {
        int tmpGID=objFuncDataVec_[iobj]->expVarGIDs_[i];
        double tmpVal=0.0;
        int root=-1;
        if (tmpGID >= 0)
        {
          tmpVal = (*nextSolVectorPtrPtr_)->getElementByGlobalIndex(tmpGID, 0);
          root = myPID;
        }
        Xyce::Parallel::AllReduce(comm.comm(), MPI_MAX, &root, 1);

        comm.bcast( &tmpVal, 1, root );

        objFuncDataVec_[iobj]->expVarVals_[i] = tmpVal;
      }
      comm.barrier();
    }
   
    //get expression value and partial derivatives
    for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
    {
      objFuncDataVec_[iobj]->expPtr_->evaluate( 
          objFuncDataVec_[iobj]->expVal_, 
          objFuncDataVec_[iobj]->expVarDerivs_, 
          objFuncDataVec_[iobj]->expVarVals_ );

      objFuncDataVec_[iobj]->objFuncEval_ = objFuncDataVec_[iobj]->expVal_;
      objFuncDataVec_[iobj]->dOdXVectorPtr_->putScalar(0.0);
      for (i=0;i<objFuncDataVec_[iobj]->numExpVars_;++i)
      {
        int tmpGID = objFuncDataVec_[iobj]->expVarGIDs_[i];
        double tmpDODX = objFuncDataVec_[iobj]->expVarDerivs_[i];

        if (DEBUG_NONLINEAR && isActive(Diag::SENS_SOLVER))
        {
          Xyce::dout() << "i="<<i<<"  gid = " << tmpGID << "  dodx = "<< tmpDODX << std::endl;
        }

        if (tmpGID >= 0)
        {
          objFuncDataVec_[iobj]->dOdXVectorPtr_->setElementByGlobalIndex(tmpGID, tmpDODX, 0);
        }
      }

      // Assuming this is zero:
      objFuncDataVec_[iobj]->dOdp_ = 0.0;
    }
  }
  else // give a warning.
  {
    Report::UserWarning0() << "Objective function was not specified";
  }


  for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
  {
    objFuncDataVec_[iobj]->dOdXVectorPtr_->fillComplete();

    if (DEBUG_NONLINEAR && isActive(Diag::SENS_SOLVER))
    {
      std::string filename = netlistFilename_ + "_dodx.txt";
      objFuncDataVec_[iobj]->dOdXVectorPtr_->writeToFile(const_cast<char *>(filename.c_str()));
    }
  }

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::solveAdjoint
// Purpose       : Solves first for the vector, lambda, of adjoint variables.
//                 Afterwards, it solves for dO/dp.
// Special Notes : 
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 11/20/02
//-----------------------------------------------------------------------------
int Sensitivity::solveAdjoint ()
{
  Stats::StatTop _solveAdjointStat("Solve Adjoint");
  Stats::TimeBlock _solveAdjointTimer(_solveAdjointStat);

  if (DEBUG_NONLINEAR && isActive(Diag::SENS_SOLVER))
  {
    Xyce::dout() << std::endl;
    Xyce::dout() << "In Sensitivity::solveAdjoint" << std::endl;
  }

  TimeIntg::DataStore & ds = *dsPtr_;

  std::vector<Linear::Vector *> & dfdpPtrVector_ = ds.nextDfdpPtrVector;
  std::vector<Linear::Vector *> & dbdpPtrVector_ = ds.nextDbdpPtrVector;
  std::vector<Linear::Vector *> & sensRHSPtrVector = ds.sensRHSPtrVector;

  // first save a copy of the rhs vector, in case we want it later.
  savedRHSVectorPtr_->update(1.0, *(rhsVectorPtr_),0.0);
  savedNewtonVectorPtr_->update(1.0, *(NewtonVectorPtr_),0.0);

  bool useTran = jacobianMatrixPtr_->useTranspose ();
  if (DEBUG_NONLINEAR && isActive(Diag::SENS_SOLVER))
  {
    std::string matrixFile = netlistFilename_ + "_adjointMatrix.txt";
    jacobianMatrixPtr_->writeToFile(const_cast<char *>(matrixFile.c_str()));
  }

  for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
  {
    // copy the current dOdx vector into the rhs vector data structure.
    rhsVectorPtr_->update(1.0, *(objFuncDataVec_[iobj]->dOdXVectorPtr_),0.0);

    int status = lasSolverPtr_->solveTranspose(reuseFactors_);
    if (status!=0)
    {
      std::string msg("Sensitivity::solveAdjoint.  Solver failed\n");
      N_ERH_ErrorMgr::report (N_ERH_ErrorMgr::DEV_FATAL, msg);
    }

    // allocate the dxdp vector for this param, and
    // copy the resulting deltax vector into the dxdp data structure.
    lambdaVectorPtr_->update(1.0, *(NewtonVectorPtr_),0.0);

    if (DEBUG_NONLINEAR && isActive(Diag::SENS_SOLVER))
    {
      std::string filename = netlistFilename_ + "_lambda.txt";
      lambdaVectorPtr_->writeToFile(const_cast<char *>(filename.c_str()));
    }

    // Now that we have lambda, get the dOdp's by doing dot products of
    // lambda * df/dp.

    // do the final dot products, one for each param.
    for (int iparam=0; iparam< numSensParams_; ++iparam)
    {
      // assuming that this is steady-state, then we need to sum together dfdp and -dbdp.
      Linear::Vector & RHSVec  = *(sensRHSPtrVector[iparam]);
      Linear::Vector & dfdpVec = *(dfdpPtrVector_[iparam]);
      Linear::Vector & dbdpVec = *(dbdpPtrVector_[iparam]);
      RHSVec.linearCombo(0.0,RHSVec,+1.0,dfdpVec);
      RHSVec.linearCombo(1.0,RHSVec,-1.0,dbdpVec);

      double tmp = -1.0 * lambdaVectorPtr_->dotProduct(RHSVec);
      ds.dOdpAdjVec_.push_back(tmp);

      // get scaled value.  dO/dp*(p/100)
      double normalize = ds.paramOrigVals_[iparam]/100.0;
      tmp *= normalize;
      ds.scaled_dOdpAdjVec_.push_back(tmp);
    }
  }

  // restore the useTranspose flag to the original setting. (false)
  jacobianMatrixPtr_->setUseTranspose (useTran);

  // Restore the RHS and Newton vectors.
  rhsVectorPtr_->update(1.0, *(savedRHSVectorPtr_),0.0);
  NewtonVectorPtr_->update(1.0, *(savedNewtonVectorPtr_),0.0);

  // set the sensitivity information to the screen:
  if (stdOutputFlag_)
  {
    stdOutput(std::string("Adjoint"), ds.paramOrigVals_, ds.dOdpAdjVec_, ds.scaled_dOdpAdjVec_);
  }
  if (fileOutputFlag_)
  {
    fileOutput(std::string("Adjoint"), ds.paramOrigVals_, ds.dOdpAdjVec_, ds.scaled_dOdpAdjVec_);
  }

#if 0
  if (dakotaFileOutputFlag_)
  {
    dakOutput(std::string("Adjoint"), ds.paramOrigVals_, ds.dOdpAdjVec_, ds.scaled_dOdpAdjVec_);
  }
#endif

  return 1;
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::setOptions
//
// Purpose       : This function processes the .SENS line
//
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 11/15/02
//-----------------------------------------------------------------------------
bool Sensitivity::setOptions(const Util::OptionBlock& OB)
{
  bool bsuccess = true;
  Util::ParamList::const_iterator iter = OB.begin();
  Util::ParamList::const_iterator end   = OB.end();

  numSensParams_ = 0;

  for ( ; iter != end; ++ iter)
  {
    if ( std::string( iter->uTag() ,0,7) == "OBJFUNC") // this is a vector
    {
      objectiveFunctionData * ofDataPtr = new objectiveFunctionData();
      ofDataPtr->objFuncString_ = iter->stringValue();
      ofDataPtr->expPtr_ = new Util::Expression(iter->stringValue());
      ofDataPtr->dOdXVectorPtr_ = lasSysPtr_->builder().createVector();
      objFuncDataVec_.push_back(ofDataPtr);

      objFuncGiven_ = true;
    }
    else if ( std::string( iter->uTag() ,0,5) == "PARAM") // this is a vector
    {
      ExtendedString tag = iter->stringValue();
      tag.toUpper();
      // set up the initial skeleton of the maps:
      ++numSensParams_;
      paramNameVec_.push_back(tag);
      int sz = tag.size();
      if (sz > maxParamStringSize_)
      {
        maxParamStringSize_ = sz;
      }
    }
    else
    {
      Xyce::Report::UserWarning() << iter->uTag() 
        << " is not a recognized sensitivity solver option.\n" << std::endl;
    }
  }
  numObjectives_ = objFuncDataVec_.size();

  // parse the expression now, so if there are any errors, they will come
  // up early in the simulation.
  if (objFuncGiven_)
  {
    for (int iobj=0;iobj<objFuncDataVec_.size();++iobj)
    {
      // setup the names:
      objFuncDataVec_[iobj]->expVarNames_.clear();

      std::vector<std::string> nodes;
      objFuncDataVec_[iobj]->expPtr_->get_names(XEXP_NODE, nodes);
      std::vector<std::string> instances;
      objFuncDataVec_[iobj]->expPtr_->get_names(XEXP_INSTANCE, instances);

      // Make the current (instance) strings all upper case.
      // The topology directory apparently requires this.
      std::vector<std::string>::iterator iter;
      for (iter=instances.begin();iter!=instances.end();++iter)
      {
        ExtendedString tmpString = *iter;
        tmpString.toUpper ();
        *iter  = tmpString;
      }

      objFuncDataVec_[iobj]->expVarNames_.insert(objFuncDataVec_[iobj]->expVarNames_.end(), nodes.begin(), nodes.end());
      objFuncDataVec_[iobj]->expVarNames_.insert(objFuncDataVec_[iobj]->expVarNames_.end(), instances.begin(), instances.end());

      // Order the names in the expression so that it agrees with the order
      // in expVarNames.
      if ( !(objFuncDataVec_[iobj]->expVarNames_.empty()) )
      {
        objFuncDataVec_[iobj]->expPtr_->order_names( objFuncDataVec_[iobj]->expVarNames_ );
      }

      objFuncDataVec_[iobj]->numExpVars_ = objFuncDataVec_[iobj]->expVarNames_.size();

      if (objFuncDataVec_[iobj]->numExpVars_<=0)
      {
        static std::string tmp = "Objective function does not contain a resolvable solution variable.\n";
        N_ERH_ErrorMgr::report( N_ERH_ErrorMgr::USR_FATAL_0, tmp);
      }
    }
  }

  if (DEBUG_NONLINEAR && isActive(Diag::SENS_SOLVER))
  {
    std::vector<std::string>::iterator iter;
    std::vector<std::string>::iterator begin = paramNameVec_.begin();
    std::vector<std::string>::iterator end   = paramNameVec_.end  ();

    for (iter=begin;iter!=end;++iter)
    {
      Xyce::dout() << *iter<<std::endl;
    }
  }

  dsPtr_->allocateSensitivityArrays(lasSysPtr_->builder(), numSensParams_);

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::setSensitivityOptions
// Purpose       : This function processes the .options SENSITIVITY line
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 06/20/2013
//-----------------------------------------------------------------------------
bool Sensitivity::setSensitivityOptions(const Util::OptionBlock &OB)
{
  bool bsuccess = true;
  Util::ParamList::const_iterator it  = OB.begin();
  Util::ParamList::const_iterator end = OB.end();
  for ( ; it != end; ++ it)
  {
    if ((*it).uTag() == "ADJOINT")
    {
      solveAdjointFlag_ = 
        static_cast<bool>((*it).getImmutableValue<bool>());
    }
    else if ((*it).uTag() == "DIRECT")
    {
      solveDirectFlag_ = 
        static_cast<bool>((*it).getImmutableValue<bool>());
    }
    else if ((*it).uTag() == "OUTPUTSCALED")
    {
      outputScaledFlag_ = 
        static_cast<bool>((*it).getImmutableValue<bool>());
    }
    else if ((*it).uTag() == "OUTPUTUNSCALED")
    {
      outputUnscaledFlag_ = 
        static_cast<bool>((*it).getImmutableValue<bool>());
    }
    else if ((*it).uTag() == "STDOUTPUT")
    {
      stdOutputFlag_ = 
        static_cast<bool>((*it).getImmutableValue<bool>());
    }
    else if ((*it).uTag() == "DAKOTAFILE")
    {
      dakotaFileOutputFlag_ = 
        static_cast<bool>((*it).getImmutableValue<bool>());
    }
    else if ((*it).uTag() == "DIAGNOSTICFILE")
    {
      fileOutputFlag_ = 
        static_cast<bool>((*it).getImmutableValue<bool>());
    }
    else if ((*it).uTag() == "FORCEFD")
    {
      forceFD_ = 
        static_cast<bool>((*it).getImmutableValue<bool>());
    }
    else if ((*it).uTag() == "DIFFERENCE")
    {
      ExtendedString sval=(*it).stringValue();
      sval.toUpper();
      if(sval=="FORWARD")
      {
        difference=SENS_FWD;
      }
      else if(sval=="REVERSE")
      {
        difference=SENS_REV;
      }
      else if(sval=="CENTRAL")
      {
        difference=SENS_CNT;
        static std::string tmp = "difference=central not supported.\n";
        N_ERH_ErrorMgr::report( N_ERH_ErrorMgr::USR_FATAL_0, tmp);
      }
      else
      {
        static std::string tmp = "difference not recognized!\n";
        N_ERH_ErrorMgr::report( N_ERH_ErrorMgr::USR_FATAL_0, tmp);
      }
    }
    else if ((*it).uTag() == "SQRTETA")
    {
      sqrtEta_ = (*it).getImmutableValue<double>();
      sqrtEtaGiven_ = true;
    }
    else if ((*it).uTag() == "REUSEFACTORS")
    {
      reuseFactors_ = (*it).getImmutableValue<double>();
    }
    else if (DEBUG_NONLINEAR && (*it).uTag() == "DEBUGLEVEL")
    {
      debugLevel_ = (*it).getImmutableValue<int>();
      Xyce::setSensitivityDebugLevel(debugLevel_);
    }
    else
    {
      Xyce::Report::UserWarning() << (*it).uTag() 
        << " is not a recognized sensitivity solver option.\n" << std::endl;
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::getMaxNormF
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and MEMS Modeling
// Creation Date : 9/28/2009
//-----------------------------------------------------------------------------
double Sensitivity::getMaxNormF() const
{
  return nls_.getMaxNormF();
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::getMaxNormFindex
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and MEMS Modeling
// Creation Date : 9/28/2009
//-----------------------------------------------------------------------------
int Sensitivity::getMaxNormFindex () const
{
  return nls_.getMaxNormFindex ();
}

//-----------------------------------------------------------------------------
// Function      : extractSENSData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 10/11/2007
//-----------------------------------------------------------------------------
bool extractSENSData(
  IO::PkgOptionsMgr &           options_manager,
  IO::CircuitBlock &            circuit_block,
  const std::string &           netlist_filename,
  const IO::TokenVector &       parsed_line)
{
  Util::OptionBlock option_block("SENS", Util::OptionBlock::ALLOW_EXPRESSIONS, netlist_filename, parsed_line[0].lineNumber_);

  int numFields = parsed_line.size();

  int parameterStartPos = 1;

  // Create an option block to temporarily store the default options.
  Util::OptionBlock defaultOptions;

  // Get the default options from metadata.
  addDefaultOptionsParameters(options_manager, defaultOptions, "SENS" );

  // Extract the parameters from parsed_line.
  int parameterEndPos = numFields - 1;
  Util::ParamList inputParameters;
  Util::Param parameter("", "");
  int intervalParameterStart = -1;
  int i = parameterStartPos;
  std::string paramBaseName;
  while (i <= parameterEndPos-1)
  {
    // Check for equal sign.
    if ( parsed_line[i+1].string_ != "=" )
    {
      // Stop after the tagged parameters have been extracted
      // from a .OPTIONS RESTART or .OPTIONS OUTPUT line, they
      // will be handled later.
      intervalParameterStart = i;
      break;
    }

    // Extract parameter name and value from parsed_line and add to
    // parameter list. Check to see if the parameter is "VECTOR"
    // valued and treat accordingly.
    parameter.set( parsed_line[i].string_, "" );
    Util::Param *parameterPtr = Util::findParameter(defaultOptions.begin(), defaultOptions.end(), parameter.tag());
    if (parameterPtr == NULL)
    {
      Report::UserWarning0().at(netlist_filename, parsed_line[0].lineNumber_)
        << "No options parameter " << parameter.tag() << " found, parameter will be ignored.";
      i+= 3;
    }
    else if (parameterPtr->stringValue() != "VECTOR")
    {
      parameter.setVal( parsed_line[i+2].string_ );
      inputParameters.push_back( parameter );
      i+= 3;

      if (i < parameterEndPos-1 && parsed_line[i].string_ == ",")
      {
        Report::UserError0().at(netlist_filename, parsed_line[0].lineNumber_)
          << "Options parameter " << parameter.tag() << " is flagged as not VECTOR, but has comma in value.";
      }
    }
    else
    {
      // We have a vector valued parameter.
      // Name the jth component of the parameter of the vector by appending
      // "j" to the parameter name.
      std::ostringstream paramName;
      paramBaseName = ExtendedString(parsed_line[i].string_).toUpper();
      int j = 1;

      paramName << paramBaseName << j;
      i += 2;
      parameter.set(paramName.str(), parsed_line[i].string_);
      option_block.addParam(parameter);

      // This while loop is dangerous if we are near the end of the
      // parsed line because it still assumed that the format is
      // option=value,value and not option=value,value,
      // that is no trailing comma.  It does work if
      // option=value,value is at the end of a line now (RLS 5/07)
      int testSize = parsed_line.size()-1;
      while ((i < testSize) && (parsed_line[i+1].string_ == ",") )
      {
        paramName.str("");
        ++j;
        paramName << paramBaseName << j;
        i += 2;
        parameter.set(paramName.str(), parsed_line[i].string_);
        option_block.addParam(parameter);
      }

      ++i;
    }
  }

  // For each input parameter, check that it is in the default
  // set and if so, set its value in "parameters" to the input
  // value, otherwise flag it as an unknown parameter.
  for (Util::ParamList::const_iterator it = inputParameters.begin(), end = inputParameters.end(); it != end; ++it)
  {
    Util::Param *parameterPtr = Util::findParameter(defaultOptions.begin(), defaultOptions.end(), (*it).tag());
    if ( parameterPtr != NULL )
    {
      parameterPtr->setVal(*it);
      option_block.addParam( *parameterPtr );
    }
    else
    {
      Report::UserWarning0().at(netlist_filename, parsed_line[0].lineNumber_)
        << "No options parameter " << (*it).tag() << " found, parameter will be ignored.";
    }
  }

  circuit_block.addOptions(option_block);

  return true; // Only get here on success.
}

//-----------------------------------------------------------------------------
// Function      : Sensitivity::populateMetadata
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Dave Baur
// Creation Date :
//-----------------------------------------------------------------------------
void Sensitivity::populateMetadata(
  IO::PkgOptionsMgr &   options_manager)
{
  {
    Util::ParamMap &parameters = options_manager.addOptionsMetadataMap("SENSITIVITY");

    parameters.insert(Util::ParamMap::value_type("DEBUGLEVEL", Util::Param("DEBUGLEVEL", 0)));
    parameters.insert(Util::ParamMap::value_type("ADJOINT", Util::Param("ADJOINT", 0)));
    parameters.insert(Util::ParamMap::value_type("DIRECT", Util::Param("DIRECT", 0)));
    parameters.insert(Util::ParamMap::value_type("OUTPUTSCALED", Util::Param("OUTPUTSCALED", 0)));
    parameters.insert(Util::ParamMap::value_type("OUTPUTUNSCALED", Util::Param("OUTPUTUNSCALED", 1)));
    parameters.insert(Util::ParamMap::value_type("STDOUTPUT", Util::Param("STDOUTPUT", 1)));
    parameters.insert(Util::ParamMap::value_type("DIAGNOSTICFILE", Util::Param("DIAGNOSTICFILE", 0)));
    parameters.insert(Util::ParamMap::value_type("DAKOTAFILE", Util::Param("DAKOTAFILE", 0)));
    parameters.insert(Util::ParamMap::value_type("DIFFERENCE", Util::Param("DIFFERENCE", 0)));
    parameters.insert(Util::ParamMap::value_type("SQRTETA", Util::Param("SQRTETA", 1.0e-8)));
    parameters.insert(Util::ParamMap::value_type("FORCEFD", Util::Param("FORCEFD", 0)));
  }

  {
    Util::ParamMap &parameters = options_manager.addOptionsMetadataMap("SENS");

    parameters.insert(Util::ParamMap::value_type("OBJFUNC", Util::Param("OBJFUNC", "VECTOR")));
    parameters.insert(Util::ParamMap::value_type("PARAM", Util::Param("PARAM", "VECTOR")));
  }

  options_manager.addCommandParser(".SENS", extractSENSData);
}

} // namespace Nonlinear
} // namespace Xyce
