//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_PDS_GlobalAccessor.h,v $
//
// Purpose        : Migrator tool using Zoltan utilities
//
// Special Notes  :
//
// Creator        : Robert Hoekstra, SNL, Parallel Computational Sciences
//
// Creation Date  : 06/07/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.24 $
//
// Revision Date  : $Date: 2015/04/08 19:18:30 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_PDS_GlobalAccessor_h
#define Xyce_N_PDS_GlobalAccessor_h

#include <list>
#include <map>
#include <vector>

#include <N_PDS_fwd.h>
#include <N_LAS_fwd.h>

class N_PDS_ParMap;
class Epetra_Distributor;

//-----------------------------------------------------------------------------
// Class         : N_PDS_GlobalAccessor
// Purpose       : Migrator utility to retrieve off processor elements
//                 of distributed objects
// Special Notes :
// Creator       : Robert Hoekstra, SNL, Parallel Compuational Sciences
// Creation Date : 06/07/00
//-----------------------------------------------------------------------------
class N_PDS_GlobalAccessor
{
public:
  // Constructors
  N_PDS_GlobalAccessor(N_PDS_Comm & comm);

  // Destructor
  ~N_PDS_GlobalAccessor();

private:
  // No public copy construction, assignment, or equality operators.

  // Copy constructor (private).
  N_PDS_GlobalAccessor(const N_PDS_GlobalAccessor & right);

  // Assignment operator (private).
  N_PDS_GlobalAccessor & operator = (const N_PDS_GlobalAccessor & right);

  // Equality operation.
  bool operator == (const N_PDS_GlobalAccessor & right) const;

  // Non-equality operation.
  bool operator != (const N_PDS_GlobalAccessor & right) const;

public:
  void registerExternGIDVector(const std::vector< std::pair< int,
                               int > > & extGIDVector);

  // Uses data from TopoLSUtil::ExternSVarList to generate migration plan using
  // Zoltan utilities.
  void generateMigrationPlan();

  // Migrates off processor elements of MultiVector using plan generated by
  // above method.
  void migrateMultiVector(Xyce::Linear::MultiVector * mVector);

  // Migrator for integer array (currently to be used for global index reset).
  void migrateIntArray(std::map< int, int > & sendMap, std::map< int,
                       int > & recvMap);

  // Migrator for integer array (currently to be used for global index reset).
  void migrateIntVecs(std::map< int, std::vector< int > > & sendMap, std::map< int,
                      std::vector< int > > & recvMap);

private:

  // PDS Comm
  N_PDS_Comm & pdsComm_;

  // Vector listing extern GIDs and their respective procs needed for nonlocal
  // dependencies.
  std::vector< std::pair< int, int > > externGIDVector_;

  // Number of objects to be received.
  int numReceiveObjs_;

  // Vector listing receive GIDs.
  std::vector< std::pair< int, int > > receiveGIDVector_;

  int * arrayReceiveGIDs_;
  int * arrayReceiveProcs_;

  // Pointer to the receiving buffer.
  char * recvBuf_;
  int recvBufSize_;

  // Number of objects to be sent.
  int numSendObjs_;

  // Vector listing sending GIDs.
  std::vector< std::pair< int, int > > sendGIDVector_;

  int * arraySendGIDs_;
  int * arraySendProcs_;

  // Pointer to the sending buffer.
  char * sendBuf_;
  int sendBufSize_;

  Epetra_Distributor * distributor_;
};

//-----------------------------------------------------------------------------
// Function      : N_PDS_GlobalAccessor::registerExternGIDVector
// Purpose       : Registers vector of external GIDs and procs.
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Compuational Sciences
// Creation Date : 06/07/00
//-----------------------------------------------------------------------------
inline void N_PDS_GlobalAccessor::registerExternGIDVector(
  const std::vector< std::pair< int, int > > & extGIDVector)
{

  // First, clear the vector...
  externGIDVector_.clear();

  // Then, assign the GIDs and processors.
  externGIDVector_.assign(extGIDVector.begin(), extGIDVector.end());
}

#endif

