//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_TIA_BackwardDifferentiation15.h,v $
//
// Purpose       : This file defines the classes for the backward
//                 differentiation, order 1-5 method.
//
// Special Notes :
//
// Creator       : Todd Coffey
//
// Creation Date : 2/16/04
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.71 $
//
// Revision Date  : $Date: 2015/06/05 20:49:43 $
//
// Current Owner  : $Author: dgbaur $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_TIA_BackwardDifferentiation15_h
#define Xyce_N_TIA_BackwardDifferentiation15_h

// ---------- Standard Includes ----------
#include <N_ANP_fwd.h>
#include <N_TIA_fwd.h>

#include <N_PDS_ParMap.h>
#include <N_TIA_DataStore.h>
#include <N_TIA_StepErrorControl.h>
#include <N_TIA_TimeIntegrationMethods.h>

#include <N_UTL_Math.h>
#include <N_UTL_IndexPair.h>

namespace Xyce {
namespace TimeIntg {

//-----------------------------------------------------------------------------
// Class         : BackwardDifferentiation15
// Purpose       : Backward Differentiation Formula(of order 1-5)
//                 Integration Class
//                (derived from TimeIntegrationMethod)
// Special Notes :
// Creator       : Todd Coffey, SNL
// Creation Date : 2/16/04
//-----------------------------------------------------------------------------
class BackwardDifferentiation15 : public TimeIntegrationMethod
{
public:
  static const int type = 6;
  static const char *name;

  static TimeIntegrationMethod * factory(
    const TIAParams &   tia_params,
    StepErrorControl &  step_error_control,
    DataStore &         data_store);

  BackwardDifferentiation15(
    const TIAParams &   tia_params,
    StepErrorControl &  step_error_control,
    DataStore &         data_store);

  ~BackwardDifferentiation15()
  {}

  const char *getName() const /* override */
  {
    return "Backward Differentiation 15";
  }

  // Predict solution at next time point(BDF1-5) /* override */.
  void obtainPredictor() /* override */;

  // Evaluate the predictor derivative formula(BDF1-5) /* override */.
  void obtainPredictorDeriv() /* override */
  {}

  // Evaluate the corrector derivative formula(BDF1-5) /* override */.
  void obtainCorrectorDeriv() /* override */
  {}

  // Compute an estimate of the error in the integration step(BDF1-5) /* override */.
  void updateDerivsBlock(
    const std::list< IndexPair > & solGIDList,
    const std::list< IndexPair > & staGIDList) /* override */
  {}

  // Compute an estimate of the error in the integration step(BDF1-5) /* override */.
  double computeErrorEstimate() /* override */ const
  {
    return sec.ck_*ds.WRMS_errorNorm() /* override */;
  }

  // Interpolate solution approximation at prescribed time point(BDF1-5) /* override */.
  bool interpolateSolution(
    double timepoint,
    Linear::Vector * tmpSolVectorPtr, std::vector<Linear::Vector*> & historyVec) /* override */;

  // Print output using interpolation when order is high
  bool printOutputSolution(
    Analysis::OutputMgrAdapter &        outputManagerAdapter,
    const TIAParams &                   tia_params,
    const double                        time,
    Linear::Vector *                      solnVecPtr,
    const bool                          doNotInterpolate,
    const std::vector<double> &         outputInterpolationTimes,
    bool                                skipPrintLineOutput) /* override */;

  // Print transient output from MPDE simulation
  bool printMPDEOutputSolution(
    Analysis::OutputMgrAdapter & outputManagerAdapter,
    const double time,
    Linear::Vector * solnVecPtr,
    const std::vector<double> & fastTimes ) /* override */;

  // Print transient output from WaMPDE simulation
  bool printWaMPDEOutputSolution(
    Analysis::OutputMgrAdapter & outputManagerAdapter,
    const double time,
    Linear::Vector * solnVecPtr,
    const std::vector<double> & fastTimes,
    const int phiGID ) /* override */;

  // .SAVE output using interpolation when order is high
  bool saveOutputSolution(
    Parallel::Machine                   comm,
    IO::InitialConditionsManager &      initial_conditions_manager,
    const NodeNameMap &                 node_name_map,
    const TIAParams &                   tia_params,
    Linear::Vector *                    solnVecPtr,
    const double                        saveTime,
    const bool                          doNotInterpolate) /* override */; 

  // Computes the step adjustment(BDF1-5) /* override */.
  double computeExpoStepAdjust(double stepadjust) /* override */;

  // Gets the time-integration order(BDF1-5) /* override */.
  int getOrder() const /* override */
  {
    return sec.currentOrder_;
  }

  int getNumberOfSteps() const /* override */
  {
    return sec.numberOfSteps_;
  }

  int getUsedOrder() const /* override */
  {
    return sec.usedOrder_;
  }

  int getNscsco() const /* override */
  {
    return sec.nscsco_;
  }

  int getMaxOrder() const /* override */
  {
    return sec.maxOrder_;
  }

  void getInitialQnorm(TwoLevelError & tle) const /* override */;
  void getTwoLevelError(TwoLevelError & tle) const /* override */;
  
  // This is for new-DAE.
  void updateStateDeriv() /* override */;
  
  // calculates dQ/dt component of store vector and adds it to store vector
  void updateLeadCurrent() /* override */;

  // calculates dQ/dt component of the lead current Q vector and adds it to lead current vector
  void updateLeadCurrentVec() /* override */;


  // Returns the current partial time derivative for either the solution or
  // state vector.
  double partialTimeDeriv() /* override */ const;

  // Gets the leading coefficient for the specified time-integration method.
  double getLeadingCoeff() const /* override */
  {
    return leadingCoeff;
  }

  // sets the leading coefficient for the specified time-integration method.
  void setLeadingCoeff(double & LC) /* override */
  {
    leadingCoeff = LC;
  }

  // 03/08/04 tscoffe:  New functions necessary for BackwardDifferentiation15
  // Evaluate corrector residual for nonlinear solver
  void obtainResidual() /* override */;

  // obtain residuals for transient sensitivity analysis
  void obtainSensitivityResiduals() /* override */;

  // obtain final derivatives w.r.t. time for transient sensitivity analysis
  void loadFinalSensitivityDerivatives() /* override */
  {}

  // Evaluate corrector Jacobian for nonlinear solver
  void obtainJacobian() /* override */;

  // Update history array after a successful step 
  void updateHistory() /* override */;

  // Restore history array after a failed step
  void restoreHistory() /* override */;

  // Update method coefficients
  void updateCoeffs() /* override */;

  // Initialize method with initial solution & step-size
  void initialize(const TIAParams &tia_params) /* override */;

  // setup 2-level data.
  void setTwoLevelTimeInfo() /* override */;

  // Reject a step(this turns back history and chooses new order & step-size) /* override */
  void rejectStep(const TIAParams &tia_params) /* override */;

  // Reject a step, but only restore the history, as the new step will be
  // imposed by an outside program. 
  void rejectStepForHabanero() /* override */;

  // Complete a step(this updates history and chooses new order & step-size) /* override */
  void completeStep(const TIAParams &tia_params) /* override */;

private:
  // Return min order of method(this should obey user option minorder) /* override */
  int getMinOrder()
  {
    return sec.minOrder_;
  }

  // Interpolate MPDE solution approximation at prescribed time point(BDF1-5).
  bool interpolateMPDESolution(
    std::vector<double>& timepoint,
    Linear::Vector * tmpSolVectorPtr);
  
  // Check whether to reduce order independent of local error test
  void checkReduceOrder();

  double                timept_;        ///< Keep track of last interpolation point in printMPDEOutputSolution
  DataStore &           ds;             ///< Reference to the TIA data-store object.
  StepErrorControl &    sec;            ///< Reference to step-error control object
  double                leadingCoeff;   ///< Time-integration method leading coefficient value.
};

} // namespace TimeIntg
} // namespace Xyce

#endif // Xyce_N_TIA_BackwardDifferentiation15_h
